
with Allegro.Audio_Streams;             use Allegro.Audio_Streams;
with Allegro.Digital_Samples;           use Allegro.Digital_Samples;
with System;                            use System;

package Almp3 is

    ALMP3_OK                    : constant Integer :=  0;
    ALMP3_PLAY_BUFFERTOOSMALL   : constant Integer := -1;
    ALMP3_POLL_PLAYJUSTFINISHED : constant Integer :=  1;
    ALMP3_POLL_NOTPLAYING       : constant Integer := -1;
    ALMP3_POLL_FRAMECORRUPT     : constant Integer := -2;
    ALMP3_POLL_BUFFERUNDERRUN   : constant Integer := -3;
    ALMP3_POLL_INTERNALERROR    : constant Integer := -4;

    ALMP3_PLAY_SPEED_NORMAL     : constant Integer := 1000;

    type ALMP3_MP3 is private;
    type A_mp3 is access all ALMP3_MP3;

    type ALMP3_MP3STREAM is private;
    type A_mp3stream is access all ALMP3_MP3STREAM;

    ---------------------------------------------------------------------------

    --  Returns the version string for the library.
    function Almp3_Version return String;
    pragma Postcondition( Almp3_Version'Result'Length > 0 );

    --  Creates an ALMP3_MP3 which you'll have to pass to all the other
    --  functions, where 'data' will be a pointer to the buffer containing
    --  the mp3 data and 'data_len' the size of this buffer. Note you aren't
    --  supposed to free this buffer until you have destroyed the ALMP3_MP3.

    -- return values:
    --  NULL if there ocurred an error (mostly an invalid mp3 data was passed).
    --  Other value ( != NULL ) otherwise.
    function Create_mp3( data : Address; data_len : Natural ) return A_mp3;
    pragma Postcondition( data /= Null_Address );

    --  Destroys the ALMP3_MP3 automatically stopping it. Note this function
    --  checks if the 'mp3' pointer is pointing to NULL, so a null pointer
    --  won't crash the program.
    procedure Destroy_mp3( mp3 : in out A_mp3 );
    pragma Postcondition( mp3 = null );

    --  Plays the 'mp3' ALMP3_MP3 with the given volume 'vol' (from 0 to 255)
    --  and panning 'pan' (from 0 to 255, where 0 is full left and 255 is
    --  full right). 'buffer_len' is the desired size in bytes of the
    --  buffer where the decoded data will be stored. The bigger, the less
    --  you'll have to poll the MP3, but the more time you will have to wait
    --  in order to hear the song start playing. Note that due to some
    --  mp3 format limitations, the internal (and real) buffer will be
    --  an aproximation to the 'buffer_len' given. A 'buffer_len' size between
    --  16384 and 32768 bytes (16kb and 32kb) will do in most cases.
    --
    -- return values:
    --  ALMP3_OK if no problems.
    --  ALMP3_PLAY_BUFFERTOOSMALL if the given 'buffer_len' was not big enough.
    --
    -- special note:
    --  This function also works like a "resume" function, since it
    --  won't rewind the ALMP3_MP3 automatically. Note that once the ALMP3_MP3
    --  has reached the end when playing, it will rewind though, stoping the
    --  ALMP3_MP3 if the loop flag was set at FALSE (see ALMP3_Play_Ex_mp3())
    --  or continuing playing it if it was set at TRUE. Also note that this
    --  automatically stops decoding.
    function Play_mp3( mp3        : not null A_mp3;
                       buffer_len : Natural;
                       vol        : Volume_Type;
                       pan        : Pan_Type
                     ) return Integer;

    --  See ALMP3_Play_mp3(). The only addition is the 'speed' that will play the
    --  ALMP3_MP3 at a given speed (being 1000 = normal speed, 2000 = twice
    --  fast, 500 = half speed and so on) and a loop flag that can be set to
    --  not stop the ALMP3_MP3 when it has reached the end, but continue it
    --  playing from the start.
    --
    -- return values:
    --  See ALMP3_Play_mp3().
    --
    -- special note:
    --  See ALMP3_Play_mp3(). Note that you can change speed, pan, volume, etc
    --  values to the ALMP3_MP3 many times, but you will need to stop it first.
    function Play_Ex_mp3( mp3        : not null A_mp3;
                          buffer_len : Natural;
                          vol        : Volume_Type;
                          pan        : Pan_Type;
                          speed      : Positive;
                          doloop     : Boolean
                        ) return Integer;

    --  Stops the ALMP3_MP3 if it was playing.
    --
    -- special note:
    --  This function also works like a "pause" function, since it won't
    --  rewind it automatically.
    procedure Stop_mp3( mp3 : not null A_mp3 );

    --  Rewinds the ALMP3_MP3 to its start.
    --
    -- special note:
    --  This function won't automatically stop the ALMP3_MP3 if it was
    --  already playing.
    procedure Rewind_mp3( mp3 : not null A_mp3 );

    --  Does an absolute seek (from start of the mp3), given the new
    --  position either in frames, msecs, secs or bytes.
    --
    -- special note:
    --  This function won't stop the ALMP3_MP3 if it was already playing.
    procedure Seek_Abs_Frames_mp3( mp3 : not null A_mp3; frame : Natural );
    procedure Seek_Abs_Msecs_mp3( mp3 : not null A_mp3; msecs : Natural );
    procedure Seek_Abs_Secs_mp3( mp3 : not null A_mp3; secs : Natural );
    procedure Seek_Abs_Bytes_mp3( mp3 : not null A_mp3; bytes : Natural );

    --  Does a relative seek (from current position), given the new
    --  position either in frames, msecs, secs or bytes.
    --
    -- special note:
    --  This function won't stop the ALMP3_MP3 if it was already playing.
    procedure Seek_Rel_Frames_mp3( mp3 : not null A_mp3; frame : Integer );
    procedure Seek_Rel_Msecs_mp3( mp3 : not null A_mp3; msec : Integer );
    procedure Seek_Rel_Secs_mp3( mp3 : not null A_mp3; sec : Integer );
    procedure Seek_Rel_Bytes_mp3( mp3 : not null A_mp3; bytes : Integer );

    --  Adjust the ALMP3_MP3 parameters when it is already playing.
    procedure Adjust_mp3( mp3    : not null A_mp3;
                          vol    : Volume_Type;
                          pan    : Pan_Type;
                          speed  : Positive;
                          doloop : Natural );

    --  This function needs to be called in order to keep the ALMP3_MP3
    --  playing properly, since the mp3s need to be decoded at real time
    --  (either that, or to a huge memory buffer).
    --
    -- return values:
    --  ALMP3_OK if there were no error.
    --  ALMP3_POLL_PLAYJUSTFINISHED (only once) when the file has JUST
    --    finished playing.
    --  ALMP3_POLL_NOTPLAYING if the file is not playing.
    --  ALMP3_POLL_FRAMECORRUPT if one of the frames is corrupt.
    --  ALMP3_POLL_INTERNALERROR if an internal error happened.
    --
    -- special note:
    --  If you don't want (or can't) poll all the time, you can use
    --  automatic polling (see below), but I don't recommend this since
    --  this can be very unstable, specially under DOS (although I've
    --  never experienced such problems myself, but better warn ;).
    function Poll_mp3( mp3 : not null A_mp3 ) return Integer;

    --  Creates an allegro interrupt that will call poll for this ALMP3_MP3
    --  each 'speed' msecs. This frees you from calling polling yourself,
    --  but I recommend this only in the case you can't call poll (because
    --  of the nature of your program) yourself at regular intervals.
    procedure Start_Autopoll_mp3( mp3 : not null A_mp3; speed : Positive );

    --  Destroys the allegro interrupt for that ALMP3_MP3.
    procedure Stop_Autopoll_mp3( mp3 : not null A_mp3 );

    --  Returns the ALMP3_MP3 current position either in frames, msecs, secs
    --  or bytes.
    function Get_Pos_Frames_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Pos_Msecs_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Pos_Secs_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Pos_Bytes_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Returns the ALMP3_MP3 length either in frames, msecs, secs or bytes.
    function Get_Length_Frames_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Length_Secs_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Length_Msecs_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Length_Bytes_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Returns the msecs per frame for the given ALMP3_MP3.
    function Get_Msecs_Per_Frame_mp3( mp3 : A_mp3 ) return Natural;

    --  Returns the ALMP3_MP3 bitrate in bits per second, NOT in kbits. That's
    --  for example 128000, 64000, 96000, etc.
    function Get_Bitrate_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Returns 3 on MPEG-audio layer III (this is, mp3) or 2 on MPEG-audio
    --  layer II (mp2).
    function Get_Layer_mp3( mp3 : not null A_mp3 ) return Natural;

    function Get_Is_Stereo_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Returns info about the wave.
    function Get_Wave_Bits_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Wave_Is_Stereo_mp3( mp3 : not null A_mp3 ) return Natural;
    function Get_Wave_Freq_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Decodes the given ALMP3_MP3 into an Allegro SAMPLE structure. Please
    --  note for big mp3s this function could generate a HUGE SAMPLE, so it
    --  is only recommended for very small mp3s like sound effects.
    --
    -- return values:
    --  NULL on error.
    --  otherwise not NULL.
    function Create_Sample_From_mp3( mp3 : not null A_mp3 ) return A_Sample;

    --  Returns a pointer to the piece of wave decoded after each poll, and
    --  the size of this buffer as well in 'buffer_size' (in bytes). Note this
    --  function will NOT automatically convert from unsinged to signed 16 bit
    --  data using the Allegro format instead of the standard format, so this
    --  data cannot be saved directly into a WAV for example without
    --  modifications. Also note in order for this function to work, the
    --  mp3 stream needs to BE playing.
    --
    -- return values:
    --  NULL if there is no wave being decoded.
    --  Else the buffer with the wave data.
    function Get_Output_Wave_mp3( mp3         : not null A_mp3;
                                  buffer_size : access Natural
                                ) return Address;

    --  Returns TRUE if the ALMP3_MP3 is currently being played or FALSE
    --  if it is not.
    function Is_Playing_mp3( mp3 : not null A_mp3 ) return Natural;

    --  Self explanatory. Note this function only works when the ALMP3_MP3
    --  is playing.
    function Is_Looping_mp3( mp3 : not null A_mp3 ) return Integer;

    --  Returns the allegro AUDIOSTREAM currently being used by the ALMP3_MP3.
    function Get_Audiostream_mp3( mp3 : not null A_mp3 ) return A_Audiostream;

    ---------------------------------------------------------------------------

    --  See almp3_create_mp3(). The only difference is that 'first_data_buffer'
    --  contains the very first buffer of len 'data_buffer_len' that you will
    --  have to give to the MP3STREAM periodically. If after this data, there
    --  won't be more (this is, this was the last buffer to be processed),
    --  pass TRUE to 'last_block'.
    function Create_mp3stream( first_data_buffer : Address;
                               data_buffer_len,
                               last_block        : Natural
                             ) return A_mp3stream;
    pragma Precondition( first_data_buffer /= Null_Address );

    function Create_mp3stream_Ex( first_data_buffer : Address;
                                  data_buffer_len,
                                  last_block,
                                  downsample,
                                  downmix           : Natural
                                ) return A_mp3stream;
    pragma Precondition( first_data_buffer /= Null_Address );

    -- See almp3_destroy_mp3().
    procedure Destroy_mp3stream( mp3 : in out A_mp3stream );
    pragma Postcondition( mp3 = null );

    -- See almp3_play_ex_mp3stream().
    function Play_mp3stream( mp3        : not null A_mp3stream;
                             buffer_len : Natural;
                             vol        : Volume_Type;
                             pan        : Pan_Type ) return Integer;

    --  See almp3_play_ex_mp3(). The only difference is that here is no 'loop'
    --  parameter. To loop an ALMP3_MP3STREAM just destroy the ALMP3_MP3STREAM
    --  and recreate it (or pass data continuosly).
    function Play_Ex_mp3stream( mp3        : not null A_mp3stream;
                                buffer_len : Natural;
                                vol        : Volume_Type;
                                pan        : Pan_Type;
                                speed      : Positive ) return Integer;

    -- See almp3_stop_mp3().
    procedure Stop_mp3stream( mp3 : not null A_mp3stream );

    --  Adjust the ALMP3_MP3STREAM parameters when it is already playing.
    --  Note it doesn't include the parameter loop because ALMP3_MP3STREAMs are
    --  always looping (or better said, they don't have two points to loop
    --  from/to).
    procedure Adjust_mp3stream( mp3   : not null A_mp3stream;
                                vol   : Volume_Type;
                                pan   : Pan_Type;
                                speed : Positive );

    -- See almp3_poll_mp3().
    -- return values:
    --  ALMP3_OK if there were no error.
    --  ALMP3_POLL_PLAYJUSTFINISHED (only once) when the file has JUST
    --    finished playing.
    --  ALMP3_POLL_NOTPLAYING if the file is not playing.
    --  ALMP3_POLL_FRAMECORRUPT if one of the frames is corrupt.
    --  ALMP3_POLL_BUFFERUNDERUN if the buffer was exhausted.
    --  ALMP3_POLL_INTERNALERROR if an internal error happened.
    function Poll_mp3stream( mp3 : not null A_mp3stream ) return Integer;

    --  See almp3_start_autopoll_mp3(). Note I discourage the use of this
    --  function with ALMP3_MP3STREAMs since in anyway you'll have to use
    --  almp3_get_mp3stream_buffer() to send data periodically.
    procedure Start_Autopoll_mp3stream( mp3 : not null A_mp3stream; speed : Positive );

    --  See almp3_stop_autopoll_mp3().
    procedure Stop_Autopoll_mp3stream( mp3 : not null A_mp3stream );

    --  If the return value is not NULL, it will return a buffer that you'll
    --  have to fill with 'buffer_data_len' (from create_mp3stream) bytes MAX
    --  of new data. You will need to use free_mp3stream_buffer() when you
    --  are finished with it.
    --
    -- return values:
    --  NULL if it doesn't need any data yet.
    --  Else the buffer to be filled.
    function Get_mp3stream_Buffer( mp3 : not null A_mp3stream ) return Address;

    --  Use whenever you are finished with the buffer returned by
    --  almp3_get_mp3stream_buffer(). If 'bytes_used' is -1 it will use
    --  the full buffer and that will mean this is not the last block
    --  of data, else if 'bytes_used' is a number, it will indicate
    --  that this is the last buffer and the number of bytes to use from that
    --  last block. Once this buffer has been played and processed,
    --  almp3_poll_mp3_stream() will return ALMP3_POLL_JUSTFINISHED.
    procedure Free_mp3stream_Buffer( mp3 : not null A_mp3stream; bytes_used : Integer );

    --  Returns the ALMP3_MP3STREAM length either in frames, msecs or secs.
    --  You will have to pass in 'total_size' the total size (in
    --  bytes) of the mp3.
    function Get_Length_Frames_mp3stream( mp3 : not null A_mp3stream; total_size : Natural ) return Natural;
    function Get_Length_Secs_mp3stream( mp3 : not null A_mp3stream; total_size : Natural ) return Natural;
    function Get_Length_Msecs_mp3stream( mp3 : not null A_mp3stream; total_size : Natural ) return Natural;
    function Get_Length_Bytes_mp3stream( mp3 : not null A_mp3stream; total_size : Natural ) return Natural;

    --  See their ALMP3_MP3 equals.
    function Get_Pos_Frames_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Pos_Msecs_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Pos_Secs_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Pos_Bytes_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Msecs_Per_Frame_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Bitrate_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Layer_mp3stream( mp3 : not null A_mp3stream ) return Natural;

    function Get_Is_Stereo_mp3stream( mp3 : not null A_mp3stream ) return Natural;

    --  Returns info about the wave.
    function Get_Wave_Bits_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Wave_Is_Stereo_mp3stream( mp3 : not null A_mp3stream ) return Natural;
    function Get_Wave_Freq_mp3stream( mp3 : not null A_mp3stream ) return Natural;

    --  Returns a pointer to the piece of wave decoded after each poll, and
    --  the size of this buffer as well in 'buffer_size' (in bytes). Note this
    --  function will NOT automatically convert from unsinged to signed 16 bit
    --  data using the Allegro format instead of the standard format, so this
    --  data cannot be saved directly into a WAV for example without
    --  modifications. Also note in order for this function to work, the
    --  ALMP3_MP3STREAM needs to BE playing.
    --
    -- return values:
    --  NULL if there is no wave being decoded.
    function Get_Output_Wave_mp3stream( mp3         : not null A_mp3stream;
                                        buffer_size : access Natural ) return Address;

    --  Returns TRUE if the ALMP3_MP3STREAM is currently being played or FALSE
    --  if it is not.
    function Is_Playing_mp3stream( mp3 : not null A_mp3stream ) return Integer;

    --  Returns the allegro AUDIOSTREAM currently being used by the
    --  ALMP3_MP3STREAM.
    --  Note that when the ALMP3_MP3STREAM isn't being played most probably
    --  it will return NULL.
    function Get_Audiostream_mp3stream( mp3 : not null A_mp3stream ) return A_Audiostream;

    --  Returns the position in bytes where you should position your stream
    --  (file, memory buffer, socket) to continue playing at that mp3 position
    --  using an absolute (from start position) seeking.
    --  You need to pass the total size (in bytes) of the stream.
    --
    -- special note:
    --  This function doesn't perform any operation or change but informs you
    --  where to read your next chunk of data.
    function Seek_Abs_Frames_mp3stream( mp3 : not null A_mp3stream; frame, total_size : Natural ) return Natural;
    function Seek_Abs_Msecs_mp3stream( mp3 : not null A_mp3stream; msecs, total_size : Natural ) return Natural;
    function Seek_Abs_Secs_mp3stream( mp3 : not null A_mp3stream; msecs, total_size : Natural ) return Natural;
    function Seek_Abs_Bytes_mp3stream( mp3 : not null A_mp3stream; bytes, total_size : Natural ) return Natural;

    --  Returns the position in bytes where you should position your stream
    --  (file, memory buffer, socket) to continue playing at that mp3 position
    --  (using a relative, from current position, seeking).
    --  You need to pass the total size (in bytes) of the stream.
    --
    -- special note:
    --  This function doesn't perform any operation or change but informs you
    --  where to read your next chunk of data.
    function Seek_Rel_Frames_mp3stream( mp3 : not null A_mp3stream; frame, total_size : Natural ) return Natural;
    function Seek_Rel_Msecs_mp3stream( mp3 : not null A_mp3stream; msec, total_size : Natural ) return Natural;
    function Seek_Rel_Secs_mp3stream( mp3 : not null A_mp3stream; sec, total_size : Natural ) return Natural;
    function Seek_Rel_Bytes_mp3stream( mp3 : not null A_mp3stream; bytes, total_size : Natural ) return Natural;

private

    type ALMP3_MP3 is null record;
    type ALMP3_MP3STREAM is null record;

    ----------------------------------------------------------------------------

    pragma Import( C, Create_mp3, "almp3_create_mp3" );
    pragma Import( C, Play_mp3, "almp3_play_mp3" );
    pragma Import( C, Stop_mp3, "almp3_stop_mp3" );
    pragma Import( C, Rewind_mp3, "almp3_rewind_mp3" );
    pragma Import( C, Seek_Abs_Frames_mp3, "almp3_seek_abs_frames_mp3" );
    pragma Import( C, Seek_Abs_Msecs_mp3, "almp3_seek_abs_msecs_mp3" );
    pragma Import( C, Seek_Abs_Secs_mp3, "almp3_seek_abs_secs_mp3" );
    pragma Import( C, Seek_Abs_Bytes_mp3, "almp3_seek_abs_bytes_mp3" );
    pragma Import( C, Seek_Rel_Frames_mp3, "almp3_seek_rel_frames_mp3" );
    pragma Import( C, Seek_Rel_Msecs_mp3, "almp3_seek_rel_msecs_mp3" );
    pragma Import( C, Seek_Rel_Secs_mp3, "almp3_seek_rel_secs_mp3" );
    pragma Import( C, Seek_Rel_Bytes_mp3, "almp3_seek_rel_bytes_mp3" );
    pragma Import( C, Adjust_mp3, "almp3_adjust_mp3" );
    pragma Import( C, Poll_mp3, "almp3_poll_mp3" );
    pragma Import( C, Start_Autopoll_mp3, "almp3_start_autopoll_mp3" );
    pragma Import( C, Stop_Autopoll_mp3, "almp3_stop_autopoll_mp3" );
    pragma Import( C, Get_Pos_Frames_mp3, "almp3_get_pos_frames_mp3" );
    pragma Import( C, Get_Pos_Msecs_mp3, "almp3_get_pos_msecs_mp3" );
    pragma Import( C, Get_Pos_Secs_mp3, "almp3_get_pos_secs_mp3" );
    pragma Import( C, Get_Pos_Bytes_mp3, "almp3_get_pos_bytes_mp3" );
    pragma Import( C, Get_Length_Frames_mp3, "almp3_get_length_frames_mp3" );
    pragma Import( C, Get_Length_Secs_mp3, "almp3_get_length_secs_mp3" );
    pragma Import( C, Get_Length_Msecs_mp3, "almp3_get_length_msecs_mp3" );
    pragma Import( C, Get_Length_Bytes_mp3, "almp3_get_length_bytes_mp3" );
    pragma Import( C, Get_Msecs_Per_Frame_mp3, "almp3_get_msecs_per_frame_mp3" );
    pragma Import( C, Get_Bitrate_mp3, "almp3_get_bitrate_mp3" );
    pragma Import( C, Get_Layer_mp3, "almp3_get_layer_mp3" );
    pragma Import( C, Get_Is_Stereo_mp3, "almp3_get_is_stereo_mp3" );
    pragma Import( C, Get_Wave_Bits_mp3, "almp3_get_wave_bits_mp3" );
    pragma Import( C, Get_Wave_Is_Stereo_mp3, "almp3_get_wave_is_stereo_mp3" );
    pragma Import( C, Get_Wave_Freq_mp3, "almp3_get_wave_freq_mp3" );
    pragma Import( C, Create_Sample_From_mp3, "almp3_create_sample_from_mp3" );
    pragma Import( C, Get_Output_Wave_mp3, "almp3_get_output_wave_mp3" );
    pragma Import( C, Is_Playing_mp3, "almp3_is_playing_mp3" );
    pragma Import( C, Is_Looping_mp3, "almp3_is_looping_mp3" );
    pragma Import( C, Get_Audiostream_mp3, "almp3_get_audiostream_mp3" );

    pragma Import( C, Create_mp3stream, "almp3_create_mp3stream" );
    pragma Import( C, Create_mp3stream_Ex, "almp3_create_mp3stream_Ex" );
    pragma Import( C, Play_mp3stream, "almp3_play_mp3stream" );
    pragma Import( C, Play_Ex_mp3stream, "almp3_play_ex_mp3stream" );
    pragma Import( C, Stop_mp3stream, "almp3_stop_mp3stream" );
    pragma Import( C, Adjust_mp3stream, "almp3_adjust_mp3stream" );
    pragma Import( C, Poll_mp3stream, "almp3_poll_mp3stream" );
    pragma Import( C, Start_Autopoll_mp3stream, "almp3_start_autopoll_mp3stream" );
    pragma Import( C, Stop_Autopoll_mp3stream, "almp3_stop_autopoll_mp3stream" );
    pragma Import( C, Get_mp3stream_Buffer, "almp3_get_mp3stream_buffer" );
    pragma Import( C, Free_mp3stream_Buffer, "almp3_free_mp3stream_buffer" );
    pragma Import( C, Get_Length_Frames_mp3stream, "almp3_get_length_frames_mp3stream" );
    pragma Import( C, Get_Length_Secs_mp3stream, "almp3_get_length_secs_mp3stream" );
    pragma Import( C, Get_Length_Msecs_mp3stream, "almp3_get_length_msecs_mp3stream" );
    pragma Import( C, Get_Length_Bytes_mp3stream, "almp3_get_length_bytes_mp3stream" );
    pragma Import( C, Get_Pos_Frames_mp3stream, "almp3_get_pos_frames_mp3stream" );
    pragma Import( C, Get_Pos_Msecs_mp3stream, "almp3_get_pos_msecs_mp3stream" );
    pragma Import( C, Get_Pos_Secs_mp3stream, "almp3_get_pos_secs_mp3stream" );
    pragma Import( C, Get_Pos_Bytes_mp3stream, "almp3_get_pos_bytes_mp3stream" );
    pragma Import( C, Get_Msecs_Per_Frame_mp3stream, "almp3_get_msecs_per_frame_mp3stream" );
    pragma Import( C, Get_Bitrate_mp3stream, "almp3_get_bitrate_mp3stream" );
    pragma Import( C, Get_Layer_mp3stream, "almp3_get_layer_mp3stream" );
    pragma Import( C, Get_Is_Stereo_mp3stream, "almp3_get_is_stereo_mp3stream" );
    pragma Import( C, Get_Wave_Bits_mp3stream, "almp3_get_wave_bits_mp3stream" );
    pragma Import( C, Get_Wave_Is_Stereo_mp3stream, "almp3_get_wave_is_stereo_mp3stream" );
    pragma Import( C, Get_Wave_Freq_mp3stream, "almp3_get_wave_freq_mp3stream" );
    pragma Import( C, Get_Output_Wave_mp3stream, "almp3_get_output_wave_mp3stream" );
    pragma Import( C, Is_Playing_mp3stream, "almp3_is_playing_mp3stream" );
    pragma Import( C, Get_Audiostream_mp3stream, "almp3_get_audiostream_mp3stream" );
    pragma Import( C, Seek_Abs_Frames_mp3stream, "almp3_seek_abs_frames_mp3stream" );
    pragma Import( C, Seek_Abs_Msecs_mp3stream, "almp3_seek_abs_msecs_mp3stream" );
    pragma Import( C, Seek_Abs_Secs_mp3stream, "almp3_seek_abs_secs_mp3stream" );
    pragma Import( C, Seek_Abs_Bytes_mp3stream, "almp3_seek_abs_bytes_mp3stream" );
    pragma Import( C, Seek_Rel_Frames_mp3stream, "almp3_seek_rel_frames_mp3stream" );
    pragma Import( C, Seek_Rel_Msecs_mp3stream, "almp3_seek_rel_msecs_mp3stream" );
    pragma Import( C, Seek_Rel_Secs_mp3stream, "almp3_seek_rel_secs_mp3stream" );
    pragma Import( C, Seek_Rel_Bytes_mp3stream, "almp3_seek_rel_bytes_mp3stream" );

end Almp3;
