
pragma Warnings( Off );
with Ada.Exceptions;                    use Ada.Exceptions;
with Alfont;
with Allegro;                           use Allegro;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Configuration;             use Allegro.Configuration;
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Allegro.Timers;                    use Allegro.Timers;
with Almp3;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Entities;
with Events.Manager;
with Font_API;                          use Font_API;
with Game_Views;                        use Game_Views;
with Game_Views.Window_Hooks;
with LoadPNG;
with Preferences;                       use Preferences;
with Renderers;                         use Renderers;
with Scaling;                           use Scaling;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Themes.Palette;
with Tiles;
with Zip;

package body Applications is

    MIN_XRES      : constant := 16;
    MIN_YRES      : constant := 16;

    MIN_XRES_WIN  : constant := 64;
    MIN_YRES_WIN  : constant := 64;
    MIN_XRES_FULL : constant := 640;
    MIN_YRES_FULL : constant := 480;

    DEFAULT_DEPTH : constant := 16;

    MAX_SCALE     : constant := 4;

    ----------------------------------------------------------------------------

    protected Global_Instance is

        -- Creates the global application if it doesn't exist. An exception is
        -- raised on error.
        procedure Create( app : out A_Application );
        pragma Postcondition( app /= null );

        -- Deletes the global application.
        procedure Delete;

        -- Returns the global application; may return null.
        function Get return A_Application;

        -- Sets the function used to allocate the global application.
        procedure Set_Allocator( alloc : not null Allocator );

    private
        g_app    : A_Application := null;
        allocate : Allocator := null;
    end Global_Instance;

    ----------------------------------------------------------------------------

    protected body Global_Instance is

        procedure Create( app : out A_Application ) is
        begin
            if g_app = null then
                -- Initialize systems here
                if allocate /= null then
                    g_app := allocate.all;
                    pragma Debug( Dbg( "Created application <" &
                                       g_app.Get_Window_Title & '>',
                                       D_INIT, INFO ) );
                else
                    raise USE_ERROR with "Application allocator not registered";
                end if;
            end if;
            app := g_app;
        end Create;

        ------------------------------------------------------------------------

        procedure Delete is
        begin
            if g_app /= null then
                pragma Debug( Dbg( "Deleting application", D_INIT, INFO ) );
                Delete( g_app );
            end if;
        end Delete;

        ------------------------------------------------------------------------

        function Get return A_Application is
        begin
            return g_app;
        end Get;

        ------------------------------------------------------------------------

        procedure Set_Allocator( alloc : not null Allocator ) is
        begin
            allocate := alloc;
        end Set_Allocator;

    end Global_Instance;

    --==========================================================================

    -- Creates the singleton Application object if it does not yet exist.
    procedure Create_Application is
        app : A_Application;
    begin
        app := Global_Instance.Get;
        if app = null then
            Global_Instance.Create( app );
        end if;
    end Create_Application;

    ----------------------------------------------------------------------------

    -- Deletes the singleton Application object
    procedure Delete_Application is
    begin
        Global_Instance.Delete;
    end Delete_Application;

    ----------------------------------------------------------------------------

    -- Returns a reference to the singleton Application object, creating it if
    -- it does not already exist.
    function Get_Application return A_Application is
        app : A_Application;
    begin
        app := Global_Instance.Get;
        if app = null then
            Global_Instance.Create( app );
        end if;
        return app;
    end Get_Application;

    ----------------------------------------------------------------------------

    -- Registers an allocator function to create an application instance
    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Global_Instance.Set_Allocator( allocate );
    end Register_Allocator;

    --==========================================================================

    -- Closes the application and release all resources.
    not overriding
    procedure Close( this : access Application ) is
    begin
        if not this.inited then
            return;
        end if;

        Delete( this.theme );
        Delete( this.game );
        Delete( this.pman );

        Entities.Finalize;
        Scaling.Finalize;
        Font_API.Finalize;
        if this.graphical then
            if this.interactive then
                pragma Debug( Dbg( "Shutting down keyboard module", D_INIT, INFO ) );
                Clear_Keybuf;
                Remove_Keyboard;
            end if;

            pragma Debug( Dbg( "Shutting down timer module", D_INIT, INFO ) );
            Remove_Timer;
            Game_Views.Window_Hooks.Finalize;
            if Set_Display_Switch_Mode( SWITCH_NONE ) /= 0 then
                null;
            end if;
        end if;

        Preferences.Finalize;
        Tiles.Finalize;

        pragma Debug( Dbg( "Shutting down Allegro", D_INIT, INFO ) );
        Allegro_Exit;
        pragma Debug( Dbg( "Application closed", D_INIT, INFO ) );
    end Close;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Application ) is
    begin
        Limited_Object(this.all).Construct;
        this.lock := new Locking_Object;

        -- Initialize systems prior to application/Allegro init
        Tiles.Initialize;
    end Construct;

    ----------------------------------------------------------------------------

    -- Deletes the members of this Application object that were created by
    -- 'construct'. Any members created during 'init' or 'run' should
    -- be deleted at 'close' instead.
    overriding
    procedure Delete( this : in out Application ) is
    begin
        Delete( this.lock );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Theme( this : not null access Application'Class ) return A_Theme is
    begin
        if this.theme = null then
            this.theme := Create_Theme;
        end if;
        return this.theme;
    end Get_Theme;

    ----------------------------------------------------------------------------

    -- Initializes the application by setting up graphics and hardware.
    not overriding
    function Init( this         : access Application;
                   app_xres,
                   app_yres     : Natural;
                   app_scale    : Positive;
                   app_windowed : Boolean ) return Boolean is
    begin
        if this.inited then
            return True;
        end if;

        -- Initialize Allegro
        if this.graphical then
            if Allegro_Initialize /= 0 then
                Allegro_Message( "Error initializing Allegro" );
                pragma Debug( Dbg( "Application.Init Error: Allegro_Initialize: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                raise INIT_EXCEPTION;
            end if;
        else
            if Allegro_Initialize_Cmdline /= 0 then
                Allegro_Message( "Error initializing Allegro" );
                pragma Dbg( Dbg( "Application.Init Error: Allegro_Initialize: " &
                                 Allegro_Error, D_INIT, ERROR ) );
                raise INIT_EXCEPTION;
            end if;
        end if;

        -- Initialize Preferences
        pragma Debug( Dbg( "Application.Init Info: Using config <" &
                           A_Application(this).Get_Name & ".cfg>", D_INIT, INFO ) );
        Preferences.Initialize( Home_Directory & A_Application(this).Get_Name & ".cfg" );

        -- Detect hardware and OS
        pragma Debug( Dbg( "", D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Vendor      : " & CPU_Vendor, D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Model       : " & To_String( CPU_Family, CPU_Model ), D_VERSION, INFO ) );
        pragma Debug( Dbg( "CPU Capabilities: " & To_String( CPU_Capabilities ), D_VERSION, INFO ) );
        pragma Debug( this.graphical,
                      Dbg( "Operating System: " & To_String( OS_Type ) & ' ' &
                                                  Image( OS_Version ) & '.' & Image( OS_Revision ),
                           D_VERSION, INFO ) );
        pragma Debug( not this.graphical,
                      Dbg( "Operating System: Undetermined", D_VERSION, INFO ) );
#if WINDOWS then
        pragma Debug( Dbg( "Compiler        : " & Standard'Compiler_Version, D_VERSION, INFO ) );
#end if;
        pragma Debug( Dbg( "", D_VERSION, INFO ) );

        -- Detect library versions
        pragma Debug( Dbg( "Libraries in use:", D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Allegro.Allegro_Id, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Almp3.Almp3_Version, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & Alfont.Alfont_Version, D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "Zip-Ada " & Zip.version & " (" & Zip.reference & ')', D_VERSION, INFO ) );
        pragma Debug( Dbg( "  " & "LoadPNG " & LoadPNG.LoadPNG_Version, D_VERSION, INFO ) );
        pragma Debug( Dbg( "", D_VERSION, INFO ) );

        -- Set window title
        Set_Window_Title( A_Application(this).Get_Window_Title );

        -- Set color depth
        declare
            depth : Integer := Get_Config_Int( "application", "bpp", Desktop_Color_Depth );
        begin
            if depth /= 15 and then
               depth /= 16 and then
               depth /= 24 and then
               depth /= 32
            then
                depth := Desktop_Color_Depth;
                if depth = 0 then
                    -- unable to get the desktop color depth!
                    depth := DEFAULT_DEPTH;
                    pragma Debug( this.graphical,
                                  Dbg( "Application.Init Warning: " &
                                       "Unable to get desktop color depth",
                                       D_INIT, WARNING ) );
                end if;
            end if;
            pragma Debug( Dbg( "Application.Init: Color depth: " &
                               Image( depth ) & " bpp", D_INIT, INFO ) );
            Set_Color_Depth( depth );
        end;

        -- Set window resolution and content scaling factor
        declare
            dxRes, dyRes,                -- current desktop resolution
            oxRes, oyRes,                -- original requested resolution
            xRes,  yRes    : Integer;    -- screen resolution to attempt
            scale          : Integer;
            gfxmode        : Gfx_Detect_Type;
            iapp_windowed  : Integer := 0;
            success        : Boolean;
        begin
            scale := Constrain( Get_Config_Int( "application", "scale", app_scale ), 1, MAX_SCALE );

            Desktop_Resolution( dxRes, dyRes, success );
            if not success then
                -- unable to get the desktop resolution!
                dxRes := app_xres;
                dyRes := app_yres;
                pragma Debug( this.graphical,
                              Dbg( "Application.Init Warning: " &
                                   "Unable to get desktop resolution",
                                   D_INIT, WARNING ) );
            end if;

            if app_windowed then
                iapp_windowed := 1;
            end if;
            if Get_Config_Int( "application", "windowed", iapp_windowed ) = 1 then
                -- if we're trying windowed mode then the default resolution is
                -- the minimum resolution.
                gfxmode := GFX_AUTODETECT_WINDOWED;
                oxRes := Constrain( Get_Config_Int( "application", "xres", app_xres ), MIN_XRES, dxRes );
                oyRes := Constrain( Get_Config_Int( "application", "yres", app_yres ), MIN_YRES, dyRes );
                xRes := Constrain( oxRes * scale, MIN_XRES_WIN, dxRes );
                yRes := Constrain( oyRes * scale, MIN_YRES_WIN, dyRes );
            else
                -- if we're trying full screen mode then the default resolution
                -- is the desktop's resolution.
                gfxmode := GFX_AUTODETECT_FULLSCREEN;
                oxRes := Constrain( Get_Config_Int( "application", "xres", app_xres ), MIN_XRES, dxRes );
                oyRes := Constrain( Get_Config_Int( "application", "yres", app_yres ), MIN_YRES, dyRes );
                xRes := Constrain( oxRes * scale, MIN_XRES_FULL, dxRes );
                yRes := Constrain( oyRes * scale, MIN_YRES_FULL, dyRes );
            end if;

            if not this.graphical then
                -- not a graphical application; don't set a graphics mode
                this.xres := oxRes;
                this.yres := oyRes;
                this.scale := scale;
            elsif Set_Gfx_Mode( gfxmode, xRes, yRes, 0, 0 ) = 0 or else
                  Set_Gfx_Mode( GFX_AUTODETECT, xRes, yRes, 0, 0 ) = 0 or else
                  Set_Gfx_Mode( GFX_AUTODETECT_FULLSCREEN, xRes, yRes, 0, 0 ) = 0 or else
                  Set_Gfx_Mode( GFX_AUTODETECT_FULLSCREEN, dxRes, dyRes, 0, 0 ) = 0 or else
                  Set_Gfx_Mode( GFX_AUTODETECT, app_xres, app_yres, 0, 0 ) = 0
            then
                -- check the actual resolution we set and adjust scale accordingly
                declare
                    swidth  : constant Integer := Get_Screen_W;
                    sheight : constant Integer := Get_Screen_H;
                begin
                    pragma Debug( Dbg( "Application.Init screen resolution:", D_INIT, INFO ) );
                    pragma Debug( Dbg( "  requested: " &
                                       Image( oxRes ) & " x " &
                                       Image( oyRes ) & " * " &
                                       Image( scale ) & " > [" &
                                       Image( oxRes * scale ) & " x " &
                                       Image( oyRes * scale ) & "]",
                                       D_INIT, INFO ) );

                    -- reduce scaling if screen size is too small
                    for i in reverse 1..scale loop
                        scale := i;
                        exit when oxRes * scale <= swidth and then oyRes * scale <= sheight;
                    end loop;

                    -- increase scaling if screen size is too big
                    for i in scale..MAX_SCALE loop
                        exit when oxRes * i > swidth or else oyRes * i > sheight;
                        scale := i;
                    end loop;

                    this.scale := scale;

                    -- set the resolution to be what the user asked for, scaled no
                    -- more than maximum. this may leave a black border around the
                    -- window if it doesn't fill the screen but it will not modify
                    -- the aspect ratio of the requested resolution.
                    this.xres := oxRes;
                    this.yres := oyRes;

                    -- modify the resolution to be 1/scale of the screen size
                    -- so that the window always fills the screen. if the window is
                    -- to small to fill the screen at maximum scale, its resolution
                    -- will is increased. the aspect ratio may be different than
                    -- what the user requested.
                    --this.xres := swidth / scale;
                    --this.yres := sheight / scale;

                    pragma Debug( Dbg( "     actual: " &
                                       Image( this.xres ) & " x " &
                                       Image( this.yres ) & " * " &
                                       Image( this.scale ) & " > [" &
                                       Image( swidth ) & " x " &
                                       Image( sheight ) & "]",
                                       D_INIT, INFO ) );
                end;
            else
                -- failed setting a graphics mode
                pragma Debug( Dbg( "Application.Init Error: Set_Gfx_Mode: " &
                                    Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error setting graphics mode." );
                raise INIT_EXCEPTION with "Error setting graphics mode";
            end if;
        end;

        -- Set display switching mode
        if this.graphical then
            if Set_Display_Switch_Mode( SWITCH_BACKGROUND ) /= 0 and then
               Set_Display_Switch_Mode( SWITCH_BACKAMNESIA ) /= 0
            then
                pragma Debug( Dbg( "Application.Init Error: Set_Display_Switch_Mode: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error setting display switch mode." );
                raise INIT_EXCEPTION with "Error setting display switch mode.";
            end if;
        end if;

        -- Install window event hooks
        if this.graphical then
            Game_Views.Window_Hooks.Initialize;
        end if;

        -- Install timers and input handlers
        if this.graphical then
            if Install_Timer /= 0 then
                pragma Debug( Dbg( "Application.Init Error: Install_Timer: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error installing timers." );
                raise INIT_EXCEPTION with "Error installing timers";
            elsif this.interactive and then
                  Install_Keyboard /= 0
            then
                pragma Debug( Dbg( "Application.Init Error: Install_Keyboard: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error installing keyboard support." );
                raise INIT_EXCEPTION with "Error installing keyboard support";
            end if;
        end if;

        -- Setup mouse driver and cursor
        if this.graphical and then this.interactive and then this.use_mouse then
            if Install_Mouse < 0 then
                pragma Debug( Dbg( "Application.Init Error: Install_Mouse: " &
                                   Allegro_Error, D_INIT, ERROR ) );
                Allegro_Message( "Error installing mouse support." );
                raise INIT_EXCEPTION with "Error installing mouse support";
            end if;
            if Is_Windowed_Mode then
                if Show_OS_Cursor( MOUSE_CURSOR_ARROW ) /= 0 then
                    Renderers.Use_Software_Mouse;
                    Set_Mouse_Cursor_Bitmap( MOUSE_CURSOR_ARROW, null );
                    pragma Debug( Dbg( "Application.Init Warning: " &
                                       "Unable to set OS cursor", D_INIT, WARNING ) );
                end if;
            else
                Renderers.Use_Software_Mouse;
                Set_Mouse_Cursor_Bitmap( MOUSE_CURSOR_ARROW, null );
                Set_Mouse_Speed( Constrain( Get_Config_Int( "application", "mouse_speed", 4 ), 1, 6 ),
                                 Constrain( Get_Config_Int( "application", "mouse_speed", 4 ), 1, 6 ) );
            end if;
        end if;

        -- Initialize systems after Allegro
        begin
            Font_API.Initialize;
        exception
            when e : others =>
                pragma Debug( Dbg( "Application.Init Error: Font_API.Initialize: " &
                                   Exception_Name( e ) & " - " & Exception_Message( e ),
                                   D_INIT, ERROR ) );
                Allegro_Message( "Error initializing font support." );
                raise INIT_EXCEPTION with "Error initializing font support";
        end;
        Scaling.Initialize( Get_Color_Depth );
        Themes.Palette.Initialize;
        Entities.Initialize;

        this.inited := True;
        return True;
    exception
        when e : others =>
            pragma Debug( Dbg( "Application.Init Exception: ...", D_INIT, ERROR ) );
            pragma Debug( Dbg( e, D_INIT, ERROR ) );

            -- because init() failed, close() won't be executed so we need to
            -- uninitialize what may have succeeded, in reverse order.
            -- finalizing a system that has not been initialized should be ok.
            Entities.Finalize;
            Font_API.Finalize;
            if this.graphical then
                if this.interactive then
                    Remove_Keyboard;
                end if;
                Remove_Timer;
                Game_Views.Window_Hooks.Finalize;
                if Set_Display_Switch_Mode( SWITCH_NONE ) /= 0 then
                    null;
                end if;
            end if;
            Preferences.Finalize;
            Allegro_Exit;
            return False;
    end Init;

    ----------------------------------------------------------------------------

    not overriding
    procedure Run( this : access Application; returnCode : in out Integer ) is
        view : A_Game_View;
    begin
        if not this.inited then
            raise USE_ERROR with "Application not initialized";
        elsif this.ran then
            raise USE_ERROR with "Application already ran";
        else
            this.ran := True;
        end if;

        pragma Debug( Dbg( "** Running application **", D_INIT, INFO ) );

        this.game := Create_Game;

        view := Create_Game_View( this.xres, this.yres, this.scale );
        this.game.Add_View( view );           -- consumes 'view'

        this.game.Start;
        Events.Manager.Run;
        this.game.Stop;

        returnCode := this.errorCode;
        pragma Debug( Dbg( "** Application run complete **", D_INIT, INFO ) );
    end Run;

    ----------------------------------------------------------------------------

    procedure Stop( this      : not null access Application'Class;
                    errorCode : Integer := NO_ERROR ) is
    begin
        Events.Manager.Stop;
        this.errorCode := errorCode;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Application ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

end Applications;
