
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Streams.Stream_IO;             use Ada.Streams.Stream_IO;
with Allegro.Files.Memory;              use Allegro.Files.Memory;
with Archives;                          use Archives;
--with Debugging;                         use Debugging;
with Resources.Images;                  use Resources.Images;
with Streams.Buffers;                   use Streams.Buffers;
with Support.Paths;                     use Support.Paths;
with Zip_Streams;                       use Zip_Streams;
with UnZip.Streams;                     use UnZip.Streams;

pragma Elaborate_All( Archives );

package body Archives.Zip_Archives is

    -- Loads a .zip from a physical path. Raises an exception on error.
    function Load_Zip_Path( path : String ) return A_Archive is
        this : A_Archive := new Zip_Archive;
    begin
        this.Construct( path );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Load_Zip_Path;

    ----------------------------------------------------------------------------

    -- Loads a .zip from a Resource_File. Raises an exception on error.
    function Load_Zip_Res( resource : not null A_Resource_File ) return A_Archive is
        this : A_Archive := new Zip_Archive;
    begin
        Zip_Archive(this.all).Construct( resource );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Load_Zip_Res;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Zip_Archive; path : String ) is
        buffer : Zip_Streams.Array_Streams.A_SEA := null;
        ft     : File_Type;
    begin
        Archive(this.all).Construct( path );

        -- Read the file contents into memory
        begin
            Open( ft, In_File, path );
            buffer := new Stream_Element_Array(1..Stream_Element_Offset(Size( ft )));
            Stream_Element_Array'Read( Stream( ft ), buffer.all );
            Close( ft );
        exception
            when others =>
                raise FILE_NOT_FOUND with "File not found";
        end;

        -- Cache the zip info catalog
        begin
            this.zipArray := new Array_Stream;
            this.zipArray.Set( buffer );
            Load( this.zipInfo, Zipstream_Class(this.zipArray) );
        exception
            when e : others =>
                raise READ_ERROR with "Error reading archive: " &
                                      Exception_Name( e ) & " - " &
                                      Exception_Message( e );
        end;
    exception
        when others =>
            Delete( buffer );
            if Is_Open( ft ) then
                Close( ft );
            end if;
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Zip_Archive;
                         resource : not null A_Resource_File ) is
        buffer : Zip_Streams.Array_Streams.A_SEA := null;
        stream : A_Buffer_Stream;
    begin
        Archive(this.all).Construct( resource.Get_Id );

        -- Copy the file contents in memory
        buffer := new Stream_Element_Array(1..Stream_Element_Offset(resource.Size));
        resource.Get_Stream( stream );
        Stream_Element_Array'Read( stream, buffer.all );
        Close( stream );

        -- Cache the zip info catalog
        begin
            this.zipArray := new Array_Stream;
            this.zipArray.Set( buffer );
            Load( this.zipInfo, Zipstream_Class(this.zipArray) );
        exception
            when e : others =>
                raise READ_ERROR with "Error reading archive: " &
                                      Exception_Name( e ) & " - " &
                                      Exception_Message( e );
        end;
    exception
        when others =>
            Close( stream );
            Delete( buffer );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Zip_Archive ) is
    begin
        Delete( this.zipInfo );
        Delete( this.zipArray );
        Archive(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    function Exists( this     : access Zip_Archive;
                     filename : String ) return Boolean is
        index        : Ada.Streams.Stream_IO.Positive_Count;
        csize, usize : File_size_type;
    begin
        Find_offset( this.zipInfo, filename, False, index, csize, usize );
        return True;
    exception
        when File_name_not_found =>
            return False;
    end Exists;

    ----------------------------------------------------------------------------

    overriding
    function Load_Bitmap( this     : access Zip_Archive;
                          filename : String ) return A_Bitmap is
        zippedFile : Zipped_File_Type;
        compSize,
        uncompSize : File_size_type;
        fileData   : A_SEA := null;
        bmp        : A_Bitmap := null;
    begin
        Open( zippedFile, this.zipInfo, filename );
        Get_sizes( this.zipInfo, filename, False, compSize, uncompSize );
        fileData := new Stream_Element_Array(1..Stream_Element_Offset(uncompSize));
        Stream_Element_Array'Read( Stream( zippedFile ), fileData.all );
        Close( zippedFile );

        bmp := Load_From_Memory( Get_Extension( filename ), fileData );
        Delete( fileData );

        return bmp;
    exception
        when others =>
            Delete( fileData );
            if Is_Open( zippedFile ) then
                Close( zippedFile );
            end if;
            raise;
    end Load_Bitmap;

    ----------------------------------------------------------------------------

    overriding
    function Open( this     : access Zip_Archive;
                   filename : String ) return A_Packfile is
        zippedFile : Zipped_File_Type;
        compSize,
        uncompSize : File_size_type;
        fileData   : A_SEA := null;
        file       : A_Packfile;
    begin
        Open( zippedFile, this.zipInfo, filename );
        Get_sizes( this.zipInfo, filename, False, compSize, uncompSize );
        fileData := new Stream_Element_Array(1..Stream_Element_Offset(uncompSize));
        Stream_Element_Array'Read( Stream( zippedFile ), fileData.all );
        Close( zippedFile );

        file := Pack_FOpen( fileData, consume => True );
        fileData := null;

        return file;
    exception
        when others =>
            Delete( fileData );
            if Is_Open( zippedFile ) then
                Close( zippedFile );
            end if;
            raise;
    end Open;

    ----------------------------------------------------------------------------

    overriding
    function Open_Chunk( this     : access Zip_Archive;
                         filename : String ) return access Stream_Element_Array is
        zippedFile : Zipped_File_Type;
        compSize,
        uncompSize : File_size_type;
        fileData   : A_SEA := null;
    begin
        Open( zippedFile, this.zipInfo, filename );
        Get_sizes( this.zipInfo, filename, False, compSize, uncompSize );
        fileData := new Stream_Element_Array(1..Stream_Element_Offset(uncompSize));
        Stream_Element_Array'Read( Stream( zippedFile ), fileData.all );
        Close( zippedFile );

        return fileData;
    exception
        when others =>
            Delete( fileData );
            if Is_Open( zippedFile ) then
                Close( zippedFile );
            end if;
            raise;
    end Open_Chunk;

begin

    Register_Format( "gfx", Load_Zip_Path'Access, Load_Zip_Res'Access );
    Register_Format( "zip", Load_Zip_Path'Access, Load_Zip_Res'Access );

end Archives.Zip_Archives;
