
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings.Hash_Case_Insensitive;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Resources;                         use Resources;

package body Entity_Factory is

    type Entity_Class_Type is
        record
            allocate : Allocator := null;
            instance : Entity_Access := null;
        end record;

    package Class_Registry is new
        Ada.Containers.Indefinite_Hashed_Maps( String, Entity_Class_Type,
                                               Ada.Strings.Hash_Case_Insensitive,
                                               "=", "=" );
    use Class_Registry;

    registry    : Class_Registry.Map;
    name        : Unbounded_String;
    initialized : Boolean := False;

    ----------------------------------------------------------------------------

    function Allocate( id : String ) return Entity_Access is
        crsr : constant Class_Registry.Cursor := registry.Find( id );
    begin
        if Has_Element( crsr ) then
            return Element( crsr ).allocate.all;
        end if;
        pragma Debug( Dbg( "Attempted to create unrecognized " & To_String( name ) &
                           " class id <" & id & ">", D_ENTITIES, ERROR ) );
        return null;
    exception
        when e : RESOURCE_ERROR =>
            pragma Debug( Dbg( "Error creating entity class <" & id & ">: " &
                               Exception_Message( e ), D_ENTITIES, ERROR ) );
            return null;
    end Allocate;

    ----------------------------------------------------------------------------

    procedure Initialize( factoryName : String ) is

        procedure Init_Class( pos : Class_Registry.Cursor ) is

            procedure Examine( id : String; elem : in out Entity_Class_Type ) is
            begin
                elem.instance := elem.allocate.all;
            exception
                when e : others =>
                    null;
                    pragma Debug( Dbg( "Error initializing " & To_String( name ) &
                                       " class <" & id & ">: " &
                                       Exception_Name( e ) & " - " &
                                       Exception_Message( e ), D_INIT, ERROR ) );
            end Examine;

        begin
            registry.Update_Element( pos, Examine'Access );
        end Init_Class;

    begin
        if not initialized then
            name := To_Unbounded_String( factoryName );
            pragma Debug( Dbg( "Initializing " & To_String( name ) &
                               " class factory", D_INIT, INFO ) );
            registry.Iterate( Init_Class'Access );
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is

        procedure Delete_Class( pos : Class_Registry.Cursor ) is

            procedure Examine( id : String; elem : in out Entity_Class_Type ) is
                pragma Unreferenced( id );
            begin
                Delete( elem.instance );
            end Examine;

        begin
            registry.Update_Element( pos, Examine'Access );
        end Delete_Class;

    begin
        if initialized then
            pragma Debug( Dbg( "Finalizing " & To_String( name ) &
                               " class factory", D_INIT, INFO ) );
            registry.Iterate( Delete_Class'Access );
            initialized := False;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    procedure Iterate_Classes( examine : access procedure( id : String ) ) is

        procedure Examine_Class( pos : Class_Registry.Cursor ) is
        begin
            examine.all( Key( pos ) );
        end Examine_Class;

    begin
        registry.Iterate( Examine_Class'Access );
    end Iterate_Classes;

    ----------------------------------------------------------------------------

    procedure Register_Class( id : String; allocate : not null Allocator ) is
    begin
        if not registry.Contains( id ) then
            declare
                elem : Entity_Class_Type;
            begin
                elem.allocate := allocate;
                elem.instance := null;
                registry.Insert( id, elem );
                return;
            end;
        end if;
        pragma Debug( Dbg( "Attempted to register duplicate " &
                           To_String( name ) & " class id <" & id & ">",
                           D_ENTITIES, WARNING ) );
    end Register_Class;

    ----------------------------------------------------------------------------

    function Template( id : String ) return Entity_Access is
        crsr : constant Class_Registry.Cursor := registry.Find( id );
    begin
        if Has_Element( crsr ) then
            return Element( crsr ).instance;
        end if;
        return null;
    end Template;

end Entity_Factory;
