
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Unchecked_Deallocation;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Manager;

package body Events.Corrals is

    function Create_Corral( name : String ) return A_Corral is
        this : constant A_Corral := new Corral;
    begin
        this.Construct( name );
        return this;
    end Create_Corral;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_Listener( this     : access Corral;
                            listener : not null A_Event_Listener;
                            evtName  : String ) is
        list : A_Listener_List;
    begin
        this.listeners.Get_List( evtName, list );
        list.Add( listener );
    end Add_Listener;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Corral; name : String ) is
    begin
        Limited_Object(this.all).Construct;
        this.name := To_Unbounded_String( name );
        this.listeners := new Listener_Registry;
        this.listeners.Init( A_Corral(this) );
        this.queue := new Event_Queue;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Corral ) is
    begin
        this.listeners.Clear;
        this.queue.Shutdown;
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Name( this : access Corral ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Corral ) return String is
    begin
        return "Corral('" & To_String( this.name ) & "')";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    not overriding
    procedure Queue_Event( this : access Corral; evt : in out A_Event ) is
    begin
        pragma Debug( Dbg( "Corral('" & To_String( this.name ) &
                           "'): Queuing " & To_String( evt ), D_EVENTS, INFO ) );
        this.queue.Add( evt );
    end Queue_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove_Listener( this     : access Corral;
                               listener : not null A_Event_Listener;
                               evtName  : String ) is
        list : A_Listener_List;
    begin
        this.listeners.Get_List( evtName, list );
        list.Remove( listener );
    end Remove_Listener;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Corral; upTime, dt : Time_Span ) is
        pragma Unreferenced( upTime, dt );
        broker   : A_Broker;
        evt      : A_Event;
        response : Response_Type;
        list     : A_Listener_List;
    begin
        this.queue.Mark_Frame;
        loop
            loop
                this.queue.Remove( broker );
                exit when broker = null;

                broker.Take_Event( evt );
                pragma Debug( Dbg( "Corral('" & To_String( this.name ) &
                                   "'): Invoking " & To_String( evt ), D_EVENTS, INFO ) );
                this.listeners.Get_List( evt.Get_Name, list );
                list.Call( evt, response );
                broker.Set_Response( response );
                -- broker is deleted by its initiating thread
            end loop;

            this.queue.Remove( evt );
            exit when evt = null;

            pragma Debug( Dbg( "Corral('" & To_String( this.name ) &
                               "'): Dispatching " & To_String( evt ), D_EVENTS, INFO ) );
            this.listeners.Get_List( evt.Get_Name, list );
            list.Call( evt, response );
        end loop;
    end Tick;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Corral ) return String is
    begin
        return "<" & this.Get_Class_Name & ":" & this.Get_Name & ">";
    end To_String;

    ----------------------------------------------------------------------------

    not overriding
    procedure Trigger_Event( this     : access Corral;
                             evt      : in out A_Event;
                             response : out Response_Type ) is
        broker : A_Broker;
    begin
        pragma Debug( Dbg( "Corral '" & To_String( this.name ) &
                           "': Triggering " & To_String( evt ), D_EVENTS, INFO ) );
        Create_Broker( evt, broker );
        this.queue.Add( broker );
        broker.Get_Response( response );
        Delete( broker );
    end Trigger_Event;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Corral ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    protected body Broker_Type is

        procedure Put_Event( evt : in out A_Event ) is
        begin
            if not initialized then
                my_event := evt;
                evt := null;
                initialized := True;
            else
                Delete( evt );
            end if;
        end Put_Event;

        ------------------------------------------------------------------------

        procedure Take_Event( evt : in out A_Event ) is
        begin
            evt := my_event;
            my_event := null;
        end Take_Event;

        ------------------------------------------------------------------------

        entry Get_Response( response : out Response_Type ) when complete is
        begin
            response := my_response;
        end Get_Response;

        ------------------------------------------------------------------------

        procedure Set_Response( response : Response_Type ) is
        begin
            my_response := response;
            complete := True;
        end Set_Response;

    end Broker_Type;

    ----------------------------------------------------------------------------

    procedure Create_Broker( evt : in out A_Event; broker : out A_Broker ) is
    begin
        broker := new Broker_Type;
        broker.Put_Event( evt );
    end Create_Broker;

    ----------------------------------------------------------------------------

    procedure Delete( broker : in out A_Broker ) is
        procedure Free is new Ada.Unchecked_Deallocation( Broker_Type, A_Broker );
        evt : A_Event := null;
    begin
        -- push an event to ensure the broker has been initialized, then take
        -- whatever event may have been originally been stored and discard it.
        broker.Put_Event( evt );
        broker.Take_Event( evt );
        Delete( evt );
        Free( broker );
    end Delete;

    --==========================================================================

    protected body Event_Queue is

        procedure Add( broker : A_Broker ) is
        begin
            if not stopped then
                brokers.Append( broker );
            else
                broker.Set_Response( No_Response );
            end if;
        end Add;

        ------------------------------------------------------------------------

        procedure Add( evt : in out A_Event ) is
        begin
            if not stopped then
                queue.Append( evt );
                evt := null;
            else
                Delete( evt );
            end if;
        end Add;

        ------------------------------------------------------------------------

        procedure Mark_Frame is
        begin
            remaining := Integer(Length( queue ));
        end Mark_Frame;

        ------------------------------------------------------------------------

        procedure Remove( broker : out A_Broker ) is
        begin
            broker := null;
            if not brokers.Is_Empty then
                broker := brokers.First_Element;
                brokers.Delete_First;
            end if;
        end Remove;

        ------------------------------------------------------------------------

        procedure Remove( evt : out A_Event ) is
        begin
            evt := null;
            if remaining > 0 then
                evt := queue.First_Element;
                queue.Delete_First;
                remaining := remaining - 1;
            end if;
        end Remove;

        ------------------------------------------------------------------------

        procedure Shutdown is
            broker : A_Broker;
            evt    : A_Event;
        begin
            while not brokers.Is_Empty loop
                broker := brokers.First_Element;
                brokers.Delete_First;
                broker.Set_Response( No_Response );
                -- brokers are be deleted by their initiating thread after
                -- receiving a response
            end loop;
            while not queue.Is_Empty loop
                evt := queue.First_Element;
                queue.Delete_First;
                Delete( evt );
            end loop;
            remaining := 0;
            stopped := True;
        end Shutdown;

    end Event_Queue;

    --==========================================================================

    not overriding
    procedure Add( this : access Listener_List; listener : not null A_Event_Listener ) is
        inserted : Boolean;
    begin
        -- verify the listener hasn't already been added to this list
        this.listeners.Append_No_Duplicate( listener, inserted );
        if inserted then
            if this.listeners.Length = 1 then
                -- the first time a listener for a new event type is
                -- added, register the corral to receive events of that
                -- type from the event manager.
                Events.Manager.Register_Listener( this.corral, To_String( this.name ) );
            end if;
        else
            null;
            pragma Debug( Dbg( "Listener <" & listener.To_String &
                               "> already added for event <" & To_String( this.name ) & '>' &
                               " in corral <" & To_String( this.corral.name ) & '>',
                               D_EVENTS, WARNING ) );
        end if;
    end Add;

    ----------------------------------------------------------------------------

    not overriding
    procedure Call( this : access Listener_List; evt : in out A_Event; response : out Response_Type ) is

          procedure Call( listener : A_Event_Listener; quit : in out Boolean ) is
          begin
              listener.Handle_Event( evt, response );
              quit := evt = null;
          exception
              when e : others =>
                  Delete( evt );
                  Set_Status( response, ST_EXCEPTION );
                  Set_Message( response,
                               "Unhandled exception: " &
                               Exception_Name( e ) & " - " &
                               Exception_Message( e ) );
                  quit := True;
          end Call;

    begin
        response := No_Response;
        this.listeners.Iterate_With_Quit( Call'Access );
        Delete( evt );
    end Call;

    ----------------------------------------------------------------------------

    not overriding
    procedure Clear( this : access Listener_List ) is
    begin
        if not this.listeners.Is_Empty then
            Events.Manager.Unregister_Listener( this.corral, To_String( this.name ) );
        end if;
        this.listeners.Clear;
    end Clear;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Listener_List; owner : not null A_Corral; evtName : String ) is
    begin
        this.corral := owner;
        this.name := To_Unbounded_String( evtName );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove( this : access Listener_List; listener : not null A_Event_Listener ) is
        crsr : Listener_Lists.Cursor;
    begin
        this.listeners.Find( listener, crsr );
        if Listener_Lists.Has_Element( crsr ) then
            this.listeners.Remove( crsr );
            if this.listeners.Is_Empty then
                -- unregister the corral with the event manager for
                -- events of this type because the corral has no
                -- more listeners
                Events.Manager.Unregister_Listener( this.corral, To_String( this.name ) );
            end if;
        else
            null;
            pragma Debug( Dbg( "Unable to remove listener " &
                               listener.To_String & " for event <" &
                               To_String( this.name ) & "> in corral <" &
                               To_String( this.corral.name ) &
                               ">: Listener not found", D_EVENTS, WARNING ) );
        end if;
    end Remove;

    --==========================================================================

    protected body Listener_Registry is

        procedure Clear is

            procedure Clear_List( pos : Type_Map.Cursor ) is
            begin
                Element( pos ).Clear;
            end Clear_List;

        begin
            registry.Iterate( Clear_List'Access );
        end Clear;

        ------------------------------------------------------------------------

        procedure Get_List( evtName : String; list : out A_Listener_List ) is
            crsr : constant Type_Map.Cursor := registry.Find( evtName );
        begin
            if Type_Map.Has_Element( crsr ) then
                list := Type_Map.Element( crsr );
            else
                list := new Listener_List;
                list.Construct( corral, evtName );
                registry.Insert( evtName, list );
            end if;
        end Get_List;

        ------------------------------------------------------------------------

        procedure Init( owner : not null A_Corral ) is
        begin
            corral := owner;
        end Init;

    end Listener_Registry;

end Events.Corrals;
