
with Events.Corrals;                    use Events.Corrals;
with Events.Listeners;                  use Events.Listeners;
with Objects;                           use Objects;
with Processes;                         use Processes;
with Widgets;                           use Widgets;
with Widgets.Containers.Windows;        use Widgets.Containers.Windows;

private with Ada.Containers.Indefinite_Hashed_Maps;
private with Ada.Strings.Hash_Case_Insensitive;
private with Ada.Real_Time;
private with Audio_Players;
private with Events;
private with Input_Handlers;
private with Processes.Managers;
private with Renderers;

package Game_Views is

    type Game_View is abstract new Limited_Object and Event_Listener and Process with private;
    type A_Game_View is access all Game_View'Class;
    pragma No_Strict_Aliasing( A_Game_View );

    function Create_Game_View( xres, yres : Positive;
                               scale      : Positive ) return A_Game_View;
    pragma Postcondition( Create_Game_View'Result /= null );

    -- Attachs a process to the view's process manager.
    procedure Attach( this : access Game_View; process : not null A_Process );

    -- Detaches a process from the view's process manager.
    procedure Detach( this : access Game_View; process : not null A_Process );

    -- Returns the view's corral. The view creates its corral at construction so
    -- this will never return null.
    function Get_Corral( this : not null access Game_View'Class ) return A_Corral;
    pragma Postcondition( Get_Corral'Result /= null );

    -- Returns a widget in the registry by id. Raises exception ID_NOT_FOUND if
    -- the widget does not exist.
    function Get_Widget( this : access Game_View; id : String ) return A_Widget;
    pragma Postcondition( Get_Widget'Result /= null );

    -- Returns the view's Window widget. This may return null if the window has
    -- not yet been set.
    function Get_Window( this : access Game_View ) return A_Window;

    -- Adds a widget to the view's widget registry. Raises exception
    -- DUPLICATE_ID if a widget with the same id already exists.
    procedure Register( this : access Game_View; widget : not null A_Widget );

    -- Sets the view's Window widget. If window is null, the view's window will
    -- be removed. The view's Renderer is deleted and recreated when the window
    -- is changed.
    procedure Set_Window( this : access Game_View; window : in out A_Window );
    pragma Postcondition( window = null );

    -- Starts the game view, attaching it to the framework. This will attach
    -- event listeners, processes, start the view's subsystems (audio, etc)
    -- and begin running the view's process manager.
    procedure Start( this : access Game_View );

    -- Stops the game view, detaching it from the framework. This must be called
    -- after Start and before deleting the object.
    procedure Stop( this : access Game_View );

    -- Removes a widget from the view's registry. If the widget was not
    -- previously registered, nothing happens.
    procedure Unregister( this : access Game_View; id : String );
    pragma Precondition( id'Length > 0 );

    procedure Delete( this : in out A_Game_View );
    pragma Postcondition( this = null );

    DUPLICATE_ID,
    ID_NOT_FOUND : exception;

private

    use Ada.Real_Time;
    use Audio_Players;
    use Events;
    use Input_Handlers;
    use Processes.Managers;
    use Renderers;

    package Widget_Registry is new Ada.Containers.Indefinite_Hashed_Maps(
        String, A_Widget, Ada.Strings.Hash_Case_Insensitive, "=", "=" );
    use Widget_Registry;

    -----------------------------------------------------------------------------

    type Game_View is abstract new Limited_Object and Event_Listener and Process with
        record
            started,
            stopped     : Boolean := False;
            paused      : Boolean := False;
            win         : A_Window := null;
            renderer    : A_Renderer := null;
            audioPlayer : A_Audio_Player := null;
            inhandler   : A_Input_Handler := null;
            pman        : A_Process_Manager := null;
            corral      : A_Corral := null;
            widgets     : Widget_Registry.Map;
        end record;

    procedure Construct( this : access Game_View );

    procedure Delete( this : in out Game_View );

    function Get_Process_Name( this : access Game_View ) return String;
    pragma Postcondition( Get_Process_Name'Result'Length > 0 );

    -- Called when the window is requested to close by the OS. The default
    -- implementation is to simply queue a Queue_Close_Window message which will
    -- exit the application. If a confirmation dialog or some other activity
    -- needs to happen first, then this procedure should be overridden.
    procedure Handle_Close_Request( this : access Game_View );

    procedure Handle_Event( this : access Game_View;
                            evt  : in out A_Event;
                            resp : out Response_Type );
    pragma Precondition( evt /= null );

    -- Called when resource loading is beginning and ending, so a loading
    -- message can be displayed.
    procedure Handle_Loading( this : access Game_View; loading : Boolean );

    -- Called when the game logic is paused or resumed. The base implementation
    -- does nothing. Override this procedure to perform actions when the game
    -- pause state changes.
    procedure Handle_Paused( this : access Game_View; paused : Boolean );

    procedure Tick( this : access Game_View; upTime, dt : Time_Span );

    ----------------------------------------------------------------------------

    type Allocator is
        access function( xres, yres : Positive;
                         scale      : Positive ) return A_Game_View;

    procedure Register_Allocator( allocate : not null Allocator );

end Game_Views;
