
with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Game;                       use Events.Game;
with Events.World;                      use Events.World;
with Preferences;                       use Preferences;
with Statuses;                          use Statuses;
with Support;                           use Support;
with Values;                            use Values;

package body Games is

    use View_Lists;

    allocate : Allocator := null;

    ----------------------------------------------------------------------------

    function Create_Game return A_Game is
    begin
        if allocate /= null then
            return allocate.all;
        else
            raise Constraint_Error with "Game allocator not registered";
        end if;
    end Create_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add_View( this : access Game; view : in out A_Game_View ) is
    begin
        this.views.Append( view );
        view := null;
    end Add_View;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Game ) is
    begin
        Limited_Object(this.all).Construct;
        this.corral := Create_Corral( "game logic" );
        this.physics := Create_Physics( this.corral );
        this.pman := Create_Process_Manager( "game",
                                             Constrain( Get_Pref( "application",
                                                                  "max_logic_fps" ),
                                                        1, 100 ), 15 );
        this.sessionVars := Create_Association;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Game ) is
    begin
        Delete( this.world );

        declare
            procedure Examine( pos : View_Lists.Cursor ) is
                view : A_Game_View := Element( pos );
            begin
                Delete( view );
            end Examine;
        begin
            this.views.Iterate( Examine'Access );
            this.views.Clear;
        end;

        Delete( this.sessionVars );
        Delete( this.pman );
        Delete( this.physics );
        Delete( this.corral );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure End_Game( this : access Game ) is
    begin
        if this.playing then
            pragma Debug( Dbg( "Game: Game session ended", D_GAME, INFO ) );
            this.playing := False;

            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );
            this.Set_World( null );

            Queue_Game_State( playing => False );
        else
            null;
            pragma Debug( Dbg( "Warning: Can't end game; no game in progress.",
                               D_GAME, WARNING ) );
        end if;
    end End_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Game_Var_Add( this : access Game; var : String; val : Integer ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, this.sessionVars.As_Integer( var ) + val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Game_Var_Add;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Corral( this : access Game ) return A_Corral is
    begin
        return this.corral;
    end Get_Corral;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Game ) return String is
        pragma Unreferenced( this );
    begin
        return "Game Logic";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Game;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is

        ------------------------------------------------------------------------

        procedure Do_Load_World( evt : not null A_Load_World_Event ) is
        begin
            this.Load_World( evt.Get_Filename );
            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );
            Set_Status( resp, ST_SUCCESS );
        exception
            when e : others =>
                Set_Status( resp, ST_FAILED );
                Set_Message( resp, Exception_Message( e ) );
        end Do_Load_World;

        ------------------------------------------------------------------------

        procedure Do_Pause( evt : not null A_Pause_Game_Event ) is
        begin
            if evt.Is_Paused = not this.paused then
                A_Game(this).Pause( not this.paused );
            end if;
        end Do_Pause;

        ------------------------------------------------------------------------

        procedure Do_View_Ready( evt : not null A_View_Ready_Event ) is
        begin
            if evt.Is_Ready then
                this.pman.Attach( A_Process(this) );
                this.pman.Attach( A_Process(this.physics) );
            end if;
        end Do_View_Ready;

        ------------------------------------------------------------------------

    begin
        if evt.Get_Id = END_GAME_ID then
            A_Game(this).End_Game;
        elsif evt.Get_Id = LOAD_WORLD_ID then
            Do_Load_World( A_Load_World_Event(evt) );
        elsif evt.Get_Id = NEW_GAME_ID then
            A_Game(this).New_Game;
        elsif evt.Get_Id = PAUSE_GAME_ID then
            Do_Pause( A_Pause_Game_Event(evt) );
        elsif evt.Get_Id = VIEW_READY_ID then
            Do_View_Ready( A_View_Ready_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Load_World( this : access Game; name : String ) is
    begin
        if not this.views.Is_Empty then
            pragma Debug( Dbg( "Game.Load_World: Loading <" & name & ">",
                               D_GAME, INFO ) );
            this.Set_World( Load_World( name ) );
        else
            null;
            pragma Debug( Dbg( "Game.Load_World Error: Game has no registered views",
                               D_GAME, ERROR ) );
        end if;
    end Load_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure New_Game( this : access Game ) is
    begin
        if not this.playing then
            pragma Debug( Dbg( "Game.New_Game: Game session begun", D_GAME, INFO ) );
            Queue_Game_State( playing => True );
            this.paused := False;
            this.playing := True;
        else
            null;
            pragma Debug( Dbg( "Game.New_Game: Can't begin game; " &
                               "game already in progress", D_GAME, ERROR ) );
        end if;
    end New_Game;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pause( this : access Game; enabled : Boolean ) is
    begin
        if this.playing and then enabled /= this.paused then
            this.paused := enabled;
            this.pman.Pause( A_Process(this), this.paused );
            this.pman.Pause( A_Process(this.physics), this.paused );
            this.pman.Pause( A_Process(this.world), this.paused );
            Queue_Game_Paused( this.paused );
        end if;
    end Pause;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Game_Var( this : access Game; var : String; val : Integer ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Game_Var( this : access Game; var : String; val : Boolean ) is
        value : A_Value;
    begin
        this.sessionVars.Set_Value( var, val );
        value := Create_Value( this.sessionVars.As_Integer( var ) );
        Queue_Game_Var_Changed( var, value );
    end Set_Game_Var;

    ----------------------------------------------------------------------------

    procedure Set_World( this : not null access Game'Class; world : A_World ) is
    begin
        if this.world /= null then
            this.world.Detach_From_Framework;
            -- Do not delete this.world here because it is still owned by the
            -- process manager. The world will be destroyed by a destructor
            -- given to the process manager by the world object.
        end if;
        this.world := world;
        if this.world /= null then
            this.world.Attach_To_Framework( this.corral, this.pman );
        end if;
    end Set_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Start( this : access Game ) is

        procedure Examine( pos : View_Lists.Cursor ) is
        begin
            Element( pos ).Start;
        end Examine;

    begin
        if not this.started and then not this.stopped then
            this.started := True;
            this.corral.Add_Listener( A_Event_Listener(this), "End_Game" );
            this.corral.Add_Listener( A_Event_Listener(this), "Load_World" );
            this.corral.Add_Listener( A_Event_Listener(this), "New_Game" );
            this.corral.Add_Listener( A_Event_Listener(this), "Pause_Game" );
            this.corral.Add_Listener( A_Event_Listener(this), "View_Ready" );
            this.pman.Attach( A_Process(this.corral) );

            this.views.Iterate( Examine'Access );
            this.pman.Start;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    not overriding
    procedure Stop( this : access Game ) is

        procedure Examine( pos : View_Lists.Cursor ) is
        begin
            Element( pos ).Stop;
        end Examine;

    begin
        if not this.stopped then
            this.stopped := True;
            this.pman.Stop;
            this.views.Iterate( Examine'Access );

            if this.world /= null then
                this.world.Detach_From_Framework;
            end if;
            this.pman.Detach( A_Process(this.physics) );
            this.pman.Detach( A_Process(this) );
            this.pman.Detach( A_Process(this.corral) );
            this.corral.Remove_Listener( A_Event_Listener(this), "View_Ready" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Pause_Game" );
            this.corral.Remove_Listener( A_Event_Listener(this), "New_Game" );
            this.corral.Remove_Listener( A_Event_Listener(this), "Load_World" );
            this.corral.Remove_Listener( A_Event_Listener(this), "End_Game" );
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Game; upTime, dt : Time_Span ) is
        pragma Unreferenced( this, upTime, dt );
    begin
        null;
    end Tick;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Game ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Register_Allocator( allocate : not null Allocator ) is
    begin
        Games.allocate := allocate;
    end Register_Allocator;

end Games;
