
with Events.Corrals;                    use Events.Corrals;
with Events.Listeners;                  use Events.Listeners;
with Game_Views;                        use Game_Views;
with Objects;                           use Objects;
with Processes;                         use Processes;

private with Ada.Containers.Indefinite_Doubly_Linked_Lists;
private with Ada.Real_Time;
private with Associations;
private with Events;
private with Physics.Managers;
private with Processes.Managers;
private with Worlds;

package Games is

    type Game is abstract new Limited_Object and Event_Listener and Process with private;
    type A_Game is access all Game'Class;

    -- Creates a new Game object, using the registered allocator.
    function Create_Game return A_Game;
    pragma Postcondition( Create_Game'Result /= null );

    -- Adds a view to the game logic. 'view' is consumed and then owned by the
    -- Game object.
    procedure Add_View( this : access Game; view : in out A_Game_View );
    pragma Precondition( view /= null );
    pragma Postcondition( view = null );

    -- Returns the Game's event corral. It is created during Game construction.
    function Get_Corral( this : access Game ) return A_Corral;
    pragma Postcondition( Get_Corral'Result /= null );

    -- Raises exception on error, clearing the current world.
    procedure Load_World( this : access Game; name : String );
    pragma Postcondition( name'Length > 0 );

    -- Starts the game logic and attaches it to the framework.
    procedure Start( this : access Game );

    -- Stops the game logic and detaches it from the framework.
    procedure Stop( this : access Game );

    -- Deletes the Game.
    procedure Delete( this : in out A_Game );
    pragma Postcondition( this = null );

private

    use Ada.Real_Time;
    use Associations;
    use Events;
    use Physics.Managers;
    use Processes.Managers;
    use Worlds;

    package View_Lists is new Ada.Containers.Indefinite_Doubly_Linked_Lists( A_Game_View, "=" );

    ----------------------------------------------------------------------------

    type Game is abstract new Limited_Object and Event_Listener and Process with
        record
            started,
            stopped     : Boolean := False;
            views       : View_Lists.List;
            corral      : A_Corral := null;
            physics     : A_Physics := null;
            pman        : A_Process_Manager := null;
            sessionVars : A_Association := null;
            world       : A_World := null;
            playing     : Boolean := False;
            paused      : Boolean := False;
        end record;

    procedure Construct( this : access Game );

    procedure Delete( this : in out Game );

    -- Ends the current game session. Game logic is stopped and the Game_State
    -- event is sent.
    procedure End_Game( this : access Game );

    -- Adds 'val' to game session variable 'var'. A Game_Var_Changed event is
    -- queued. An exception is raised on error.
    procedure Game_Var_Add( this : access Game; var : String; val : Integer );
    pragma Precondition( var'Length > 0 );

    function Get_Process_Name( this : access Game ) return String;
    pragma Postcondition( Get_Process_Name'Result'Length > 0 );

    procedure Handle_Event( this : access Game;
                            evt  : in out A_Event;
                            resp : out Response_Type );
    pragma Precondition( evt /= null );

    -- Begins a new game session from the start. Override this procedure to
    -- setup initial game state and load the first game world. Call this
    -- procedure from the overriding one first.
    procedure New_Game( this : access Game );

    -- Pauses the game logic. If overriding this procedure, call it explicitly
    -- first in the overriding Pause procedure. It will set the paused flag in
    -- the game object to the value of 'enabled'.
    procedure Pause( this : access Game; enabled : Boolean );

    -- Sets game session variable 'var' to 'val'. A Game_Var_Changed event is
    -- queued.
    procedure Set_Game_Var( this : access Game; var : String; val : Integer );
    pragma Precondition( var'Length > 0 );

    -- Sets game session variable 'var' to 'val'. A Game_Var_Changed event is
    -- queued.
    procedure Set_Game_Var( this : access Game; var : String; val : Boolean );
    pragma Precondition( var'Length > 0 );

    -- Sets the game's current world and attaches the world to the framework.
    -- The previous world will be detached from the framework and deleted.
    procedure Set_World( this : not null access Game'Class; world : A_World );

    procedure Tick( this : access Game; upTime, dt : Time_Span );

    ----------------------------------------------------------------------------

    type Allocator is access function return A_Game;

    procedure Register_Allocator( allocate : not null Allocator );

end Games;
