
with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Unchecked_Deallocation;
with Allegro.Keyboard;                  use Allegro.Keyboard;
with Allegro.Mouse;                     use Allegro.Mouse;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Input;                      use Events.Input;
with Interfaces;                        use Interfaces;
with Keyboard;                          use Keyboard;
with Mouse;                             use Mouse;

package body Input_Handlers is

    UNKNOWN : constant := 16#7FFFFFFF#;

    ----------------------------------------------------------------------------

    type Click_Count is mod 2;

    type Key_Array is array (1..KEY_MAX) of Boolean;

    type Mouse_Button is (LEFT, RIGHT, MIDDLE);

    type MB_State is
        record
            pressTime : Time := Clock;
            pressX,
            pressY    : Integer := UNKNOWN;
            lastHeld  : Time := Clock;
            clickTime : Time := Clock;
            clickX,
            clickY    : Integer := UNKNOWN;
            clicks    : Click_Count := 0;
        end record;

    type MB_States is array (Mouse_Button) of MB_State;

    ----------------------------------------------------------------------------

    CLICK_DRIFT       : constant := 10;                              -- allowed mouse drift during click
    CLICK_DELAY       : constant Time_Span := Milliseconds( 400 );   -- delay between press and release
    DOUBLECLICK_DELAY : constant Time_Span := Milliseconds( 600 );   -- delay between 1st click and 2nd click
    HELD_REPEAT_DELAY : constant Time_Span := Milliseconds( 1000 / 25 ); -- delay between mouse held events

    ----------------------------------------------------------------------------

    function Create_Input_Handler( hertz : Positive ) return A_Input_Handler is
        this : constant A_Input_Handler := new Input_Handler;
    begin
        this.Construct( hertz );
        return this;
    end Create_Input_Handler;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Input_Handler; hertz : Positive ) is
    begin
        Limited_Object(this.all).Construct;
        this.tickDelta := Milliseconds( 1000 / hertz );
        this.process := new Input_Task;
        this.process.Init( A_Input_Handler(this) );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Input_Handler ) is
    begin
        this.Stop;
        Delete( this.process );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Start( this : not null access Input_Handler'Class ) is
    begin
        if not this.started and then not this.stopped then
            this.started := True;
            this.process.Start;
        end if;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Input_Handler'Class ) is
    begin
        if not this.stopped then
            this.stopped := True;
            this.process.Stop;

            for count in 1..10 loop
                if Is_Terminated( this.process'Identity ) then
                    exit;
                end if;
                delay To_Duration( this.tickDelta );
            end loop;
            pragma Debug( not Is_Terminated( this.process'Identity ),
                          Dbg( "Input_Handler.Stop: Input_Task failed to stop",
                               D_INIT, ERROR ) );
        end if;
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Input_Handler.Stop Exception: ...", D_INIT, ERROR ) );
            pragma Debug( Dbg( e, D_INIT, ERROR ) );
    end Stop;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Input_Handler ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    task body Input_Task is
        -- range of key scan codes to poll
        FIRST_POLL_KEY : constant := 1;
        LAST_POLL_KEY  : constant := 91;

        mb             : Unsigned_32 := UNKNOWN;
        mx, my, mz     : Integer := UNKNOWN;
        mbtn           : MB_States;
        keys           : Key_Array := (others => False);
        prevkeys       : Key_Array := (others => False);

        ------------------------------------------------------------------------

        procedure Tick_Keyboard( modifiers : out Modifiers_Array ) is
            ok       : constant Integer := Poll_Keyboard;
            capslock : constant Boolean := (Get_Key_Shifts and KB_CAPSLOCK_FLAG) > 0;
        begin
            if ok /= 0 then
                -- keyboard driver not installed
                return;
            end if;

            -- Handle modifier key events separately first
            for k in 115..120 loop
                prevkeys(k) := keys(k);
                keys(k) := Get_Key_Status( k ) /= 0;
                if keys(k) then
                    if not prevkeys(k) then
                        -- key press
                        Queue_Key_Press( k, MODIFIERS_NONE, False );
                    end if;
                    -- key held
                    Queue_Key_Held( k, MODIFIERS_NONE, False );
                elsif not keys(k) and then prevkeys(k) then
                    -- key release
                    Queue_Key_Release( k, MODIFIERS_NONE, False );
                end if;
            end loop;

            -- Set the modifiers for any key events
            modifiers(ALT) := keys( KEY_ALT ) or else keys( KEY_ALTGR );
            modifiers(CTRL) := keys( KEY_LCONTROL ) or else keys( KEY_RCONTROL );
            modifiers(SHIFT) := keys( KEY_LSHIFT ) or else keys( KEY_RSHIFT );

            -- Poll all remaining keys that we care about, with modifiers
            for k in FIRST_POLL_KEY..LAST_POLL_KEY loop
                prevkeys(k) := keys(k);
                keys(k) := Get_Key_Status( k ) /= 0;
                if keys(k) then
                    if not prevkeys(k) then
                        -- key press
                        Queue_Key_Press( k, modifiers, capslock );
                    end if;
                    -- key held
                    Queue_Key_Held( k, modifiers, capslock );
                elsif not keys(k) and then prevkeys(k) then
                    -- key release
                    Queue_Key_Release( k, modifiers, capslock );
                end if;
            end loop;
        end Tick_Keyboard;

        ------------------------------------------------------------------------

        procedure Tick_Mouse( modifiers : Modifiers_Array ) is
            nmx : constant Integer := Mouse_X;
            nmy : constant Integer := Mouse_Y;
            nmz : constant Integer := Mouse_Z;
            nmb : constant Unsigned_32 := Unsigned_32(Mouse_B);
            clk : constant Time := Clock;
        begin
            if mx = UNKNOWN or else my = UNKNOWN then
                -- first time around we get the mouse's location
                -- it is not a movement event
                null;
            elsif nmx /= mx or else nmy /= my then
                -- movement event
                Queue_Mouse_Move( nmx, nmy );
            end if;

            if mx /= UNKNOWN and then nmz /= mz then
                Queue_Mouse_Scroll( nmx, nmy, mz - nmz );
            end if;

            if mb /= UNKNOWN then
                if (nmb and MOUSE_BUTTON_LEFT) > 0 then
                    -- left mouse is down
                    if (mb and MOUSE_BUTTON_LEFT) = 0 then
                        -- event: left mouse press
                        Queue_Mouse_Press( nmx, nmy, Mouse_Left, modifiers );
                        mbtn(LEFT).pressTime := clk;
                        mbtn(LEFT).pressX := nmx;
                        mbtn(LEFT).pressY := nmy;
                        mbtn(LEFT).lastHeld := clk;
                    elsif clk - mbtn(LEFT).pressTime >= CLICK_DELAY and then
                          clk - mbtn(LEFT).lastHeld >= HELD_REPEAT_DELAY
                    then
                        -- event: left mouse still held
                        Queue_Mouse_Held( nmx, nmy, Mouse_Left );
                        mbtn(LEFT).lastHeld := clk;
                    end if;
                else
                    -- left mouse is up
                    if (mb and MOUSE_BUTTON_LEFT) > 0 then
                        -- event: left mouse release
                        Queue_Mouse_Release( nmx, nmy, Mouse_Left );
                        if clk - mbtn(LEFT).pressTime < CLICK_DELAY and then
                           abs( mbtn(LEFT).pressX - nmx ) < CLICK_DRIFT and then
                           abs( mbtn(LEFT).pressY - nmy ) < CLICK_DRIFT
                        then
                            -- event: left mouse click
                            Queue_Mouse_Click( nmx, nmy, Mouse_Left );
                            if mbtn(LEFT).clicks = 1 and then
                               clk - mbtn(LEFT).clickTime < DOUBLECLICK_DELAY and then
                               abs( mbtn(LEFT).clickX - nmx ) < CLICK_DRIFT and then
                               abs( mbtn(LEFT).clickY - nmy ) < CLICK_DRIFT
                            then
                                -- event: left mouse double click
                                Queue_Mouse_Doubleclick( nmx, nmy, Mouse_Left );
                            end if;
                            mbtn(LEFT).clicks := mbtn(LEFT).clicks + 1;
                            mbtn(LEFT).clickTime := clk;
                            mbtn(LEFT).clickX := nmx;
                            mbtn(LEFT).clickY := nmy;
                        end if;
                    end if;
                end if;

                if (nmb and MOUSE_BUTTON_RIGHT) > 0 then
                    -- right mouse is down
                    if (mb and MOUSE_BUTTON_RIGHT) = 0 then
                        -- event: right mouse press
                        Queue_Mouse_Press( nmx, nmy, Mouse_Right, modifiers );
                        mbtn(RIGHT).pressTime := clk;
                        mbtn(RIGHT).pressX := nmx;
                        mbtn(RIGHT).pressY := nmy;
                    elsif clk - mbtn(RIGHT).pressTime >= CLICK_DELAY and then
                          clk - mbtn(RIGHT).lastHeld >= HELD_REPEAT_DELAY
                    then
                        -- event: right mouse still held
                        Queue_Mouse_Held( nmx, nmy, Mouse_Right );
                        mbtn(RIGHT).lastHeld := clk;
                    end if;
                else
                    -- right mouse is up
                    if (mb and MOUSE_BUTTON_RIGHT) > 0 then
                        -- event: right mouse release
                        Queue_Mouse_Release( nmx, nmy, Mouse_Right );
                        if clk - mbtn(RIGHT).pressTime < CLICK_DELAY and then
                           abs( mbtn(RIGHT).pressX - nmx ) < CLICK_DRIFT and then
                           abs( mbtn(RIGHT).pressY - nmy ) < CLICK_DRIFT
                        then
                            -- event: right mouse click
                            Queue_Mouse_Click( nmx, nmy, Mouse_Right );
                        end if;
                    end if;
                end if;

                if (nmb and MOUSE_BUTTON_MIDDLE) > 0 then
                    -- middle mouse is down
                    if (mb and MOUSE_BUTTON_MIDDLE) = 0 then
                        -- event: middle mouse press
                        Queue_Mouse_Press( nmx, nmy, Mouse_Middle, modifiers );
                        mbtn(MIDDLE).pressTime := clk;
                        mbtn(MIDDLE).pressX := nmx;
                        mbtn(MIDDLE).pressY := nmy;
                    elsif clk - mbtn(MIDDLE).pressTime >= CLICK_DELAY and then
                          clk - mbtn(MIDDLE).lastHeld >= HELD_REPEAT_DELAY
                    then
                        -- event: middle mouse still held
                        Queue_Mouse_Held( nmx, nmy, Mouse_Middle );
                        mbtn(MIDDLE).lastHeld := clk;
                    end if;
                else
                    -- middle mouse is up
                    if (mb and MOUSE_BUTTON_MIDDLE) > 0 then
                        -- event: middle mouse release
                        Queue_Mouse_Release( nmx, nmy, Mouse_Middle );
                        if clk - mbtn(MIDDLE).pressTime < CLICK_DELAY and then
                           abs( mbtn(MIDDLE).pressX - nmx ) < CLICK_DRIFT and then
                           abs( mbtn(MIDDLE).pressY - nmy ) < CLICK_DRIFT
                        then
                            -- event: middle mouse click
                            Queue_Mouse_Click( nmx, nmy, Mouse_Middle );
                        end if;
                    end if;
                end if;
            end if;

            mb := nmb;
            mx := nmx;
            my := nmy;
            mz := nmz;
        end Tick_Mouse;

        ------------------------------------------------------------------------

        handler   : A_Input_Handler;
        done      : Boolean := False;
        now,
        nextTick  : Time;
        modifiers : Modifiers_Array;
    begin
        pragma Debug( Dbg( "Input_Handlers.Input_Task created", D_INIT, INFO ) );

        accept Init( in_handler : A_Input_Handler ) do
            handler := in_handler;
        end Init;

        select
            accept Start do
                null;
                pragma Debug( Dbg( "Input_Handlers.Input_Task starting", D_INIT, INFO ) );
            end Start;
        or
            accept Stop do
                pragma Debug( Dbg( "Input_Handlers.Input_Task stopping", D_INIT, INFO ) );
                done := True;
            end Stop;
        end select;

        while not done loop
            select
                accept Stop do
                    pragma Debug( Dbg( "Input_Handlers.Input_Task stopping", D_INIT, INFO ) );
                    done := True;
                end Stop;
            else
                nextTick := Clock + handler.tickDelta;

                Tick_Keyboard( modifiers );
                Tick_Mouse( modifiers );

                now := Clock;
                if now < nextTick then
                    delay To_Duration( nextTick - now );
                end if;
            end select;
        end loop;

        pragma Debug( Dbg( "Input_Handlers.Input_Task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            null;
            pragma Debug( Dbg( "Input_Handlers.Input_Task Exception: ...", D_INPUT, ERROR ) );
            pragma Debug( Dbg( e, D_INPUT, ERROR ) );
    end Input_Task;

    ----------------------------------------------------------------------------

    procedure Delete( intask : in out A_Input_Task ) is
        procedure Free is new Ada.Unchecked_Deallocation( Input_Task, A_Input_Task );
    begin
        Free( intask );
    end Delete;

end Input_Handlers;
