
with Debugging;                         use Debugging;
with Directions;                        use Directions;
with Engine_Debugging;                  use Engine_Debugging;
with Events.Entities;                   use Events.Entities;
with Physics.Managers;                  use Physics.Managers;
with Support;                           use Support;

package body Physics.Bodies is

    -- At framerate 40 Hz:
    --
    -- friction = 0.8      vx := vx * friction;
    -- gravity = 0.9       vy := vy + gravity;  36 pixels / sec^2

    -- player max vx = 3.5 pixels * 40 fps
    -- player max vy = 8.0 pixels * 40 fps

    -- Test for map collision before adjusting velocities. Test a side of the
    -- entity as a whole. Test all blocks on the right side from top to bottom
    -- of entity and handle appropriately. Only a few tile types apply to the
    -- sides and only a few tile types apply to the feet and head.

    -- slope - feet, head
    -- one way floor - feet and falling
    -- gem socket - feet
    -- wall - any
    -- hazzard - any
    -- end level - any
    -- scuba gear - any

    -- We really only care about walls, slopes and one-way floors for clipping.

    -- The other tile types should be sent as events.
    -- Map_Collision( entity id, map x, map y )

    ----------------------------------------------------------------------------

    function Create_Corpus( id       : Entity_Id;
                            width,
                            height   : Natural;
                            physical : Boolean;
                            clipped  : Boolean;
                            manager  : access Physics.Managers.Physics_Manager'Class
                          ) return A_Corpus is
        this : constant A_Corpus := new Corpus;
    begin
        this.Construct( id, width, height, physical, clipped, manager );
        return this;
    end Create_Corpus;

    ----------------------------------------------------------------------------

    procedure Apply_Forces( this : not null access Corpus'Class; dt : Float )  is
    begin
        -- apply gravitational acceleration
        if this.physical then
            if not this.standing then
                this.vy := this.vy + (dt * this.manager.Gravity);
            end if;
        end if;

        -- approach target velocity
        if this.ax > 0.0 then
            if this.tvx > this.vx then
                if this.ax >= INSTANT_ACCELERATION then
                    this.vx := this.tvx;
                else
                    this.vx := this.vx + (dt * this.ax);
                    if this.vx > this.tvx then
                        this.vx := this.tvx;
                    end if;
                end if;
            elsif this.tvx < this.vx then
                if this.ax >= INSTANT_ACCELERATION then
                    this.vx := this.tvx;
                else
                    this.vx := this.vx - (dt * this.ax);
                    if this.vx < this.tvx then
                        this.vx := this.tvx;
                    end if;
                end if;
            end if;
        end if;
        if this.ay > 0.0 then
            if this.tvy > this.vy then
                if this.ay >= INSTANT_ACCELERATION then
                    this.vy := this.tvy;
                else
                    this.vy := this.vy + (dt * this.ay);
                    if this.vy > this.tvy then
                        this.vy := this.tvy;
                    end if;
                end if;
            elsif this.tvy < this.vy then
                if this.ay >= INSTANT_ACCELERATION then
                   this.vy := this.tvy;
                else
                    this.vy := this.vy - (dt * this.ay);
                    if this.vy < this.tvy then
                        this.vy := this.tvy;
                    end if;
                end if;
            end if;
        end if;

        -- constrain velocity
        if this.vy > this.manager.Max_VY then
            this.vy := this.manager.Max_VY;
        elsif this.vy < -this.manager.Max_VY then
            this.vy := -this.manager.Max_VY;
        end if;
        if this.vx > this.manager.Max_VX then
            this.vx := this.manager.Max_VX;
        elsif this.vx < -this.manager.Max_VX then
            this.vx := -this.manager.Max_VX;
        end if;

        -- move according to velocity
        declare
            xDist : constant Float := dt * this.vx;
            yDist : constant Float := dt * this.vy;
        begin
            if xDist /= 0.0 or else yDist /= 0.0 then
                this.x := this.x + xDist;
                this.y := this.y + yDist;
                this.collide := True;
            end if;
        end;

        -- apply friction
        if this.physical then
            if this.tvx = 0.0 and then this.manager.Friction_X > 0.0 then
                this.vx := this.vx * this.manager.Friction_X;
                if this.vx > -8.0 and then this.vx < 8.0 then
                    this.vx := 0.0;
                end if;
            end if;
            if this.tvy = 0.0 and then this.manager.Friction_Y > 0.0 then
                this.vy := this.vy * this.manager.Friction_Y;
                if this.vy > -8.0 and then this.vy < 8.0 then
                    this.vy := 0.0;
                end if;
            end if;
        end if;
    end Apply_Forces;

    ----------------------------------------------------------------------------

    procedure Clip( this : not null access Corpus'Class; map : not null A_Clip_Map ) is

        ------------------------------------------------------------------------

        procedure Notify_Hit_Wall( dir : Cardinal_Direction ) is
        begin
            case dir is
                when Left | Right =>
                    if this.x /= this.oldX then
                        Queue_Entity_Hit_Wall( this.eid, dir );
                    end if;
                when Up | Down =>
                    if this.y /= this.oldY then
                        Queue_Entity_Hit_Wall( this.eid, dir );
                    end if;
            end case;
        end Notify_Hit_Wall;

        ------------------------------------------------------------------------

        procedure Clip_Feet is

            --------------------------------------------------------------------

            -- Clips the given point (in world coordinates) downward, returning
            -- the vertical adjustment to the point in pixels that is necessary
            -- to keep it out of the ground.
            function Clip_Down( x, y : Float; tileWidth : Integer ) return Integer is
                adjustY : Integer;
            begin
                Clip_Downward( map.Get( x, y ), tileWidth,
                               Integer(x) mod tileWidth,
                               Integer(y) mod tileWidth, adjustY );
                return adjustY;
            end Clip_Down;
            pragma Inline( Clip_Down );

            --------------------------------------------------------------------

            tileWidth   : constant Integer := map.Tile_Width;
            wasStanding : constant Boolean := this.standing;
            feetX       : constant Float := this.x;
            feetY       : constant Float := this.y + this.height / 2;
            adjustY     : Integer := 0;
            adjustY2    : Integer := 0;
        begin
            this.standing := False;
            if this.vy = 0.0 and then this.manager.Gravity /= 0.0 then
                -- check if entity is standing on some kind of surface
                adjustY := Clip_Down( feetX, feetY + 1.0, tileWidth );
                if adjustY < 0 then
                    if adjustY = -1 then
                        -- entity is standing perfectly on top of the ground.
                        -- it's still possible that the tile directly above the
                        -- ground here is a slope, so we check the location
                        -- directly at the entity's feet.
                        adjustY := Clip_Down( feetX, feetY, tileWidth );
                        if adjustY /= 0 then
                            -- need to move the entity's feet upward.
                            this.y := this.y + Float(adjustY);
                            Notify_Hit_Wall( Down );
                            this.collide := True;
                            this.vy := 0.0;
                        end if;
                    else
                        -- entity is standing slightly below the ground. it's
                        -- probably moving up a slope. entity is still standing
                        -- and we need to move its feet upward. we add 1 to
                        -- adjustY here because adjustY was calculated for the
                        -- location one pixel beneath the entity's feet.
                         this.y := this.y + Float(adjustY + 1);
                         Notify_Hit_Wall( Down );
                         this.collide := True;
                         this.vy := 0.0;
                    end if;
                    this.standing := True;
                elsif wasStanding then
                    -- entity is above the ground but was standing before now.
                    -- this means either the ground disappeared or the entity
                    -- is walking down a slope and needs to be clamped down to
                    -- the ground. check a short distance underneath the
                    -- entity's feet for some ground to clamp to.
                    adjustY := Clip_Down( feetX, feetY + Float(tileWidth / 4), tileWidth );
                    if adjustY /= 0 then
                        this.y := this.y + Float(tileWidth / 4) + Float(adjustY);
                        this.collide := True;
                        this.standing := True;
                    end if;
                end if;
            elsif this.vy >= 0.0 then
                -- clipping for bottom edge (single point)
                adjustY := Clip_Down( feetX, feetY, tileWidth );
                if adjustY /= 0 then
                    adjustY2 := Clip_Down( feetX, feetY - Float(tileWidth), tileWidth );
                    if adjustY2 /= 0 then
                        this.y := this.y + Float(adjustY2 - tileWidth);
                    else
                        this.y := this.y + Float(adjustY);
                    end if;
                    Notify_Hit_Wall( Down );
                    this.vy := 0.0;
                    this.standing := True;
                    this.collide := True;
                end if;
            end if;

            -- check for grounding events
            if this.physical and then this.manager.Gravity /= 0.0 then
                if not wasStanding and then this.standing then
                    Queue_Entity_Grounded( this.eid, True );
                    pragma Debug( Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                                       " landed on the ground",
                                       D_PHYSICS, INFO ) );
                elsif wasStanding and then not this.standing then
                    Queue_Entity_Grounded( this.eid, False );
                    pragma Debug( Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                                       " left the ground",
                                       D_PHYSICS, INFO ) );
                end if;
            end if;
        end Clip_Feet;

        ------------------------------------------------------------------------

        procedure Clip_Head is
            tileWidth : constant Integer := map.Tile_Width;
            headX     : constant Float := this.x;
            headY     : constant Float := this.y - this.height / 2;
            headYint  : constant Integer := Integer(Float'Floor( headY ));
            adjustY   : Integer;
        begin
            -- clipping for top edge (single point)
            Clip_Upward( map.Get( this.x, Float'Floor( headY ) ),
                         map.Tile_Width, Integer(headX) mod tileWidth,
                         headYint mod tileWidth, adjustY );
            if adjustY /= 0 then
                this.y := this.y + adjustY;
                Notify_Hit_Wall( Up );
                this.collide := True;
                this.vy := 0.0;
                pragma Debug( this.manager.Gravity /= 0.0,
                              Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                                   " hit the ceiling",
                                   D_PHYSICS, INFO ) );
            end if;
        end Clip_Head;

        ------------------------------------------------------------------------

        procedure Clip_Sides is
            topTile   : constant Integer := (Integer(this.y) - this.height_2 + 4) / map.Tile_Width;
            botTile   : constant Integer := (Integer(this.y) + this.height_2 - 4) / map.Tile_Width;
            leftX     : constant Float := this.x - this.width_2;
            leftXint  : constant Integer := Integer(Float'Floor( leftX ));
            rightX    : Float;
            rightXint : Integer;
        begin
            -- clipping for left edge
            for yTile in topTile..botTile loop
                if map.Get( leftXint / map.Tile_Width, yTile ) = Wall then
                    this.x := this.x + (Float((leftXint / map.Tile_Width + 1) * map.Tile_Width) - leftX);
                    Notify_Hit_Wall( Left );
                    this.vx := 0.0;
                    this.collide := True;
                    exit;
                end if;
            end loop;

            -- clipping for right edge
            rightX := this.x + this.width_2;
            rightXint := Integer(Float'Ceiling( rightX ));
            for yTile in topTile..botTile loop
                if map.Get( rightXint / map.Tile_Width, yTile ) = Wall then
                    this.x := this.x - (rightX - Float((rightXint - (rightXint mod map.Tile_Width)) - 1));
                    Notify_Hit_Wall( Right );
                    this.vx := 0.0;
                    this.collide := True;
                    exit;
                end if;
            end loop;
        end Clip_Sides;

        ------------------------------------------------------------------------

    begin
        -- stay within map boundaries
        if this.x < 0.0 then
            this.x := 0.0;
            Notify_Hit_Wall( Left );
            this.vx := 0.0;
            this.collide := True;
        elsif this.x > Float(map.Get_Width - 1) then
            this.x := Float(map.Get_Width - 1);
            Notify_Hit_Wall( Right );
            this.vx := 0.0;
            this.collide := True;
        end if;
        if this.y < 0.0 then
            this.y := 0.0;
            Notify_Hit_Wall( Up );
            this.collide := True;
            this.vy := 0.0;
        elsif this.y > Float(map.Get_Height - 1) then
            this.y := Float(map.Get_Height - 1);
            Notify_Hit_Wall( Down );
            this.collide := True;
            this.vy := 0.0;
        end if;

        -- clip to map tiles
        if this.clipped then
            Clip_Feet;
            if this.vy <= 0.0 then
                Clip_Head;
            end if;
            Clip_Sides;
        end if;

        -- todo: move attachments
    end Clip;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Corpus;
                         id       : Entity_Id;
                         width,
                         height   : Natural;
                         physical : Boolean;
                         clipped  : Boolean;
                         manager  : access Physics.Managers.Physics_Manager'Class ) is
    begin
        Object(this.all).Construct;
        this.manager := manager;
        this.eid := id;
        this.width := width;
        this.width_2 := width / 2;
        this.height := height;
        this.height_2 := height / 2;
        this.physical := physical;
        this.clipped := clipped;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Corpus ) is
    begin
        this.touching.Clear;
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Detect_Collision( this : not null access Corpus'Class;
                               that : not null A_Corpus ) return Boolean is
        this_x : constant Integer := Integer(this.x);
        this_y : constant Integer := Integer(this.y);
        that_x : constant Integer := Integer(that.x);
        that_y : constant Integer := Integer(that.y);
    begin
        if this_y + this.height_2 < that_y - that.height_2 or else    -- bottom1 < top2
           this_y - this.height_2 > that_y + that.height_2 or else    -- top1 > bottom2
           this_x + this.width_2 < that_x - that.width_2 or else      -- right1 < left2
           this_x - this.width_2 > that_x + that.width_2              -- left1 > right2
        then
            return False;
        end if;
        return True;
    end Detect_Collision;

    ----------------------------------------------------------------------------

    function Get_Id( this : not null access Corpus'Class ) return Entity_Id is
    begin
        return this.eid;
    end Get_Id;

    ----------------------------------------------------------------------------

    function Is_Touching( this : not null access Corpus'Class;
                          that : not null A_Corpus ) return Boolean is
    begin
        return this.touching.Contains( that.eid );
    end Is_Touching;

    ----------------------------------------------------------------------------

    procedure Mark_Contact( this      : not null access Corpus'Class;
                            that      : not null A_Corpus;
                            inContact : Boolean ) is
        pos      : Corps_Maps.Cursor;
        inserted : Boolean;
        pragma Unreferenced( pos, inserted );
    begin
        if inContact then
            this.touching.Insert( that.eid, that, pos, inserted );
        else
            this.touching.Delete( that.eid );
        end if;
    exception
        when others =>
            null;
    end Mark_Contact;

    ----------------------------------------------------------------------------

    procedure Move_To( this : not null access Corpus'Class;
                       map  : not null A_Clip_Map;
                       x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
        this.collide := True;
        this.vx := 0.0;
        this.vy := 0.0;
        this.Clip( map );
        Queue_Entity_Moved( this.eid, this.x, this.y, this.vx, this.vy );
    end Move_To;

    ----------------------------------------------------------------------------

    procedure Set_AX( this : not null access Corpus'Class; ax : Float ) is
    begin
        this.ax := ax;
    end Set_AX;

    ----------------------------------------------------------------------------

    procedure Set_AY( this : not null access Corpus'Class; ay : Float ) is
    begin
        this.ay := ay;
    end Set_AY;

    ----------------------------------------------------------------------------

    procedure Set_Clipped( this    : not null access Corpus'Class;
                           clipped : Boolean ) is
    begin
        this.clipped := clipped;
        pragma Debug( Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                           " clipping is " & Boolean'Image( this.clipped ),
                           D_PHYSICS, INFO ) );
        this.collide := True;
        if not this.clipped and then this.standing then
            -- entities leave the ground when no longer clipped
            this.standing := False;
            Queue_Entity_Grounded( this.eid, False );
            pragma Debug( Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                               " left the ground",
                               D_PHYSICS, INFO ) );
        end if;
    end Set_Clipped;

    ----------------------------------------------------------------------------

    procedure Set_Location( this : not null access Corpus'Class;
                            map  : not null A_Clip_Map;
                            x, y : Float ) is
    begin
        this.x := x;
        this.y := y;
        this.collide := True;
        this.Clip( map );
    end Set_Location;

    ----------------------------------------------------------------------------

    procedure Set_Physical( this     : not null access Corpus'Class;
                            physical : Boolean ) is
    begin
        this.physical := physical;
        pragma Debug( Dbg( "Entity" & Entity_Id'Image( this.eid ) &
                           " physical is " & Boolean'Image( this.physical ),
                           D_PHYSICS, INFO ) );
    end Set_Physical;

    ----------------------------------------------------------------------------

    procedure Set_Size( this   : not null access Corpus'Class;
                        map    : not null A_Clip_Map;
                        width,
                        height : Natural ) is
    begin
        this.width := width;
        this.height := height;
        this.width_2 := this.width / 2;
        this.height_2 := this.height / 2;
        this.collide := True;
        this.Clip( map );
        Queue_Entity_Resized( this.eid, this.width, this.height );
    end Set_Size;

    ----------------------------------------------------------------------------

    procedure Set_Target_VX( this : not null access Corpus'Class; vx : Float ) is
    begin
        this.tvx := vx;
    end Set_Target_VX;

    ----------------------------------------------------------------------------

    procedure Set_Target_VY( this : not null access Corpus'Class; vy : Float ) is
    begin
        this.tvy := vy;
    end Set_Target_VY;

    ----------------------------------------------------------------------------

    procedure Set_Velocity( this : not null access Corpus'Class; vx, vy : Float ) is
    begin
        this.vx := vx;
        this.vy := vy;
    end Set_Velocity;

    ----------------------------------------------------------------------------

    procedure Tick( this    : not null access Corpus'Class;
                    map     : not null A_Clip_Map;
                    dt      : Float;
                    collide : out Boolean ) is
    begin
        this.oldX := this.x;
        this.oldY := this.y;

        this.Apply_Forces( dt );
        this.Clip( map );

        if this.x /= this.oldX or else this.y /= this.oldY then
            Queue_Entity_Moved( this.eid, this.x, this.y, this.vx, this.vy );
        end if;

        collide := this.collide;
        this.collide := False;
    end Tick;

    ----------------------------------------------------------------------------

    function To_String( this : access Corpus'Class ) return String is
    begin
        if this /= null then
            return "<Corpus - id:" & Entity_Id'Image( this.eid ) &
                   ", x: " & Image( this.x ) & ", y: " & Image( this.y ) &
                   ", physical: " & Boolean'Image( this.physical ) & ">";
        else
            return "<Corpus - null>";
        end if;
    end To_String;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Corpus ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Physics.Bodies;
