
with Locking_Objects;                   use Locking_Objects;
with Objects;                           use Objects;
with Processes;                         use Processes;

private with Ada.Containers.Doubly_Linked_Lists;
private with Ada.Strings.Unbounded;

package Processes.Managers is

    type Process_Manager is new Limited_Object with private;
    type A_Process_Manager is access all Process_Manager'Class;

    -- Reference to a procedure that is responsible for destroying a process.
    type A_Destructor is access procedure( process : in out A_Process );

    -- An update rate of 0 hertz means maximum update rate, without delay
    -- between frames. 'minHertz' is not the minimum actual rate of the process
    -- manager because it is unable to interrupt or skip processes; it is the
    -- minimum time delta reported to the processes on tick. If some processes
    -- can't handle large time deltas but sometimes they do happen, 'minHertz'
    -- will prevent the processes from getting a dt that's too big.
    function Create_Process_Manager( name     : String;
                                     hertz    : Natural := 0;
                                     minHertz : Natural := 0 ) return A_Process_Manager;
    pragma Postcondition( Create_Process_Manager'Result /= null );

    -- Attach a process to the process manager for execution. It will be
    -- executed last in the rotation of attached processes. This can be done
    -- before or after starting the process manager.
    procedure Attach( this : not null access Process_Manager'Class;
                      proc : not null A_Process );

    -- Detaches a process from the process manager. This can be done at any time.
    procedure Detach( this       : not null access Process_Manager'Class;
                      proc       : not null A_Process;
                      destructor : A_Destructor := null );

    -- Returns the name of the process manager as given at creation.
    function Get_Name( this : not null access Process_Manager'Class ) return String;

    -- Returns the rate in Hz that the process manager is ticking its processes.
    function Get_Rate( this : not null access Process_Manager'Class ) return Natural;

    -- Pauses/resumes the given process if it's attached to this manager. If
    -- paused, Tick will not be called until after the process has been resumed.
    procedure Pause( this   : not null access Process_Manager'Class;
                     proc   : not null A_Process;
                     paused : Boolean );

    -- Starts execution of the processes.
    procedure Start( this : not null access Process_Manager'Class );

    -- Stops execution of the processes.
    procedure Stop( this : not null access Process_Manager'Class );

    -- Instructs the process manager to track the rate at which it is executing
    -- its processes. This is useful for determining frame rates or finding
    -- which process manager threads are running too slowly.
    procedure Track_Rate( this    : not null access Process_Manager'Class;
                          enabled : Boolean );

    -- Deletes a process manager. Its attached processes, if it has any at the
    -- time of deletion, are detached and left unchanged.
    procedure Delete( this : in out A_Process_Manager );
    pragma Postcondition( this = null );

private

    use Ada.Strings.Unbounded;

    type Execution is
        record
            process   : A_Process := null;
            firstTick : Time := Time_First;
            lastTick  : Time := Time_First;
            paused    : Boolean := False;
            pauseTime : Time := Time_First;
        end record;
    type A_Execution is access all Execution;

    procedure Delete( exec : in out A_Execution );

    function Eq( l, r : A_Execution ) return Boolean;

    package Execution_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Execution, Eq );
    use Execution_Lists;

    ----------------------------------------------------------------------------

    type Operation is abstract tagged
        record
            process : A_Process := null;
        end record;
    type A_Operation is access all Operation'Class;

    procedure Execute( this  : access Operation;
                       plist : in out Execution_Lists.List ) is abstract;

    procedure Delete( this : in out A_Operation );

    type Attach_Operation is new Operation with null record;

    procedure Execute( this     : access Attach_Operation;
                       execList : in out Execution_Lists.List );

    type Detach_Operation is new Operation with
        record
            destroy : A_Destructor := null;
        end record;

    procedure Execute( this     : access Detach_Operation;
                       execList : in out Execution_Lists.List );

    type Pause_Operation is new Operation with null record;

    procedure Execute( this     : access Pause_Operation;
                       execList : in out Execution_Lists.List );

    type Resume_Operation is new Operation with null record;

    procedure Execute( this     : access Resume_Operation;
                       execList : in out Execution_Lists.List );

    function Create_Attach_Operation( process : not null A_Process ) return A_Operation;

    function Create_Detach_Operation( process    : not null A_Process;
                                      destructor : A_Destructor ) return A_Operation;

    function Create_Pause_Operation( process : not null A_Process ) return A_Operation;

    function Create_Resume_Operation( process : not null A_Process ) return A_Operation;

    package Operation_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Operation, "=" );

    ----------------------------------------------------------------------------

    task type Ticker_Task is
        entry Init( pman : A_Process_Manager; hertz, minHertz : Natural );
        entry Start;
        entry Stop;
    end Ticker_Task;
    type A_Ticker_Task is access all Ticker_Task;

    procedure Delete( ticker : in out A_Ticker_Task );

    ----------------------------------------------------------------------------

    type Process_Manager is new Limited_Object with
        record
            lock       : A_Locking_Object;       -- protects all other fields
            name       : Unbounded_String;
            operations : Operation_Lists.List;
            trackRate  : Boolean := False;
            rate       : Natural := 0;
            ticker     : A_Ticker_Task := null;
            started    : Boolean := False;
            stopped    : Boolean := False;
        end record;

    procedure Construct( this     : access Process_Manager;
                         name     : String;
                         hertz    : Natural;
                         minHertz : Natural );

    procedure Delete( this : in out Process_Manager );

end Processes.Managers;
