
with Ada.Unchecked_Conversion;
with System.Memory;

package body Recycling_Pools is

    package SSE renames System.Storage_Elements;
    use type SSE.Storage_Offset;

    Pointer_Size  : constant SSE.Storage_Offset := Address'Size / Storage_Unit;
    Pointers_Size : constant SSE.Storage_Offset := 2 * Pointer_Size;

    type Acc_Address is access all Address;
    function To_Acc_Address is new Ada.Unchecked_Conversion( Address, Acc_Address );

    ----------------------------------------------------------------------------

    protected body Simple_Lock is

        entry Lock when unlocked is
        begin
            unlocked := False;
        end Lock;

        procedure Unlock is
        begin
            unlocked := True;
        end Unlock;

    end Simple_Lock;

    ----------------------------------------------------------------------------

    -- Given an address of a block, return an access to the next block
    function Next( A : Address ) return Acc_Address is
    begin
        return To_Acc_Address( A );
    end Next;
    pragma Inline( Next );

    ----------------------------------------------------------------------------

    -- Given an address of a block, return an access to the previous block
    function Prev( A : Address ) return Acc_Address is
    begin
        return To_Acc_Address( A + Pointer_Size );
    end Prev;
    pragma Inline( Prev );

    ----------------------------------------------------------------------------

    procedure Allocate( this        : in out Recycling_Pool;
                        address     : out System.Address;
                        storageSize : SSE.Storage_Count;
                        alignment   : SSE.Storage_Count ) is
        -- The call to Alloc returns an address whose alignment is compatible
        -- with the worst case alignment requirement for the machine; thus the
        -- alignment argument can be safely ignored.
        pragma Unreferenced( alignment );

        allocated : System.Address;
    begin
        this.lock.Lock;

        if this.freed = Null_Address then
            allocated := Memory.Alloc( Memory.size_t(storageSize + Pointers_Size) );

            if allocated = Null_Address then
                raise Storage_Error;
            else
                -- the requested block begins after the next/prev pointers
                address := allocated + Pointers_Size;

                -- add the allocated block to the used list
                --
                -- allocated.next := this.used;
                -- allocated.prev := null;
                -- if this.used /= null then
                --     this.used.prev := allocated;
                -- end if;
                Next( allocated ).all := this.used;
                Prev( allocated ).all := Null_Address;
                if this.used /= Null_Address then
                    Prev( this.used ).all := allocated;
                end if;
                this.used := allocated;
            end if;
        else
            raise Storage_Error;
        end if;

        this.lock.Unlock;
    exception
        when others =>
            this.lock.Unlock;
            raise;
    end Allocate;

    ----------------------------------------------------------------------------

    procedure Deallocate( this        : in out Recycling_Pool;
                          address     : System.Address;
                          storageSize : SSE.Storage_Count;
                          alignment   : SSE.Storage_Count ) is
        pragma Unreferenced( storageSize );
        pragma Unreferenced( alignment );

        allocated : constant System.Address := address - Pointers_Size;
    begin
        this.lock.Lock;

        -- remove block from the used list
        --
        -- if allocated.prev = null then
        --     used := allocated.next;
        -- else
        --     allocated.prev.next := allocated.next;
        -- end if;
        -- if allocated.next /= null then
        --     allocated.next.prev = allocated.prev;
        -- end if;
        if Prev( allocated ).all = Null_Address then
            this.used := Next( allocated ).all;
            Prev( this.used ).all := Null_Address;
        else
            Next( Prev( allocated ).all ).all := Next( allocated ).all;
        end if;
        if Next( allocated ).all /= Null_Address then
            Prev( Next( allocated ).all).all := Prev( allocated ).all;
        end if;

        Memory.Free( allocated );

        this.lock.Unlock;
    exception
        when others =>
            this.lock.Unlock;
            raise;
    end Deallocate;

    ----------------------------------------------------------------------------

    procedure Finalize( this : in out Recycling_Pool ) is
        n         : System.Address;
        allocated : System.Address;
    begin
        this.lock.Lock;

        -- clear the used list
        n := this.used;
        while n /= Null_Address loop
            allocated := n;
            n := Next( n ).all;
            Memory.Free( allocated );
        end loop;

        -- clear the freed list
        n := this.freed;
        while n /= Null_Address loop
            allocated := n;
            n := Next( n ).all;
            Memory.Free( allocated );
        end loop;

        this.lock.Unlock;
    exception
        when others =>
            this.lock.Unlock;
            raise;
    end Finalize;

    ----------------------------------------------------------------------------

    function Storage_Size( this : Recycling_Pool ) return SSE.Storage_Count is
        pragma Unreferenced( this );
    begin
        return SSE.Storage_Count'Last;
    end Storage_Size;

end Recycling_Pools;
