
with Allegro.Graphics;                  use Allegro.Graphics;
with Allegro.Sprites;                   use Allegro.Sprites;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with hqx;                               use hqx;
with Interfaces;                        use Interfaces;
with Scale_X;                           use Scale_X;
with Super2xSaI;                        use Super2xSaI;
with Support;                           use Support;

package body Scaling is

    initialized : Boolean := False;
    finalized   : Boolean := False;
    buffer      : A_Bitmap := null;

    ----------------------------------------------------------------------------

    procedure Initialize( color_depth : Integer ) is
    begin
        if not initialized then
            pragma Debug( Dbg( "Initializing scaling filters", D_INIT, INFO ) );
            if Init_2xSaI( color_depth ) /= 0 then
                raise INIT_EXCEPTION with "Failed to initialize 2xSaI";
            end if;
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is
    begin
        if not finalized then
            pragma Debug( Dbg( "Finalizing scaling filters", D_INIT, INFO ) );
            Destroy_Bitmap( buffer );
            finalized := True;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    procedure Scale( src,
                     dst           : not null A_Bitmap;
                     src_width,
                     src_height    : Positive;
                     magnification : Float;
                     filter        : Filter_Type := Filter_Nearest ) is
        actualFilter : Filter_Type := filter;
    begin
        if src_width > Get_Width( src ) or else
           src_height > Get_Height( src ) or else
           Integer( src_width * magnification ) > Get_Width( dst ) or else
           Integer( src_height * magnification ) > Get_Height( dst )
        then
            -- prevent something stupid from happening
            pragma Debug( Dbg( "Illegal scaling: [" &
                               Image( Get_Width( src ) ) & ", " &
                               Image( Get_Height( src ) ) & "] " &
                               Image( src_width ) & ", " &
                               Image( src_height ) & " x " &
                               Image( magnification ) & " -> [" &
                               Image( Get_Width( dst ) ) & ", " &
                               Image( Get_Height( dst ) ) & "]",
                               D_GUI, ERROR ) );
            return;
        end if;

        -- dynamically choose best quality filter
        if actualFilter = Filter_Quality then
            if magnification = 2.0 then
                if Get_Color_Depth = 32 then
                    actualFilter := Filter_HQX;
                else
                    actualFilter := Filter_2xSaI;
                end if;
            elsif magnification = 3.0 then
                if Get_Color_Depth = 32 then
                    actualFilter := Filter_HQX;
                else
                    actualFilter := Filter_ScaleX;
                end if;
            elsif magnification = 4.0 then
                if Get_Color_Depth = 32 then
                    actualFilter := Filter_HQX;
                else
                    -- may not work if source bitmap isn't large enough
                    actualFilter := Filter_2xSaI;
                end if;
            end if;
        end if;

        case actualFilter is

            when Filter_Nearest | Filter_Quality => null;

            when Filter_HQX =>
                if Get_Color_Depth = 32 then
                    if magnification = 2.0 then
                        hq2x( Get_Dat( src ), Get_Width( src ),
                              Get_Dat( dst ), Get_Width( dst ),
                              src_width, src_height );
                        return;
                    elsif magnification = 3.0 then
                        hq3x( Get_Dat( src ), Get_Width( src ),
                              Get_Dat( dst ), Get_Width( dst ),
                              src_width, src_height );
                        return;
                    elsif magnification = 4.0 then
                        hq4x( Get_Dat( src ), Get_Width( src ),
                              Get_Dat( dst ), Get_Width( dst ),
                              src_width, src_height );
                        return;
                    end if;
                end if;

            when Filter_2xSaI =>
                if magnification = 2.0 then
                    Super2xSaI.Super2xSaI( src, dst, 0, 0, 0, 0, src_width, src_height );
                    return;
                elsif magnification = 4.0 then
                    -- 4xSaI is just 2xSaI applied twice
                    -- The buffer bitmap must be at least twice the size of the
                    -- source image to handle the 2x intermediate.
                    if not finalized then
                        if buffer = null or else
                           Get_Width( buffer ) < src_width * 2 or else
                           Get_Height( buffer ) < src_height * 2
                        then
                            Destroy_Bitmap( buffer );
                            buffer := Create_Bitmap( src_width * 2, src_height * 2 );
                        end if;

                        Super2xSaI.Super2xSaI( src, dst, 0, 0, 0, 0, src_width, src_height );
                        Blit( dst, buffer, 0, 0, 0, 0, src_width * 2, src_height * 2 );
                        Super2xSaI.Super2xSaI( buffer, dst, 0, 0, 0, 0, src_width * 2, src_height * 2 );
                        return;
                    end if;
                end if;

            when Filter_Eagle =>
                if magnification = 2.0 then
                    Super2xSaI.SuperEagle( src, dst, 0, 0, 0, 0, src_width, src_height );
                    return;
                end if;

            when Filter_ScaleX =>
                if Get_Color_Depth = 32 then
                    if magnification = 2.0 then
                        Scale2x( Get_Dat( dst ), Unsigned_32(Get_Width( dst )),
                                 Get_Dat( src ), Unsigned_32(Get_Width( src )),
                                 src_width, src_height );
                        return;
                    elsif magnification = 3.0 then
                        Scale3x( Get_Dat( dst ), Unsigned_32(Get_Width( dst )),
                                 Get_Dat( src ), Unsigned_32(Get_Width( src )),
                                src_width, src_height );
                        return;
                    elsif magnification = 4.0 then
                        -- Scale4x is just Scale2x applied twice
                        -- The source bitmap must be at least twice the size of the
                        -- original image to handle the 2x blit back.
                        if Get_Width( src ) >= src_width * 2 and then
                           Get_Height( src ) >= src_height * 2
                        then
                            Scale2x( Get_Dat( dst ), Unsigned_32(Get_Width( dst )),
                                     Get_Dat( src ), Unsigned_32(Get_Width( src )),
                                     src_width, src_height );
                            Blit( dst, src, 0, 0, 0, 0, src_width * 2, src_height * 2 );
                            Scale2x( Get_Dat( dst ), Unsigned_32(Get_Width( dst )),
                                     Get_Dat( src ), Unsigned_32(Get_Width( src )),
                                     src_width * 2, src_height * 2 );
                            return;
                        end if;
                    end if;
                end if;

        end case;

        -- Nearest-neighbor filter
        Stretch_Blit( src, dst, 0, 0, src_width, src_height, 0, 0,
                      Integer(src_width * magnification),
                      Integer(src_height * magnification) );
    end Scale;

    ----------------------------------------------------------------------------

    function To_Filter( name    : String;
                        default : Filter_Type := Filter_Type'First
                      ) return Filter_Type is
        filter : Filter_Type;
    begin
        filter := Filter_Type'Value( "Filter_" & name );
        if not filter'Valid then
            return default;
        end if;
        return filter;
    exception
        when others =>
            return default;
    end To_Filter;

end Scaling;
