
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
--with Debugging;                         use Debugging;
with GNAT.Directory_Operations;         use GNAT.Directory_Operations;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;

package body Support.Paths is

    -- Returns the location of the '.' character which begins the path's file
    -- extension. If no extension is found, path'First - 1 is returned.
    function Find_Extension( path : String ) return Integer is

        ------------------------------------------------------------------------

        function Is_Dir_Sep( c : Character ) return Boolean is
        begin
            return c = '/' or else c = '\' or else c = '#';
        end Is_Dir_Sep;

        ------------------------------------------------------------------------

        pos : Integer := path'Last;
    begin
        while pos > path'First and then
              not Is_Dir_Sep( path(pos) ) and then
              path(pos) /= '.'
        loop
            pos := pos - 1;
        end loop;
        -- to find the extension, we must be somewhere after the first character,
        -- at or before the last character, at a dot character and the character
        -- before the dot can't be a directory separator.
        if pos + 1 <= path'Last and then pos > path'First and then
           path(pos) = '.' and then not Is_Dir_Sep( path(pos-1) )
        then
           return pos;
        end if;
        return path'First - 1;
    end Find_Extension;

    ----------------------------------------------------------------------------

    function Add_Extension( path,
                            ext   : String;
                            force : Boolean := False ) return String is
    begin
        if ext'Length > 0 then
            if Get_Extension( path )'Length = 0 or else
               (force and then not Ends_With( To_Lower( path ), "." & To_Lower( ext ) ))
            then
                -- add the extension if we're forcing it, or if there isn't an
                -- extension
                return path & "." & ext;
            end if;
        end if;
        return path;
    end Add_Extension;

    ----------------------------------------------------------------------------

    function Ancestor_Dir( fullpath, filename : String ) return String is
    begin
        if Ends_With( fullpath, filename ) then
            return fullpath(fullpath'First..fullpath'Last-filename'Length);
        end if;
        return fullpath;
    end Ancestor_Dir;

    ----------------------------------------------------------------------------

    function Executable_Extension return String renames OS.Executable_Path;

    ----------------------------------------------------------------------------

    function Get_Directory( path : String ) return String is
        pos : constant Integer := Index( path, Slash, Backward );
    begin
        if pos >= path'First then
            return path(path'First..pos);
        end if;
        return "";
    end Get_Directory;

    ----------------------------------------------------------------------------

    function Get_Extension( path : String ) return String is
        pos : constant Integer := Find_Extension( path );
    begin
        if pos >= path'First then
            return path(pos+1..path'Last);
        end if;
        return "";
    end Get_Extension;

    ----------------------------------------------------------------------------

    function Get_Filename( path : String ) return String is
        pos : constant Integer := Index( path, Slash, Backward );
    begin
        if pos >= path'First and then pos < path'Last then
            return path(pos+1..path'Last);
        end if;
        return path;
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Home_Directory return String renames OS.Home_Directory;

    ----------------------------------------------------------------------------

    function Path_Append( basePath, extra : String ) return String is

        ------------------------------------------------------------------------

        function No_Leading( str : String ) return String is
        begin
            if str'Length > 0 and then
               (str(str'First) = '/' or else str(str'First) = '\')
            then
                return No_Leading( str(str'First+1..str'Last) );
            end if;
            return str;
        end No_Leading;

        ------------------------------------------------------------------------

        function No_Trailing( str : String ) return String is
        begin
            if str'Length > 0 and then
               (str(str'Last) = '/' or else str(str'Last) = '\')
            then
                return No_Trailing( str(str'First..str'Last-1) );
            end if;
            return str;
        end No_Trailing;

        ------------------------------------------------------------------------

    begin
        if basePath'Length = 0 and then extra'Length = 0 then
            return "";
        elsif extra'Length = 0 then
            return basePath;
        elsif basePath'Length = 0 then
            return extra;
        end if;

        return No_Trailing( basePath ) & Slash & No_Leading( extra );
    end Path_Append;

    ----------------------------------------------------------------------------

    function Remove_Extension( path : String ) return String is
        pos : constant Integer := Find_Extension( path );
    begin
        if pos >= path'First then
            return path(path'First..pos-1);
        end if;
        return path;
    end Remove_Extension;

    ----------------------------------------------------------------------------

    procedure Set_Working_Directory( path : String := "" ) is
    begin
        if path'Length = 0 then
#if OSX then
            declare
                dir : constant String := Get_Directory( OS.Executable_Path );
            begin
                if Index( dir, ".app/Contents/MacOS", Backward ) > dir'First then
                    -- Inside an app bundle, set the working directory to the
                    -- Resources directory
                    Change_Dir( Normalize_Pathname( dir & "../Resources/" ) );
                else
                    -- Not inside an app bundle, set the working directory to
                    -- the location of the executable
                    Change_Dir( dir );
                end if;
            end;
#else
            Change_Dir( Get_Directory( OS.Executable_Path ) );
#end if;
        else
            Change_Dir( path );
        end if;
    exception
        when Directory_Error =>
            -- failed to change working directory
            null;
    end Set_Working_Directory;

    ----------------------------------------------------------------------------

    function Slash return String is
        result : constant String (1..1) := (1 => GNAT.OS_Lib.Directory_Separator);
    begin
        return result;
    end Slash;

    ----------------------------------------------------------------------------

    function System_Font_Directory return String renames OS.System_Font_Directory;

end Support.Paths;
