
with Ada.Unchecked_Deallocation;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Ada.Real_Time;                     use Ada.Real_Time;
with Ada.Streams;                       use Ada.Streams;
with Objects;                           use Objects;
with Physics;                           use Physics;

private with Ada.Containers.Doubly_Linked_Lists;
private with Ada.Strings.Unbounded;
private with Archives;
private with Locking_Objects;
private with Resources;

package Tiles is

    procedure Initialize;

    procedure Finalize;

    ----------------------------------------------------------------------------

    type Tile_Id_Array is array (Integer range <>) of Natural;
    type A_Tile_Id_Array is access all Tile_Id_Array;

    function A_Tile_Id_Array_Input( stream : access Root_Stream_Type'Class ) return A_Tile_Id_Array;
    for A_Tile_Id_Array'Input use A_Tile_Id_Array_Input;

    procedure A_Tile_Id_Array_Output( stream : access Root_Stream_Type'Class; tia : A_Tile_Id_Array );
    for A_Tile_Id_Array'Output use A_Tile_Id_Array_Output;

    function Copy( src : A_Tile_Id_Array ) return A_Tile_Id_Array;
    pragma Postcondition( Copy'Result /= src or else src = null );

    procedure Delete is new Ada.Unchecked_Deallocation( Tile_Id_Array, A_Tile_Id_Array );

    ----------------------------------------------------------------------------

    type Tile_Object is abstract new Object with private;
    type A_Tile is access all Tile_Object'Class;

    -- Creates a new tile using register allocator. Returns null if no allocator
    -- has been registered.
    function Create_Tile return A_Tile;

    -- Returns the animation frame delay for the tile.
    function Get_Anm_Delay( this : not null access Tile_Object'Class ) return Time_Span;

    -- Returns the id of the tile within its library.
    function Get_Id( this : not null access Tile_Object'Class ) return Natural;

    -- Returns the name of the tile within its library.
    function Get_Name( this : not null access Tile_Object'Class ) return String;

    -- Returns the id of the next frame in the tile's animation if it's a frame
    -- in a single shot animation.
    function Get_Next_Frame( this : not null access Tile_Object'Class ) return Natural;

    -- Returns a reference to the tile's looped animation frame list, or null if
    -- the tile does not have a looping animation. Do not delete the array, it
    -- belongs to the tile.
    function Get_Frame_List( this : not null access Tile_Object'Class ) return A_Tile_Id_Array;

    -- Returns True if the file is animated.
    function Is_Animated( this : not null access Tile_Object'Class ) return Boolean;

    function Object_Input( stream : access Root_Stream_Type'Class ) return Tile_Object is abstract;

    procedure Object_Output( stream : access Root_Stream_Type'Class; obj : Tile_Object ) is abstract;

    -- Sets an attribute on the tile by name. Only certain attribute names are
    -- supported and each has its own value constraints. False is returned in
    -- 'found' if the attribute name is not recognized. An exception is raised
    -- if the value is invalid or if there is a semantic error caused by setting
    -- the attribute.
    procedure Set_Attribute( this  : in out Tile_Object;
                             found : out Boolean;
                             name  : String;
                             val   : String := "" );

    -- Sets the id of the tile.
    procedure Set_Id( this : access Tile_Object; id : Natural );

    -- Sets the name of the tile.
    procedure Set_Name( this : access Tile_Object; name : String );
    pragma Precondition( name'Length > 0 );

    -- Deletes the tile.
    procedure Delete( this : in out A_Tile );
    pragma Postcondition( this = null );

    ATTRIBUTE_ERROR : exception;

private

    use Ada.Strings.Unbounded;
    use Archives;
    use Locking_Objects;
    use Resources;

    ----------------------------------------------------------------------------

    type Tile_Object is abstract new Object with
        record
            id        : Natural := 0;
            name      : Unbounded_String;

            lock      : A_Locking_Object := null;    -- protects bmp, priority
            bmp       : A_Bitmap := null;
            priority  : Integer := 0;

            loadState : A_Async_Operation := null;
            clipping  : Clip_Type := Passive;
            anmDelay  : Time_Span := Time_Span_Zero; -- animation frame delay in ms
            nextFrame : Natural := 0;                -- tile id of next frame in one-shot animation
            frames    : A_Tile_Id_Array := null;     -- tile id list in looping animation
        end record;

    -- Raises COPY_NOT_ALLOWED.
    procedure Adjust( this : access Tile_Object );

    procedure Construct( this : access Tile_Object );

    procedure Delete( this : in out Tile_Object );

    -- Returns null if the tile's bitmap file isn't found or hasn't loaded yet.
    function Get_Bitmap( this : not null access Tile_Object'Class ) return A_Bitmap;

    -- Returns the physics clipping type that the tile represents.
    function Get_Clipping( this : not null access Tile_Object'Class ) return Clip_Type;

    -- Returns True if the tile's bitmap has been loaded, successfully or otherwise.
    function Is_Loaded( this : not null access Tile_Object'Class ) return Boolean;

    -- Loads the tile's bitmap from an archive.
    procedure Load_Bitmap( this    : not null access Tile_Object'Class;
                           archive : not null A_Archive );

    -- Blocks the caller until the tile's bitmap has been loaded.
    procedure Wait_For_Load( this : not null access Tile_Object'Class );

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Tile_Object );
    for Tile_Object'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Tile_Object );
    for Tile_Object'Write use Object_Write;

    ----------------------------------------------------------------------------

    function A_Tile_Input( stream : access Root_Stream_Type'Class ) return A_Tile;
    for A_Tile'Input use A_Tile_Input;

    procedure A_Tile_Output( stream : access Root_Stream_Type'Class; tile : A_Tile );
    for A_Tile'Output use A_Tile_Output;

    procedure A_Tile_Read( stream : access Root_Stream_Type'Class; tile : out A_Tile );
    for A_Tile'Read use A_Tile_Read;

    procedure A_Tile_Write( stream : access Root_Stream_Type'Class; tile : A_Tile );
    for A_Tile'Write use A_Tile_Write;

    ----------------------------------------------------------------------------

    package Tile_Lists is new Ada.Containers.Doubly_Linked_Lists( A_Tile, "=" );
    use Tile_Lists;

    ----------------------------------------------------------------------------

    type Allocator is access function return A_Tile;

    function Tile_Identifier return String;
    pragma Postcondition( Tile_Identifier'Result'Length > 0 );

    -- Returns 0 if a tile version has not yet been registered.
    function Tile_Version return Natural;

    procedure Register_Allocator( allocate : not null Allocator );

    procedure Register_Identifier( identifier : String );
    pragma Precondition( identifier'Length > 0 );

    procedure Register_Version( version : Positive );

end Tiles;
