
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Support;                           use Support;

package body Values is

    not overriding
    function As_Boolean( this : access Value ) return Boolean is
    begin
        raise INVALID_CONVERSION with "Can't convert " &
                                      this.Get_Class_Name &
                                      " to boolean";
        return False;
    end As_Boolean;

    ----------------------------------------------------------------------------

    not overriding
    function As_Integer( this : access Value ) return Integer is
    begin
        raise INVALID_CONVERSION with "Can't convert " &
                                      this.Get_Class_Name &
                                      " to integer";
        return 0;
    end As_Integer;

    ----------------------------------------------------------------------------

    not overriding
    function As_String( this : access Value ) return String is
    begin
        raise INVALID_CONVERSION with "Can't convert " &
                                      this.Get_Class_Name &
                                      " to string";
        return "";
    end As_String;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Value ) is
    begin
        Object'Read( stream, Object(obj) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Value ) is
    begin
        Object'Write( stream, Object(obj) );
    end Object_Write;

    ----------------------------------------------------------------------------

    function A_Value_Input( stream : access Root_Stream_Type'Class ) return A_Value is
        obj : A_Value := null;
    begin
        if Boolean'Input( stream ) then
            A_Value'Read( stream, obj );
        end if;
        return obj;
    end A_Value_Input;

    ----------------------------------------------------------------------------

    procedure A_Value_Output( stream : access Root_Stream_Type'Class; obj : A_Value ) is
    begin
        Boolean'Output( stream, obj /= null );
        if obj /= null then
            A_Value'Write( stream, obj );
        end if;
    end A_Value_Output;

    ----------------------------------------------------------------------------

    procedure A_Value_Read( stream : access Root_Stream_Type'Class; obj : out A_Value ) is
    begin
        obj := new Value'Class'(Value'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown value class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_Value_Read;

    ----------------------------------------------------------------------------

    procedure A_Value_Write( stream : access Root_Stream_Type'Class; obj : A_Value ) is
    begin
         Value'Class'Output( stream, obj.all );
    end A_Value_Write;

    ----------------------------------------------------------------------------

    function Copy( src : A_Value ) return A_Value is
    begin
        return A_Value(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Value ) is
    begin
        Delete( A_Object(this) );
    end Delete;

    --==========================================================================

    function Create_Value( val : Boolean ) return A_Value is
        this : constant A_Value := new Val_Bool;
    begin
        this.Construct;
        Val_Bool(this.all).bool := val;
        return this;
    end Create_Value;

    ----------------------------------------------------------------------------

    overriding
    function As_Boolean( this : access Val_Bool ) return Boolean is
    begin
        return this.bool;
    end As_Boolean;

    ----------------------------------------------------------------------------

    overriding
    function As_Integer( this : access Val_Bool ) return Integer is
    begin
        if this.bool then
            return 1;
        end if;
        return 0;
    end As_Integer;

    ----------------------------------------------------------------------------

    overriding
    function As_String( this : access Val_Bool ) return String is
    begin
        if this.bool then
            return "True";
        end if;
        return "False";
    end As_String;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_Bool is
        this : aliased Val_Bool;
    begin
        this.Construct;
        Val_Bool'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_Bool ) is
    begin
        Value'Read( stream, Value(obj) );
        obj.bool := Boolean'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_Bool ) is
    begin
        Value'Write( stream, Value(obj) );
        Boolean'Output( stream, obj.bool );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Val_Bool ) return String is
    begin
        if this.bool then
            return "True";
        end if;
        return "False";
    end To_String;

    --==========================================================================

    function Create_Value( val : Integer ) return A_Value is
        this : constant A_Value := new Val_Int;
    begin
        this.Construct;
        Val_Int(this.all).int := val;
        return this;
    end Create_Value;

    ----------------------------------------------------------------------------

    overriding
    function As_Boolean( this : access Val_Int ) return Boolean is
    begin
        return this.int /= 0;
    end As_Boolean;

    ----------------------------------------------------------------------------

    overriding
    function As_Integer( this : access Val_Int ) return Integer is
    begin
        return this.int;
    end As_Integer;

    ----------------------------------------------------------------------------

    overriding
    function As_String( this : access Val_Int ) return String is
    begin
        return Image( this.int );
    end As_String;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_Int is
        this : aliased Val_Int;
    begin
        this.Construct;
        Val_Int'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_Int ) is
    begin
        Value'Read( stream, Value(obj) );
        obj.int := Integer'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_Int ) is
    begin
        Value'Write( stream, Value(obj) );
        Integer'Output( stream, obj.int );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Val_Int ) return String is
    begin
        return Image( this.int );
    end To_String;

    --==========================================================================

    function Create_Value( val : String ) return A_Value is
        this : constant A_Value := new Val_String;
    begin
        this.Construct;
        Val_String(this.all).str := To_Unbounded_String( val );
        return this;
    end Create_Value;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Val_String ) is
    begin
        Value(this.all).Adjust;
        this.str := To_Unbounded_String( To_String( this.str ) );
    end Adjust;

    ----------------------------------------------------------------------------

    overriding
    function As_String( this : access Val_String ) return String is
    begin
        return To_String( this.str );
    end As_String;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Val_String is
        this : aliased Val_String;
    begin
        this.Construct;
        Val_String'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Val_String ) is
    begin
        Value'Read( stream, Value(obj) );
        obj.str := To_Unbounded_String( String'Input( stream ) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Val_String ) is
    begin
        Value'Write( stream, Value(obj) );
        String'Output( stream, To_String( obj.str ) );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Val_String ) return String is
    begin
        return """" & To_String( this.str ) & """";
    end To_String;

end Values;
