
with Applications;                      use Applications;
--with Debugging;                         use Debugging;
with Themes.Palette;                    use Themes.Palette;
with Widgets.Containers;                use Widgets.Containers;
with Widgets.Layouts;                   use Widgets.Layouts;

package body Widgets.Containers.Panels is

    procedure Make_Child_Layout( layout       : in out A_Layout;
                                 contentTop   : out Integer;
                                 border       : Border_Type;
                                 titled       : Boolean;
                                 titlePadding : Natural;
                                 textHeight   : Natural ) is
        theme  : constant A_Theme := Get_Application.Get_Theme;
        bw     : constant Integer := theme.Border_Width;
        left,
        top,
        right,
        bottom : Integer := 0;
    begin
        if border /= Border_None then
            left := bw;
            top := bw;
            right := bw;
            bottom := bw;
        end if;
        if titled then
            top := top + titlePadding * 2 + textHeight;
        end if;
        Delete( layout );
        layout := Create_Layout_LTRB( left, top, right, bottom );
        contentTop := top;
    end Make_Child_Layout;

    --==========================================================================

    function Create_Panel( view  : not null access Game_Views.Game_View'Class;
                           id    : String;
                           title : String := "";
                           icon  : String := "" ) return A_Panel is
        this : A_Panel := new Panel;
    begin
        this.Construct( view, id, title, icon );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Panel;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this  : access Panel;
                         view  : not null access Game_Views.Game_View'Class;
                         id    : String;
                         title : String;
                         icon  : String ) is
    begin
        Container(this.all).Construct( view, id );
        this.border := Border_Raised;
        Make_Child_Layout( this.childLayout, this.contentTop, this.border,
                           False, this.titlePadding,
                           Get_Application.Get_Theme.Get_Text_Height( To_String( this.fontName ),
                                                                      this.fontSize ) );
        this.Set_Title( title );
        this.Set_Icon( icon );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Panel; dc : Drawing_Context ) is
        theme : constant A_Theme := Get_Application.Get_Theme;
        th    : constant Integer := theme.Get_Text_Height(  To_String( this.fontName ), this.fontSize );
        tp    : constant Natural := this.titlePadding;
        bw    : Integer := theme.Border_Width;
    begin
        theme.Draw_Box( dc, 0, 0, this.cwidth - 1, this.cheight - 1,
                        this.colors(Foreground), this.border );

        if this.icon /= 0 or else Length( this.title ) > 0 then
            if this.border = Border_None then
                bw := 0;
            end if;
            theme.Draw_Box( dc, bw, bw, this.cwidth - 1 - bw,
                            bw + tp * 2 + th - 1, this.colors(Selected) );
            theme.Draw_Label( dc, bw + tp, bw + tp,
                              this.cwidth - 1 - bw - tp, bw + tp + th,
                              theme.Get_Bitmap( this.icon ),
                              To_String( this.title ),
                              To_String( this.fontName ), this.fontSize,
                              Hard_Contrast( this.colors(Selected) ),
                              Align_Left, smooth => this.fontSmooth );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Panel ) return Natural is
        theme  : A_Theme := null;
        height : Integer := 0;
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        else
            theme := Get_Application.Get_Theme;
            if this.border /= Border_None then
                height := height + theme.Border_Width * 2;
            end if;
            if this.icon /= 0 or else Length( this.title ) > 0 then
                height := height + this.titlePadding * 2 +
                          theme.Label_Height( theme.Get_Bitmap( this.icon ),
                                              To_String( this.title ),
                                              To_String( this.fontName ), this.fontSize );
            end if;
            return height;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Panel ) return Natural is
        theme : A_Theme := null;
        width : Integer := 0;
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        else
            theme := Get_Application.Get_Theme;
            if this.border /= Border_None then
                width := width + theme.Border_Width * 2;
            end if;
            if this.icon /= 0 or else Length( this.title ) > 0 then
                width := width + this.titlePadding * 2 +
                         theme.Label_Width( theme.Get_Bitmap( this.icon ),
                                            To_String( this.title ),
                                            To_String( this.fontName ), this.fontSize );
            end if;
            return width;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Title( this : access Panel ) return String is
    begin
        return To_String( this.title );
    end Get_Title;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Border( this : access Panel; border : Border_Type ) is
        updateLayout : constant Boolean :=
            border /= this.border and then
            (border = Border_None or else this.border = Border_None);
    begin
        Container(this.all).Set_Border( border );
        if updateLayout then
            Make_Child_Layout( this.childLayout,
                               this.contentTop,
                               this.border,
                               titled => Length( this.title ) > 0 or else
                                         this.icon /= 0,
                               titlePadding => this.titlePadding,
                               textHeight => Get_Application.Get_Theme.Get_Text_Height( To_String( this.fontName ),
                                                                                        this.fontSize ) );
        end if;
    end Set_Border;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Color( this    : access Panel;
                         purpose : Color_Purpose;
                         color   : Color_Type ) is
    begin
        Container(this.all).Set_Color( purpose, color );
        if purpose = Foreground then
            this.transparent := (color = base_colors(Transparent));
        end if;
    end Set_Color;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Icon( this : access Panel; icon : String ) is
        id : Integer := 0;
    begin
        if icon'Length > 0 then
            id := Get_Application.Get_Theme.Get_ID( icon );
        end if;

        if id /= this.icon then
            if Length( this.title ) = 0 then
                -- if we don't have a title then adding/removing an icon will
                -- change the child layout
                if this.icon = 0 or else id = 0 then
                    Make_Child_Layout( this.childLayout,
                                       this.contentTop,
                                       this.border,
                                       titled => id /= 0,
                                       titlePadding => this.titlePadding,
                                       textHeight => Get_Application.Get_Theme.Get_Text_Height( To_String( this.fontName ),
                                                                                                this.fontSize ) );
                    this.Pack;
                end if;
            end if;
            this.icon := id;
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Icon;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Title( this : access Panel; title : String ) is
    begin
        if title /= this.title then
            if this.icon = 0 then
                -- if we don't have an icon then adding/removing a title will
                -- change the child layout
                if (title'Length > 0 and then Length( this.title ) = 0) or else
                   (title'Length = 0 and then Length( this.title ) > 0)
                then
                    Make_Child_Layout( this.childLayout,
                                       this.contentTop,
                                       this.border,
                                       titled => title'Length > 0,
                                       titlePadding => this.titlePadding,
                                       textHeight => Get_Application.Get_Theme.Get_Text_Height( To_String( this.fontName ),
                                                                                                this.fontSize ) );
                    this.Pack;
                end if;
            end if;
            this.title := To_Unbounded_String( title );
            A_Widget(this).Set_Dirty;
        end if;
    end Set_Title;

end Widgets.Containers.Panels;
