
with Actions;                           use Actions;
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Applications;                      use Applications;
--with Debugging;                         use Debugging;
with Events.Application;                use Events.Application;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Game_Views;                        use Game_Views;
with Game_Views.Keen.Actions;           use Game_Views.Keen.Actions;
with Game_Views.Keen.Actions.Keys;      use Game_Views.Keen.Actions.Keys;
with Impulses;                          use Impulses;
with Keyboard;                          use Keyboard;
with Palette_16;
with Preferences;                       use Preferences;
with Scaling;                           use Scaling;
with Support;                           use Support;
with Themes;                            use Themes;
with Themes.Palette;                    use Themes.Palette;
with Values;                            use Values;
with Widgets.Buttons;                   use Widgets.Buttons;
with Widgets.Buttons.Pushes;            use Widgets.Buttons.Pushes;
with Widgets.Containers.Panels;         use Widgets.Containers.Panels;
with Widgets.Containers.Scenes;         use Widgets.Containers.Scenes;
with Widgets.Containers.Scenes.Keen;    use Widgets.Containers.Scenes.Keen;
with Widgets.Containers.Keen_Panels;    use Widgets.Containers.Keen_Panels;
with Widgets.Labels;                    use Widgets.Labels;
with Widgets.Progress_Boards;           use Widgets.Progress_Boards;
with Widgets.Scoreboards;               use Widgets.Scoreboards;

pragma Elaborate_All( Game_Views );
pragma Elaborate_All( Preferences );

package body Game_Views.Keen is

    MIN_LOAD_DELAY : constant Time_Span := Seconds( 2 );

    ----------------------------------------------------------------------------

    -- This is called when the scene receives a key event.
    procedure Scene_Key_Handler( action  : A_Key_Action;
                                 handled : out Boolean ) is
        this : constant A_Keen_View := A_Keen_View(action.Get_Source.Get_View);
        key  : constant Integer := action.Get_Key;
    begin
        if action.Get_Id = Key_Press then
            handled := True;
            this.ignoreKey(key) := False;

            if key = KEY_ENTER then
                A_Progress_Board(this.Get_Widget( "progressBoard" )).Toggle;
            elsif key = KEY_P then
                this.Pause_Game( not this.paused );
            elsif key = KEY_ALT then
                Queue_Impulse( this.playerId, Toggle_Pogo );
            elsif key = KEY_SPACE then
                Queue_Impulse( this.playerId, Shoot );
            elsif key = KEY_ESC then
                this.Return_To_Menu;
            else
                handled := False;
            end if;

        elsif action.Get_Id = Key_Held then
            handled := True;
            if this.ignoreKey(key) then
                return;
            end if;

            -- if the game is paused and an action key is pressed then resume
            if this.paused then
                case key is
                    when KEY_LEFT     |
                         KEY_RIGHT    |
                         KEY_UP       |
                         KEY_DOWN     |
                         KEY_SPACE    |
                         KEY_LCONTROL |
                         KEY_ALT =>
                        this.Pause_Game( False );
                        return;
                    when others => null;
                end case;
            end if;

            -- handle the held state of an action key
            case key is
                when KEY_LEFT     => Queue_Impulse( this.playerId, Move_Left_Start );
                when KEY_RIGHT    => Queue_Impulse( this.playerId, Move_Right_Start );
                when KEY_UP       => Queue_Impulse( this.playerId, Move_Up_Start );
                when KEY_DOWN     => Queue_Impulse( this.playerId, Move_Down_Start );
                when KEY_LCONTROL => Queue_Impulse( this.playerId, Jump_Start );
                when others       => handled := False;
            end case;

        elsif action.Get_Id = Key_Release then
            handled := True;
            if this.ignoreKey(key) then
                return;
            end if;

            -- if the key performs an on-going action, send the impulse to stop
            -- it on key release.
            case key is
                when KEY_LEFT     => Queue_Impulse( this.playerId, Move_Left_Stop );
                when KEY_RIGHT    => Queue_Impulse( this.playerId, Move_Right_Stop );
                when KEY_UP       => Queue_Impulse( this.playerId, Move_Up_Stop );
                when KEY_DOWN     => Queue_Impulse( this.playerId, Move_Down_Stop );
                when KEY_LCONTROL => Queue_Impulse( this.playerId, Jump_Stop );
                when others       => handled := False;
            end case;

        else
            -- won't happen
            handled := False;
        end if;
    end Scene_Key_Handler;

    ----------------------------------------------------------------------------

    -- This is called when the progress board opens or closes.
    procedure Progress_Board_Handler( action : A_PBoard_Action ) is
        this : constant A_Keen_View := A_Keen_View(action.Get_Source.Get_View);
    begin
        if action.Get_Id = Opened then
            this.Auto_Pause( True );

        elsif action.Get_Id = Closed then
            this.Auto_Pause( False );

        end if;
    end Progress_Board_Handler;

    --==========================================================================

    function Create_Game_View( xres, yres : Positive;
                               scale      : Positive ) return A_Game_View is
        this : constant A_Keen_View := new Keen_View;
    begin
        this.Construct;
        this.Populate_View( xres, yres, scale );
        return A_Game_View(this);
    end Create_Game_View;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Game_State_Event ) is
    begin
        if evt.Is_Playing then
            -- A game has begun - - - - - - - - - - - - - - - - - - - - - - - -
            this.gameInProgress := True;

            -- clear paused state
            this.Get_Widget( "pnlPaused" ).Set_Visible( False );
            this.autoPaused := False;
            this.pauseEnabled := True;
            this.paused := False;

            this.Show_Game( False );
            this.Show_Menu( False );

        else
            -- A game has ended - - - - - - - - - - - - - - - - - - - - - - - -
            this.gameInProgress := False;

            this.Show_Game( False );
            this.Show_Menu( True );

        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Game_Var_Changed_Event ) is
        var : constant String := To_Lower( evt.Get_Var );
        val : A_Value;
    begin
        evt.Copy_Value( val );
        if var = "points" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Points( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Points( val.As_Integer );
        elsif var = "drops" then
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Drops( val.As_Integer );
        elsif var = "ammo" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Ammo( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Ammo( val.As_Integer );
        elsif var = "lives" then
            A_Scoreboard(this.Get_Widget( "scoreboard" )).Set_Lives( val.As_Integer );
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Lives( val.As_Integer );
        elsif var = "nextlife" then
            A_Progress_Board(this.Get_Widget( "progressBoard" )).Set_Next_Life( val.As_Integer );
        end if;
        Delete( val );
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_New_World_Event ) is
        pragma Unreferenced( evt );
    begin
        -- begin ignoring all action keys until the next press event
        this.ignoreKey(KEY_LEFT)     := True;
        this.ignoreKey(KEY_RIGHT)    := True;
        this.ignoreKey(KEY_UP)       := True;
        this.ignoreKey(KEY_DOWN)     := True;
        this.ignoreKey(KEY_SPACE)    := True;
        this.ignoreKey(KEY_ENTER)    := True;
        this.ignoreKey(KEY_LCONTROL) := True;
        this.ignoreKey(KEY_ALT)      := True;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_Scroll_View_Event ) is
    begin
        A_Scene(this.Get_Widget( "scene" )).Scroll( evt.Get_X, evt.Get_Y );
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Keen_View'Class;
                      evt  : not null A_World_Property_Changed_Event ) is
    begin
        if evt.Get_Property_Name = "music" then
            if evt.Get_Value'Length > 0 then
                Queue_Play_Music( evt.Get_Value );
            else
                Queue_Stop_Music;
            end if;
        elsif evt.Get_Property_Name = "introduction" then
            this.loadScreenStarted := Clock;
            this.Set_Loading_Text( evt.Get_Value );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Close_Request( this : access Keen_View ) is
         pragma Unreferenced( this );
    begin
         Queue_Close_Window;
    end Handle_Close_Request;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Keen_View;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        Game_View(this.all).Handle_Event( evt, resp );
        if evt /= null then

            if evt.Get_Id = GAME_VAR_CHANGED_ID then
                this.Handle( A_Game_Var_Changed_Event(evt) );

            elsif evt.Get_Id = FOLLOW_ENTITY_ID then
                this.playerId := A_Follow_Entity_Event(evt).Get_Id;

            elsif evt.Get_Id = SCROLL_VIEW_ID then
                this.Handle( A_Scroll_View_Event(evt) );

            elsif evt.Get_Id = NEW_WORLD_ID then
                this.Handle( A_New_World_Event(evt) );

            elsif evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
                this.Handle( A_World_Property_Changed_Event(evt) );

            elsif evt.Get_Id = GAME_STATE_ID then
                this.Handle( A_Game_State_Event(evt) );

            end if;
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Loading( this : access Keen_View; loading : Boolean ) is
    begin
        this.loading := loading;
        if this.gameInProgress then
            -- the loading screen is shown when loading starts, but it isn't
            -- hidden here when loading ends. the Tick procedure is responsible
            -- for hiding the screen.
            if loading then
                this.Show_Game( False );
                this.Show_Loading( True, "One moment" );
            end if;
        end if;
    end Handle_Loading;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Paused( this : access Keen_View; paused : Boolean ) is
    begin
        this.Get_Widget( "pnlPaused" ).Set_Visible( paused );
    end Handle_Paused;

    ----------------------------------------------------------------------------

    overriding
    procedure Start( this : access Keen_View ) is
    begin
        Game_View(this.all).Start;
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Follow_Entity" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "New_World" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );
    end Start;

    ----------------------------------------------------------------------------

    overriding
    procedure Stop( this : access Keen_View ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Scroll_View" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "New_World" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_Var_Changed" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Game_State" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Follow_Entity" );
        Game_View(this.all).Stop;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Tick( this : access Keen_View; upTime, dt : Time_Span ) is
    begin
        Game_View(this.all).Tick( upTime, dt );
        if not this.loading and then Clock - MIN_LOAD_DELAY > this.loadScreenStarted then
            this.Ready_To_Play;
            this.loadScreenStarted := Time_Last;
        end if;
    end Tick;

    --==========================================================================

    procedure Action_Title_OK( this : not null access Keen_View'Class ) is
    begin
        this.Get_Widget( "pnlTitle" ).Set_Visible( False );
        this.Show_Menu( True );
    end Action_Title_OK;

    ----------------------------------------------------------------------------

    procedure Action_New_Game( this : not null access Keen_View'Class ) is
    begin
        if this.gameInProgress then
            Queue_End_Game;
        end if;
        Queue_New_Game;
    end Action_New_Game;

    ----------------------------------------------------------------------------

    procedure Action_Resume_Game( this : not null access Keen_View'Class ) is
    begin
        if this.gameInProgress then
            this.Show_Menu( False );
            this.Show_Game( True );

            -- auto-resume
            this.Auto_Pause( False );
        end if;
    end Action_Resume_Game;

    ----------------------------------------------------------------------------

    procedure Action_Quit( this : not null access Keen_View'Class ) is
        pragma Unreferenced( this );
    begin
        Queue_Close_Window;
    end Action_Quit;

    --==========================================================================

    procedure Auto_Pause( this  : not null access Keen_View'Class;
                          pause : Boolean ) is
    begin
        if this.gameInProgress then
            if pause then
                if not this.paused then
                    this.Pause_Game( True, force => True );
                    this.autoPaused := True;
                end if;
            else
                if this.autoPaused then
                    this.Pause_Game( False, force => True );
                    this.autoPaused := False;
                end if;
            end if;
            this.pauseEnabled := not pause;
        end if;
    end Auto_Pause;

    ----------------------------------------------------------------------------

    procedure Pause_Game( this  : not null access Keen_View'Class;
                          pause : Boolean;
                          force : Boolean := False ) is
    begin
        if this.gameInProgress then
            if pause = not this.paused then
                if this.pauseEnabled or else force then
                    Queue_Pause_Game( pause );
                end if;
            end if;
        end if;
    end Pause_Game;

    ----------------------------------------------------------------------------

    procedure Populate_View( this  : not null access Keen_View'Class;
                             xres,
                             yres  : Positive;
                             scale : Positive ) is
        win : A_Window := Create_Window( this, "win",
                                         xres, yres, scale,
                                         To_Filter( Get_Pref( "scaling_filter" ) ) );
        scene         : A_Scene;
        scoreboard    : A_Scoreboard;
        progressBoard : A_Progress_Board;
        label         : A_Label;
        panel,
        panel2        : A_Panel;
        kpanel        : A_Keen_Panel;
        button        : A_Button;
    begin
        win.Add_Listener( KEY_F4, (ALT=>Yes,others=>No), Action_Quit_Key'Access );

        -- the background color seen during loading
        panel := Create_Panel( this, "background" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Color( Foreground, base_colors(Black) );
        panel.Set_Visible( False );
        win.Add( A_Widget(panel) );

        -- container for the scene and other gameplay widgets
        panel := Create_Panel( this, "pnlGame" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Visible( False );

            -- game scene
            scene := Create_Scene( this, "scene" );
            scene.Set_Layout_LTRB( 0, 0, 0, 0 );
            scene.Set_Enabled( False );
            scene.Add_Listener( KEY_LEFT,     Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_RIGHT,    Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_UP,       Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_DOWN,     Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_SPACE,    Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_LCONTROL, Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_RCONTROL, Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_ALT,      Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_ALTGR,    Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_ENTER,    Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_P,        Scene_Key_Handler'Access );
            scene.Add_Listener( KEY_ESC,      Scene_Key_Handler'Access );
            panel.Add( A_Widget(scene) );

            -- scoreboard
            scoreboard := Create_Scoreboard( this, "scoreboard", "interface" );
            scoreboard.Set_Layout_LTWH( 7, 4, 0, 0 );
            scoreboard.Set_Next( "scene" );
            panel.Add( A_Widget(scoreboard) );

            -- paused dialog
            kpanel := Create_Panel( this, "pnlPaused", "interface" );
            kpanel.Set_Layout_Center( 64, 40 );
            kpanel.Set_Focusable( False );
            kpanel.Set_Next( "scene" );
            kpanel.Set_Visible( False );
                -- paused text label
                label := Create_Label( this, "lblPaused", "Paused" );
                label.Set_Layout_Center( 0, 0 );
                label.Set_Color( Foreground, kpanel.Get_Color(Background) );
                label.Set_Focusable( False );
                kpanel.Add( A_Widget(label) );
            panel.Add( A_Widget(kpanel) );

            -- drop down progress board
            progressBoard := Create_Progress_Board( this, "progressBoard", "interface" );
            progressBoard.Add_Listener( Progress_Board_Handler'Access );
            progressBoard.Set_Next( "scene" );
            panel.Add( A_Widget(progressBoard) );
        win.Add( A_Widget(panel) );

        -- main menu
        panel := Create_Panel( this, "pnlMenuMain" );
        panel.Set_Layout_LTRB( 0, 0, 0, 0 );
        panel.Set_Border( Border_None );
        panel.Set_Color( Foreground, base_colors(Black) );
        panel.Set_Visible( False );
            -- main menu background
            label := Create_Label( this, "lblMenuMainBG", icon=>"main_menu" );
            label.Set_Layout_LTRB( 0, 0, 0, 0 );
            panel.Add( A_Widget(label) );
            -- main menu options
            panel2 := Create_Panel( this, "pnlMenuMainOptions" );
            panel2.Set_Layout_LTRB( 77, 62, -231, -130 );
            panel2.Set_Border( Border_None );
            panel2.Set_Color( Foreground, base_colors(Transparent) );
                -- New Game button
                button := Create_Push_Button( this, "btnNewGame", "New Game" );
                button.Set_Layout_LTRB( 0, 0, 0, -19 );
                button.Set_Border( Border_None );
                button.Set_Align( Align_Left );
                button.Set_Color( Text, Palette_16.Light_Green );
                button.Set_Color( Selected, Palette_16.Dark_Grey );
                button.Set_Color( Foreground, Palette_16.Dark_Grey );
                button.Add_Listener( Action_New_Game_Button'Access );
                panel2.Add( A_Widget(button) );
                -- Resume Game button
                button := Create_Push_Button( this, "btnResumeGame", "Resume Game" );
                button.Set_Layout_LTRB( 0, 20, 0, -39 );
                button.Set_Border( Border_None );
                button.Set_Align( Align_Left );
                button.Set_Color( Text, Palette_16.Light_Green );
                button.Set_Color( Selected, Palette_16.Dark_Grey );
                button.Set_Color( Foreground, Palette_16.Dark_Grey );
                button.Set_Enabled( False );
                button.Add_Listener( Action_Resume_Game_Button'Access );
                panel2.Add( A_Widget(button) );
                -- Quit Game button
                button := Create_Push_Button( this, "btnQuitGame", "Quit Game" );
                button.Set_Layout_LTRB( 0, 40, 0, -59 );
                button.Set_Border( Border_None );
                button.Set_Align( Align_Left );
                button.Set_Color( Text, Palette_16.Light_Green );
                button.Set_Color( Selected, Palette_16.Dark_Grey );
                button.Set_Color( Foreground, Palette_16.Dark_Grey );
                button.Add_Listener( Action_Quit_Button'Access );
                panel2.Add( A_Widget(button) );
            panel.Add( A_Widget(panel2) );
        win.Add( A_Widget(panel) );

        -- title screen
        button := Create_Push_Button( this, "pnlTitle", icon=>"title" );
        button.Set_Layout_LTRB( 0, 0, 0, 0 );
        button.Set_Border( Border_None );
        button.Set_Visible( True );
        button.Add_Listener( Action_Title_OK_Button'Access );
        win.Add( A_Widget(button) );
        win.Give_Focus( this.Get_Widget( "pnlTitle" ) );

        -- loading dialog
        kpanel := Create_Panel( this, "pnlLoading", "interface" );
        kpanel.Set_Layout_Center( 222, 78 );
        kpanel.Set_Visible( False );
            -- loading icon
            label := Create_Label( this, "lblLoadIcon", icon=>"loading_keen1" );
            label.Set_Layout_LTWH( 1, 10, 0, 0 );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            kpanel.Add( A_Widget(label) );
            -- loading labels
            label := Create_Label( this, "lblLoadText1" );
            label.Set_Layout_LTRB( 51, 4, 2, -19 );
            label.Set_Align( Align_Left );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( size => 11 );
            kpanel.Add( A_Widget(label) );
            label := Create_Label( this, "lblLoadText2" );
            label.Set_Layout_LTRB( 51, 20, 2, -35 );
            label.Set_Align( Align_Left );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( size => 11 );
            kpanel.Add( A_Widget(label) );
            label := Create_Label( this, "lblLoadText3" );
            label.Set_Layout_LTRB( 51, 36, 2, -51 );
            label.Set_Align( Align_Left );
            label.Set_Color( Foreground, kpanel.Get_Color(Background) );
            label.Set_Font( size => 11 );
            kpanel.Add( A_Widget(label) );
        win.Add( A_Widget(kpanel) );

        this.Set_Window( win );
    end Populate_View;

    ----------------------------------------------------------------------------

    procedure Ready_To_Play( this : not null access Keen_View'Class ) is
    begin
        this.Show_Loading( False );
        this.Show_Game( True );
        Queue_View_Ready( True );
    end Ready_To_Play;

    ----------------------------------------------------------------------------

    procedure Return_To_Menu( this : not null access Keen_View'Class ) is
    begin
        this.Auto_Pause( True );
        this.Show_Game( False  );
        this.Show_Menu( True );
    end Return_To_Menu;

    ----------------------------------------------------------------------------

    procedure Set_Loading_Text( this : not null access Keen_View'Class;
                                text : String ) is
        theme     : constant A_Theme := Get_Application.Get_Theme;
        maxLines  : constant := 3;
        line      : Positive := 1;
        label     : A_Label := A_Label(this.Get_Widget( "lblLoadText" & Image( line ) ));
        fontName  : constant String := label.Get_Font_Name;
        fontSize  : constant Positive := label.Get_Font_Size;
        lineWidth : constant Natural := label.Get_Width - theme.Border_Width * 2;
        lineText  : Unbounded_String;

        ------------------------------------------------------------------------

        procedure Add_Word( word : String ) is
        begin
            -- add the word to the current line if we're on the last line or
            -- nothing is on this line yet or the text will fit on this line.
            if line = maxLines or else
               Length( lineText ) = 0 or else
               theme.Get_Text_Width( To_String( lineText ) & ' ' & word, fontName, fontSize ) < lineWidth
            then
                -- add word to the current line
                if Length( lineText ) = 0 then
                    Append( lineText, word );
                else
                    Append( lineText, ' ' & word );
                end if;
            else
                -- complete current line and add word to the next line
                label.Set_Text( To_String( lineText ) );
                lineText := To_Unbounded_String( word );
                line := line + 1;
                label := A_Label(this.Get_Widget( "lblLoadText" & Image( line ) ));
            end if;
        end Add_Word;

        ------------------------------------------------------------------------

    begin
        Iterate_Words( text, Add_Word'Access );

        if Length( lineText ) > 0 then
            -- complete the phrase to the current line
            label.Set_Text( To_String( lineText ) );
            line := line + 1;
        end if;

        -- clear any remaining lines
        for i in line..maxLines loop
            A_Label(this.Get_Widget( "lblLoadText" & Image( i ) )).Set_Text( "" );
        end loop;
    end Set_Loading_Text;

    ----------------------------------------------------------------------------

    procedure Show_Game( this : not null access Keen_View'Class; show : Boolean ) is
    begin
        if show then
            this.Get_Widget( "scene" ).Set_Enabled( True );
            this.Get_Widget( "pnlGame" ).Set_Visible( True );
            this.win.Give_Focus( this.Get_Widget( "scene" ) );
        else
            this.Get_Widget( "pnlGame" ).Set_Visible( False );
            this.Get_Widget( "scene" ).Set_Enabled( False );
        end if;
    end Show_Game;

    ----------------------------------------------------------------------------

    procedure Show_Loading( this : not null access Keen_View'Class;
                            show : Boolean;
                            text : String := "" ) is
    begin
        if show and then not this.Get_Widget( "pnlLoading" ).Is_Visible then
            this.Get_Widget( "background" ).Set_Visible( True );
            this.Set_Loading_Text( text );
            this.win.Set_Modal( this.Get_Widget( "pnlLoading" ) );
        elsif not show and then this.Get_Widget( "pnlLoading" ).Is_Visible then
            this.win.Set_Modal( null );
            this.Get_Widget( "background" ).Set_Visible( False );
        end if;
    end Show_Loading;

    ----------------------------------------------------------------------------

    procedure Show_Menu( this : not null access Keen_View'Class; show : Boolean ) is
    begin
        if show then
            this.Get_Widget( "btnResumeGame" ).Set_Enabled( this.gameInProgress );
        end if;
        this.Get_Widget( "pnlMenuMain" ).Set_Visible( show );
    end Show_Menu;

begin

    Preferences.Set_Default( "scaling_filter", "Nearest" );

    Register_Allocator( Create_Game_View'Access );

end Game_Views.Keen;
