
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
--with Debugging;                         use Debugging;
with Entities.Players;
with Events.Audio;                      use Events.Audio;
with Events.Entities;                   use Events.Entities;
with Events.Game;                       use Events.Game;
with Events.Keen;                       use Events.Keen;
with Impulses;                          use Impulses;
with Physics;
with Preferences;
with Worlds;                            use Worlds;

pragma Elaborate_All( Entities.Players );
pragma Elaborate_All( Preferences );

package body Entities.Players.Keen is

    -- The amount of time Keen spends walking into a doorway when teleporting
    TELEPORT_DURATION : constant Time_Span := Milliseconds( 600 );

    -- The number of pixels that Keen moves up into a doorway when teleporting
    TELEPORT_DEPTH : constant Float := 16.0;

    -- Time between shots
    SHOOT_DURATION : constant Time_Span := Milliseconds( 215 );

    -- Speed of blaster shots in world units per second
    SHOT_SPEED : constant Float := 160.0;

    -- Distance that Keen should move (in either axis) while dying, before death
    DEATH_DISTANCE : constant Float := 200.0;

    Finish_Teleporting : constant Hashed_String := To_Hashed_String( "Finish_Teleporting" );

    ----------------------------------------------------------------------------

    type Frame_Type is
        record
            filename : String_Access := null;
            id       : Integer := -1;
        end record;

    type Frame_Array is array (Integer range <>) of Frame_Type;

    type Direction_Array is array (Direction_8 range <>) of access Frame_Array;

    type Animation( directional : Boolean ) is
        record
            looping    : Boolean;
            frameDelay : Time_Span;
            case directional is
                when True =>
                    dir : access Direction_Array := null;
                when False =>
                    frames : access Frame_Array := null;
            end case;
        end record;

    ----------------------------------------------------------------------------

    type Visible_Action is (
        Standing,
        Standing_and_Shooting,
        Looking,
        Walking,
        Jumping,                -- jumping upward
        Jumping_Midair,         -- hanging in midair during a jump
        Falling,                -- falling downward
        Jumping_and_Shooting,   -- shooting in the air (jumping or falling)
        Pogoing_Midair,
        Pogoing_On_Ground,
        Hanging_From_Pole,
        Hanging_From_Pole_And_Shooting,
        Climbing_Pole,
        Hanging_From_Ledge,
        Climbing_Ledge,
        Looking_Bored,
        Checking_Watch_Bored,
        Shrugging_Bored,
        Sitting_Bored,
        Reading_Bored,
        Rising_Bored,
        Entering_Doorway,
        Flipping_Switch,
        Dying
    );

    ----------------------------------------------------------------------------

    type Idle_Animation is
        record
            visibleAction : Visible_Action;
            idleDelay     : Time_Span;
            anmDuration   : Time_Span;
        end record;

    type Idle_Animation_Array is array (Positive range <>) of Idle_Animation;

    INDEX_LOOKING_BORED   : constant := 1;
    INDEX_CHECKING_WATCH  : constant := 2;
    INDEX_SHRUGGING_BORED : constant := 3;
    INDEX_SITTING_BORED   : constant := 4;
    INDEX_READING_BORED   : constant := 5;

    idle_animations : constant Idle_Animation_Array := Idle_Animation_Array'(
        INDEX_LOOKING_BORED   => (Looking_Bored,        idleDelay => Milliseconds( 4000 ), anmDuration => Milliseconds( 700 )),
        INDEX_CHECKING_WATCH  => (Checking_Watch_Bored, idleDelay => Milliseconds( 7700 ), anmDuration => Milliseconds( 1000 )),
        INDEX_SHRUGGING_BORED => (Shrugging_Bored,      idleDelay => Milliseconds( 11750 ), anmDuration => Milliseconds( 125 ) * 35),
        INDEX_SITTING_BORED   => (Sitting_Bored,        idleDelay => Milliseconds( 19000 ), anmDuration => Milliseconds( 115 ) * 4),
        INDEX_READING_BORED   => (Reading_Bored,        idleDelay => Milliseconds( 19000 ), anmDuration => Seconds( 84600 ))
    );

    ----------------------------------------------------------------------------

    type Action_Frames is array (Visible_Action) of access Animation;

    frames : constant Action_Frames := Action_Frames'(
        Standing => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("stand_left.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("stand_right.bmp"), others => <>)))),
        Standing_and_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_left.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_right.bmp"), others => <>)),
                D8_Up    => new Frame_Array'(
                    1 => (filename => new String'("shoot_up.bmp"), others => <>)))),
        Looking => new Animation'(
            directional => True,
            looping     => False,
            frameDelay  => Milliseconds( 180 ),
            dir         => new Direction_Array'(
                D8_Up   => new Frame_Array'(
                    1 => (filename => new String'("look_up.bmp"), others => <>)),
                D8_Down => new Frame_Array'(
                    1 => (filename => new String'("look_down1.bmp"), others => <>),
                    2 => (filename => new String'("look_down2.bmp"), others => <>)))),
        Walking => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("walk_left1.bmp"), others => <>),
                    (filename => new String'("walk_left2.bmp"), others => <>),
                    (filename => new String'("walk_left3.bmp"), others => <>),
                    (filename => new String'("walk_left4.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("walk_right1.bmp"), others => <>),
                    (filename => new String'("walk_right2.bmp"), others => <>),
                    (filename => new String'("walk_right3.bmp"), others => <>),
                    (filename => new String'("walk_right4.bmp"), others => <>)))),
        Jumping => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left1.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right1.bmp"), others => <>)))),
        Jumping_Midair => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left2.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right2.bmp"), others => <>)))),
        Falling => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("jump_left3.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("jump_right3.bmp"), others => <>)))),
        Jumping_and_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left       => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_left.bmp"), others => <>)),
                D8_Right      => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_right.bmp"), others => <>)),
                D8_Up         => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up.bmp"), others => <>)),
                D8_Up_Left    => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up.bmp"), others => <>)),
                D8_Up_Right   => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_up.bmp"), others => <>)),
                D8_Down       => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down.bmp"), others => <>)),
                D8_Down_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down.bmp"), others => <>)),
                D8_Down_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_jump_down.bmp"), others => <>)))),
        Pogoing_Midair => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pogo_left_air.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pogo_right_air.bmp"), others => <>)))),
        Pogoing_On_Ground => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pogo_left_ground.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pogo_right_ground.bmp"), others => <>)))),
        Hanging_From_Pole => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("pole_left1.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("pole_right1.bmp"), others => <>)))),
        Hanging_From_Pole_And_Shooting => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left       => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_left.bmp"), others => <>)),
                D8_Right      => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_right.bmp"), others => <>)),
                D8_Up         => null,
                D8_Down       => null,
                D8_Down_Left  => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_down_left.bmp"), others => <>)),
                D8_Down_Right => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_down_right.bmp"), others => <>)),
                D8_Up_Left    => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_up_left.bmp"), others => <>)),
                D8_Up_Right   => new Frame_Array'(
                    1 => (filename => new String'("shoot_pole_up_right.bmp"), others => <>)))),
        Climbing_Pole => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("pole_left1.bmp"), others => <>),
                    (filename => new String'("pole_left2.bmp"), others => <>),
                    (filename => new String'("pole_left3.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("pole_right1.bmp"), others => <>),
                    (filename => new String'("pole_right2.bmp"), others => <>),
                    (filename => new String'("pole_right3.bmp"), others => <>)),
                D8_Up    => null,
                D8_Down  => new Frame_Array'(
                    (filename => new String'("pole_down1.bmp"), others => <>),
                    (filename => new String'("pole_down2.bmp"), others => <>),
                    (filename => new String'("pole_down3.bmp"), others => <>),
                    (filename => new String'("pole_down4.bmp"), others => <>)))),
        Hanging_From_Ledge => new Animation'(
            directional => True,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    1 => (filename => new String'("ledge_left1.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    1 => (filename => new String'("ledge_right1.bmp"), others => <>)))),
        Climbing_Ledge => new Animation'(
            directional => True,
            looping     => False,
            frameDelay  => Milliseconds( 115 ),
            dir         => new Direction_Array'(
                D8_Left  => new Frame_Array'(
                    (filename => new String'("ledge_left2.bmp"), others => <>),
                    (filename => new String'("ledge_left3.bmp"), others => <>),
                    (filename => new String'("ledge_left4.bmp"), others => <>),
                    (filename => new String'("ledge_left5.bmp"), others => <>)),
                D8_Right => new Frame_Array'(
                    (filename => new String'("ledge_right2.bmp"), others => <>),
                    (filename => new String'("ledge_right3.bmp"), others => <>),
                    (filename => new String'("ledge_right4.bmp"), others => <>),
                    (filename => new String'("ledge_right5.bmp"), others => <>)))),
        Looking_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_LOOKING_BORED).anmDuration,
            frames      => new Frame_Array'(
                1 => (filename => new String'("look_up.bmp"), others => <>))),
        Shrugging_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_SHRUGGING_BORED).anmDuration / 35,
            frames      => new Frame_Array'(
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug2.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug2.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug1.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>),
                (filename => new String'("bored_shrug3.bmp"), others => <>))),
        Checking_Watch_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_CHECKING_WATCH).anmDuration,
            frames      => new Frame_Array'(
                1 => (filename => new String'("bored_watch.bmp"), others => <>))),
        Sitting_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => idle_animations(INDEX_SITTING_BORED).anmDuration / 4,
            frames      => new Frame_Array'(
                (filename => new String'("bored_sit1.bmp"), others => <>),
                (filename => new String'("bored_sit2.bmp"), others => <>),
                (filename => new String'("bored_sit3.bmp"), others => <>),
                (filename => new String'("bored_sit4.bmp"), others => <>))),
        Reading_Bored => new Animation'(
            directional => False,
            looping     => True,
            frameDelay  => Milliseconds( 250 ),
            frames      => new Frame_Array'(
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read1.bmp"), others => <>),
                (filename => new String'("bored_read2.bmp"), others => <>),
                (filename => new String'("bored_read3.bmp"), others => <>))),
        Rising_Bored => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => Milliseconds( 125 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("bored_rise1.bmp"), others => <>),
                2 => (filename => new String'("bored_rise2.bmp"), others => <>))),
        Entering_Doorway => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => TELEPORT_DURATION / 5,
            frames      => new Frame_Array'(
                (filename => new String'("doorway1.bmp"), others => <>),
                (filename => new String'("doorway2.bmp"), others => <>),
                (filename => new String'("doorway3.bmp"), others => <>),
                (filename => new String'("doorway4.bmp"), others => <>),
                (filename => new String'("doorway5.bmp"), others => <>))),
        Flipping_Switch => new Animation'(
            directional => False,
            looping     => False,
            frameDelay  => Milliseconds( 115 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("switch.bmp"), others => <>))),
        Dying => new Animation'(
            directional => False,
            looping     => True,
            frameDelay  => Milliseconds( 115 ),
            frames      => new Frame_Array'(
                1 => (filename => new String'("die1.bmp"), others => <>)))
    );

    ----------------------------------------------------------------------------

    function Get_Pref( shortName : String ) return Float is
    begin
        return Preferences.Get_Pref( "development", "entity.player.keen." & shortName );
    end Get_Pref;
    pragma Inline( Get_Pref );

    ----------------------------------------------------------------------------

    function Get_Pref( shortName : String ) return String is
    begin
        return Preferences.Get_Pref( "development", "entity.player.keen." & shortName );
    end Get_Pref;
    pragma Inline( Get_Pref );

    ----------------------------------------------------------------------------

    function To_X( xv : Float; dir : Direction_Type ) return Direction_8 is
    begin
        if xv < 0.0 then
            return D8_Left;
        elsif xv > 0.0 then
            return D8_Right;
        end if;
        return To_X( dir );
    end To_X;
    pragma Inline( To_X );

    ----------------------------------------------------------------------------

    function To_D8( xv : Float; dir : Direction_Type ) return Direction_8 is
    begin
        if xv < 0.0 then
            return To_D8( dir + Left );
        elsif xv > 0.0 then
            return To_D8( dir + Right );
        end if;
        return To_D8( dir );
    end To_D8;
    pragma Inline( To_D8 );

    --==========================================================================

    function Create_Player return A_Entity is
        this : constant A_Entity := new Keen;
    begin
        this.Construct;
        return this;
    end Create_Player;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen ) is
    begin
        Player(this.all).Construct( 8, 32, Get_Pref( "library" ) );
        this.minJumpHeight := Get_Pref( "min_jump_height" );
        this.maxJumpHeight := Get_Pref( "max_jump_height" );
        this.minPogoHeight := Get_Pref( "min_pogo_height" );
        this.maxPogoHeight := Get_Pref( "max_pogo_height" );
        this.maxBounceHeight := Get_Pref( "max_bounce_height" );
        this.lookSpeed := Get_Pref( "look_speed" );
        this.Update_Frame( notify => False );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Die( this : access Keen ) is
        val : A_Value;
    begin
        -- keen is temporarily invincible while teleporting
        if this.action = Act_Teleporting then
            return;
        end if;

        this.Start_Action( Act_Dying );
        this.deathX := this.x;
        this.deathY := this.y;
        this.jumping := True;
        this.jumpStartY := this.y;

        Queue_Follow_Entity( 0 );
        val := Create_Value( False );
        Queue_Set_Entity_Attribute( this.id, "clipped", val );
        Queue_Accelerate( this.id, Up,
                          Get_Pref( "jump_speed" ),
                          Physics.INSTANT_ACCELERATION );
        Queue_Accelerate( this.id, Right,
                          Get_Pref( "walk_speed" ) / 2.0,
                          Physics.INSTANT_ACCELERATION );
        Queue_Play_Sound( "keen_die.wav" );

        -- TODO: Disable save, return to menu, and pause functions in the view
    end Die;

    ----------------------------------------------------------------------------

    overriding
    procedure Face( this : access Keen; dir : Direction_Type ) is
    begin
        if dir = Dir_Left or else dir = Dir_Right then
            this.dir := dir;
            this.Update_Frame;
        end if;
    end Face;

    ----------------------------------------------------------------------------

    overriding
    procedure Impulse( this : access Keen; name : Hashed_String ) is

        procedure Activate_Entities is

            procedure Activate( e : A_Entity ) is
            begin
                e.On_Activate( A_Entity(this) );
            end Activate;

        begin
            -- only activate entities if the player is standing on the ground
            -- and not already activating something.
            if not this.activating and then
               this.action = Act_Moving and then
               this.grounded and then
               not this.jumping and then
               not this.onPogo
            then
                this.Iterate_Touching( Activate'Access );
            end if;
        end Activate_Entities;

    begin
        if name = Move_Left_Start then
            this.Move_Start( Left );

        elsif name = Move_Left_Stop then
            this.Move_Stop( Left );

        elsif name = Move_Right_Start then
            this.Move_Start( Right );

        elsif name = Move_Right_Stop then
            this.Move_Stop( Right );

        elsif name = Move_Up_Start then
            Activate_Entities;
            this.Move_Start( Up );

        elsif name = Move_Up_Stop then
            this.activating := False;
            this.Move_Stop( Up );

        elsif name = Move_Down_Start then
            this.Move_Start( Down );

        elsif name = Move_Down_Stop then
            this.Move_Stop( Down );

        elsif name = Jump_Start then
            -- keen is attempting to jump
            if this.onPogo or else not this.attemptingJump then
                -- keen is allowed to jump if he wasn't already attempting, or
                -- if he's on the pogo stick. (don't allow keen to normal jump
                -- multiple times while holding the jump key)
                this.attemptingJump := True;
                this.Start_Jump;
            end if;

        elsif name = Jump_Stop then
            -- keen is no longer attempting to jump
            this.attemptingJump := False;
            if this.jumping then
                this.Stop_Jump;
            end if;

        elsif name = Toggle_Pogo then
            -- toggle the pogostick on or off if possible
            if this.action = Act_Moving then
                -- toggle the pogo
                this.onPogo := not this.onPogo;

            elsif this.action = Act_Idle then
                -- interrupt the idle animation if possible
                if this.idleNumber = INDEX_READING_BORED then
                    this.Start_Action( Act_Rising,
                                       frames(Rising_Bored).frameDelay *
                                           frames(Rising_Bored).frames'Length );
                    this.Reset_Idle;
                elsif this.idleNumber /= INDEX_SITTING_BORED then
                    this.Start_Action( Act_Moving );
                    this.Reset_Idle;
                end if;

            end if;

        elsif name = Shoot then
            this.Shoot_Gun;

        elsif name = Finish_Teleporting then
            if this.action = Act_Teleporting then
                this.Start_Action( Act_Moving );
                this.lastTeleport := this.age;
            end if;

        end if;
    end Impulse;

    ----------------------------------------------------------------------------

    function Is_Busy( this : not null access Keen'Class ) return Boolean is
    begin
        if this.actionDuration > Time_Span_Zero then
            return this.actionStart + this.actionDuration > this.age;
        end if;
        return False;
    end Is_Busy;

    ----------------------------------------------------------------------------

    overriding
    procedure Move_Start( this : access Keen; dir : Cardinal_Direction ) is
        acc : Float;
    begin
        case dir is
            when Left | Right =>
                if this.action = Act_Looking then
                    -- if keen is looking then he must stop it before moving
                    if this.dir and Up then
                        -- stop looking up
                        this.Move_Stop( Up );
                    else
                        -- stop looking down
                        this.Move_Stop( Down );
                    end if;
                    return;

                elsif this.action = Act_Moving then
                    -- keen can begin moving if he isn't already moving in a direction
                    if not this.shooting and then
                       not this.moving(Left) and then not this.moving(Right) then
                        this.moving(dir) := True;
                        this.dir := this.dir + dir;

                        if this.onPogo then
                            -- keen is moving in the air on his pogo stick
                            acc := Get_Pref( "midair_pogo_acceleration" );
                        elsif not this.grounded then
                            -- keen is moving in the air without his pogo stick
                            acc := Get_Pref( "midair_acceleration" );
                        else
                            -- keen is walking along the ground
                            acc := Get_Pref( "walk_acceleration" );
                        end if;

                        Queue_Accelerate( this.id, dir, Get_Pref( "walk_speed" ), acc );
                        this.Update_Frame;
                    end if;

                elsif this.action = Act_Idle then
                    -- keen can begin moving if he isn't sitting down reading
                    if this.idleNumber = INDEX_READING_BORED then
                        this.Start_Action( Act_Rising,
                                           frames(Rising_Bored).frameDelay *
                                               frames(Rising_Bored).frames'Length );
                        this.Reset_Idle;
                    elsif this.idleNumber /= INDEX_SITTING_BORED then
                        this.Start_Action( Act_Moving );
                        this.Reset_Idle;
                    end if;

                end if;

            when Up | Down =>
                if this.action = Act_Moving then
                    this.dir := this.dir + dir;

                    -- if keen is on the ground
                    if not this.jumping and then
                       this.grounded and then
                       not this.onPogo and then
                       not this.shooting
                    then
                        -- todo: check if keen should climb a pole

                        -- if keen is standing still
                        if not this.moving then
                            -- begin the looking action and freeze the player
                            this.Start_Action( Act_Looking,
                                               frames(Looking).frameDelay *
                                                   frames(Looking).dir(To_D8( dir ))'Length );
                       end if;
                    end if;

                elsif this.action = Act_Idle then
                    -- keen can abort idle animations if he isn't sitting down
                    -- reading
                    if this.idleNumber = INDEX_READING_BORED then
                        this.Start_Action( Act_Rising,
                                           frames(Rising_Bored).frameDelay *
                                               frames(Rising_Bored).frames'Length );
                        this.Reset_Idle;
                    elsif this.idleNumber /= INDEX_SITTING_BORED then
                        this.Start_Action( Act_Moving );
                        this.Reset_Idle;
                    end if;

                end if;

        end case;
    end Move_Start;

    ----------------------------------------------------------------------------

    overriding
    procedure Move_Stop( this : access Keen; dir : Cardinal_Direction ) is
    begin
        case dir is
            when Left | Right =>
                if this.action = Act_Moving then
                    -- stop moving if 'dir' is the direction keen was moving
                    if this.moving(dir) then
                        this.moving(dir) := False;
                        Queue_Accelerate( this.id, dir, 0.0, 0.0 );
                        this.Update_Frame;
                    end if;
                end if;

            when Up | Down =>
                if this.action = Act_Moving then
                    this.dir := this.dir - dir;

                elsif this.action = Act_Looking then
                    -- stop looking if 'dir' is the direction keen was looking
                    if this.dir and dir then
                        -- begin the action of looking back. use one frame fewer
                        -- in the animation because the last frame of the looking
                        -- animation is the steady looking frame.
                        this.Start_Action( Act_Looking_Back,
                                           frames(Looking).frameDelay *
                                               (frames(Looking).dir(To_D8( dir ))'Length - 1) );
                    end if;
                end if;
        end case;
    end Move_Stop;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Keen is
        this : aliased Keen;
    begin
        this.Construct;
        Keen'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Keen ) is
    begin
        Player'Read( stream, Player(obj) );
        obj.jumping := Boolean'Input( stream );
        obj.jumpStartY := Float'Input( stream );
        obj.onPogo := Boolean'Input( stream );
        obj.shooting := Boolean'Input( stream );
        obj.shootStart := Time_Span'Input( stream );
        obj.action := Action_Type'Input( stream );
        obj.actionStart := Time_Span'Input( stream );
        obj.actionDuration := Time_Span'Input( stream );
        obj.idleStart := Time_Span'Input( stream );
        obj.idleNumber := Positive'Input( stream );
        obj.teleportX := Float'Input( stream );
        obj.teleportY := Float'Input( stream );
        obj.lastTeleport := Time_Span'Input( stream );
        obj.deathX := Float'Input( stream );
        obj.deathY := Float'Input( stream );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Keen ) is
    begin
        Player'Write( stream, Player(obj) );
        Boolean'Output( stream, obj.jumping );
        Float'Output( stream, obj.jumpStartY );
        Boolean'Output( stream, obj.onPogo );
        Boolean'Output( stream, obj.shooting );
        Time_Span'Output( stream, obj.shootStart );
        Action_Type'Output( stream, obj.action );
        Time_Span'Output( stream, obj.actionStart );
        Time_Span'Output( stream, obj.actionDuration );
        Time_Span'Output( stream, obj.idleStart );
        Positive'Output( stream, obj.idleNumber );
        Float'Output( stream, obj.teleportX );
        Float'Output( stream, obj.teleportY );
        Time_Span'Output( stream, obj.lastTeleport );
        Float'Output( stream, obj.deathX );
        Float'Output( stream, obj.deathY );
    end Object_Write;

    ----------------------------------------------------------------------------

    overriding
    procedure On_Hit_Wall( this : access Keen; dir : Cardinal_Direction ) is
    begin
        if dir = Up and then this.jumping then
            this.Stop_Jump;
        end if;
    end On_Hit_Wall;

    ----------------------------------------------------------------------------

    procedure Reset_Idle( this : not null access Keen'Class ) is
    begin
        this.idleStart := this.age;
        this.idleNumber := 1;
    end Reset_Idle;

    ----------------------------------------------------------------------------

    procedure Shoot_Gun( this : not null access Keen'Class ) is
    begin
        if not this.shooting then
            if this.action = Act_Moving then
                -- standing/moving on the ground or in the air
                this.onPogo := False;
                this.shooting := True;
                this.shootStart := this.age;

                if this.grounded then
                    if this.moving(Left) then
                        this.Move_Stop( Left );
                    elsif this.moving(Right) then
                        this.Move_Stop( Right );
                    end if;
                end if;
                this.Reset_Idle;

                Queue_Give_Ammo( -1 );
                Queue_Play_Sound( "keen_shoot.wav" );
                if this.dir and Up then
                    Queue_Spawn_Entity( "sprite.Stunner Shot",
                                        this.x, this.y - Float(this.height / 2),
                                        yv => -SHOT_SPEED );
                elsif this.dir and Down then
                    Queue_Spawn_Entity( "sprite.Stunner Shot",
                                        this.x, this.y + Float(this.height / 2),
                                        yv => SHOT_SPEED );
                elsif this.dir and Left then
                    Queue_Spawn_Entity( "sprite.Stunner Shot",
                                        this.x - Float(this.width / 2), this.y,
                                        xv => -SHOT_SPEED );
                elsif this.dir and Right then
                    Queue_Spawn_Entity( "sprite.Stunner Shot",
                                        this.x + Float(this.width / 2), this.y,
                                        xv => SHOT_SPEED );
                end if;

            elsif this.action = Act_Looking and then (this.dir and Up) then
                -- standing on the ground and looking up
                this.shooting := True;
                this.shootStart := this.age;

                Queue_Give_Ammo( -1 );
                Queue_Play_Sound( "keen_shoot.wav" );
                Queue_Spawn_Entity( "sprite.Stunner Shot",
                                    this.x, this.y - Float(this.height / 2),
                                    yv => -SHOT_SPEED );

            end if;
        end if;
    end Shoot_Gun;

    ----------------------------------------------------------------------------

    procedure Start_Action( this   : not null access Keen'Class;
                            action : Action_Type;
                            length : Time_Span := Time_Span_Zero ) is
    begin
        this.action := action;
        this.actionStart := this.age;
        this.actionDuration := length;
        this.Update_Frame;
    end Start_Action;

    ----------------------------------------------------------------------------

    procedure Start_Jump( this : not null access Keen'Class ) is
    begin
        -- check if keen is allowed to start a jump
        if this.action = Act_Looking and then (this.dir and Up) then
            -- if player is looking up then switch to the moving action so
            -- he can jump immediately. he can't jump if he is looking down.
            this.Start_Action( Act_Moving );
        end if;

        if this.action = Act_Moving then
            -- keen can only start jumping if he's standing on the ground and
            -- hasn't started a jump before now.
            if not this.jumping and then this.grounded and then not this.shooting then
                this.jumping := True;
                this.jumpStartY := this.y;
                Queue_Accelerate( this.id, Up,
                                  Get_Pref( "jump_speed" ),
                                  Get_Pref( "jump_acceleration" ) );
                if this.onPogo then
                    Queue_Play_Sound( "keen_pogo.wav" );
                else
                    Queue_Play_Sound( "keen_jump.wav" );
                end if;
            end if;

        elsif this.action = Act_Idle then
            -- keen can abort idle animations if he isn't sitting down reading
            if this.idleNumber = INDEX_READING_BORED then
                this.Start_Action( Act_Rising,
                                   frames(Rising_Bored).frameDelay *
                                       frames(Rising_Bored).frames'Length );
                 this.Reset_Idle;
            elsif this.idleNumber /= INDEX_SITTING_BORED then
                this.Start_Action( Act_Moving );
                this.Reset_Idle;
            end if;

        end if;
    end Start_Jump;

    ----------------------------------------------------------------------------

    procedure Stop_Jump( this : not null access Keen'Class ) is
    begin
        -- check if keen is allowed to stop the jump (he can't normally jump
        -- less than the minimum jump height)
        if this.action = Act_Moving and then this.jumping then
            if (this.onPogo and then this.jumpStartY - this.y > this.minPogoHeight) or else
               (not this.onPogo and then this.jumpStartY - this.y > this.minJumpHeight)
            then
                this.jumping := False;
                this.jumpStartY := 0.0;
                Queue_Accelerate( this.id, Up, 0.0, 0.0 );
            end if;
        end if;
    end Stop_Jump;

    ----------------------------------------------------------------------------

    not overriding
    procedure Teleport( this         : access Keen;
                        fromX, fromY : Float;
                        toX, toY     : Float ) is
        pragma Unreferenced( fromY );
    begin
        this.dir := this.dir - Up;  -- have to look up to teleport; stop looking

        -- stop moving left or right and move toward the teleporter
        if this.moving(Left) then
            this.moving(Left) := False;
        elsif this.moving(Right) then
            this.moving(Right) := False;
        end if;
        if fromX < this.x then
            Queue_Accelerate( this.id, Left,
                              (this.x - fromX) / Float(To_Duration( TELEPORT_DURATION * 2 )),
                              Physics.INSTANT_ACCELERATION );
        elsif fromX > this.x then
            Queue_Accelerate( this.id, Right,
                              (fromX - this.x) / Float(To_Duration( TELEPORT_DURATION * 2 )),
                              Physics.INSTANT_ACCELERATION );
        else
            Queue_Accelerate( this.id, Left, 0.0, Physics.INSTANT_ACCELERATION );
        end if;

        this.teleportX := toX;
        this.teleportY := toY;
        this.Start_Action( Act_Teleporting, TELEPORT_DURATION );
        Queue_Set_Entity_Attribute( this.id, "physical", False );
        Queue_Accelerate( this.id, Up,
                          TELEPORT_DEPTH / Float(To_Duration( TELEPORT_DURATION )),
                          Physics.INSTANT_ACCELERATION );
        this.activating := True;
        this.Update_Frame;
    end Teleport;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Keen; upTime, dt : Time_Span ) is
    begin
        Player(this.all).Tick( upTime, dt );

        case this.action is
            when Act_Moving =>  -- - - - - - - - - - - - - - - - - - - - - - - -
                if this.onPogo then
                    -- keen is on his pogo stick; he needs to auto-bounce
                    this.Start_Jump;
                end if;

                if this.jumping then
                    -- if keen isn't attempting to jump any more, check if he's
                    -- reached the minimum height before stopping the jump.
                    if not this.attemptingJump then
                        this.Stop_Jump;
                    end if;

                    -- if keen has jumped to the maximum height, stop the jump.
                    if (this.onPogo and then this.jumpStartY - this.y > this.maxPogoHeight) or else
                       (not this.onPogo and then this.jumpStartY - this.y > this.maxJumpHeight)
                    then
                        this.Stop_Jump;
                    end if;
                end if;

                if this.grounded and then not
                   this.jumping and then
                   not this.onPogo and then
                   not this.moving and then
                   not this.editingMode
                then
                    -- keen is standing still on the ground
                    -- check if its time to start an idle animation
                    if this.age >= this.idleStart + idle_animations(this.idleNumber).idleDelay then
                        this.Start_Action( Act_Idle, idle_animations(this.idleNumber).anmDuration );
                    end if;
                else
                    this.Reset_Idle;
                end if;

            when Act_Looking => -- - - - - - - - - - - - - - - - - - - - - - - -
                if not this.shooting and then not this.Is_Busy then
                    -- if keen is steadily looking up or down, scroll the view
                    Queue_Scroll_View( 0.0, Float(To_Duration( dt )) * (Float(this.dir.y) * this.lookSpeed) );
                end if;
                this.Reset_Idle;

            when Act_Looking_Back =>  -- - - - - - - - - - - - - - - - - - - - -
                -- if keen is done looking back then go back to moving
                if not this.Is_Busy then
                    this.Start_Action( Act_Moving );
                    this.dir.y := 0;
                end if;
                this.Reset_Idle;

            when Act_Teleporting => -- - - - - - - - - - - - - - - - - - - - - -
                -- keen is teleporting
                -- check if he's done yet
                if not this.Is_Busy then
                    -- stop moving into the doorway
                    -- become subject to gravity again
                    -- jump to the destination
                    --   this performs clipping and may queue a grounded event
                    -- end the teleporting action after the jump has been made
                    Queue_Accelerate( this.id, Left, 0.0, 0.0 );
                    Queue_Accelerate( this.id, Up, 0.0, 0.0 );
                    Queue_Set_Entity_Attribute( this.id, "physical", True );
                    Queue_Move_Entity( this.id, this.teleportX, this.teleportY );
                    Queue_Impulse( this.id, Finish_Teleporting );
                end if;
                this.Reset_Idle;

            when Act_Idle =>  -- - - - - - - - - - - - - - - - - - - - - - - - -
                -- if keen is in the air, stop being idle immediately (this can
                -- happen if the floor disappears beneath him)
                if not this.grounded then
                    this.Start_Action( Act_Moving );
                    this.Reset_Idle;

                -- check if the idle animation is done
                elsif not this.Is_Busy then
                    -- set the number for the next idle animation
                    this.Start_Action( Act_Moving );
                    if this.idleNumber = idle_animations'Last then
                        this.Reset_Idle;
                    else
                        this.idleNumber := idle_animations'First + (this.idleNumber mod idle_animations'Last);
                    end if;

                    -- check if another idle animation should begin immediately
                    if this.age >= this.idleStart + idle_animations(this.idleNumber).idleDelay then
                        this.Start_Action( Act_Idle, idle_animations(this.idleNumber).anmDuration );
                    end if;
                end if;

            when Act_Rising =>  -- - - - - - - - - - - - - - - - - - - - - - - -
                -- if keen is done looking back then go back to moving
                if not this.Is_Busy then
                    this.Start_Action( Act_Moving );
                end if;
                this.Reset_Idle;

            when Act_Dying => -- - - - - - - - - - - - - - - - - - - - - - - - -
                -- cap the height of the death bounce
                if this.jumping and then this.jumpStartY - this.y > this.maxBounceHeight then
                    this.jumping := False;
                    Queue_Accelerate( this.id, Up, 0.0, 0.0 );
                end if;

                -- check for death complete
                if not this.dead and then
                   (this.x > this.deathX + DEATH_DISTANCE or else
                    this.y > this.deathY + DEATH_DISTANCE or else
                    this.x >= Float(this.world.Get_Width - 1) or else
                    this.y >= Float(this.world.Get_Height - 1))
                then
                    Queue_Play_Sound( "keen_dead.wav" );

                    this.deathX := this.x;
                    this.deathY := this.y;
                    this.dead := True;   -- temporary to keep sound from repeating
                    Queue_End_Game;      -- hard quit. this is temporary.
                end if;

                -- if he's off the map then hard quit. this is temporary
                --if this.dead and then
                --   (this.x >= Float(this.world.Get_Width - 1) or else
                --    this.y >= Float(this.world.Get_Height - 1))
                --then
                --    Queue_End_Game;
                --end if;

                this.Reset_Idle;

        end case;

        if this.shooting and then this.age > this.shootStart + SHOOT_DURATION then
            this.shooting := False;
        end if;

        this.upTime := upTime;
        this.Update_Frame;
    end Tick;

    ----------------------------------------------------------------------------

    overriding
    procedure Update_Frame( this : access Keen; notify : Boolean := True ) is

        ------------------------------------------------------------------------

        function Get_Frame_Id( action   : Visible_Action;
                               dir      : Direction_8 := D8_Up;
                               reversed : Boolean := False ) return Natural is

            --------------------------------------------------------------------

            procedure Resolve_Id( id : out Natural; frame : in out Frame_Type ) is
            begin
                if frame.id < 0 then
                    frame.id := this.lib.Get_Id( frame.filename.all );
                end if;
                id := frame.id;
            end Resolve_Id;

            --------------------------------------------------------------------

            frameArray : access Frame_Array;
            frameNum   : Natural := 0;
            id         : Natural := 0;
        begin
            if frames(action).directional then
                frameArray := frames(action).dir(dir);
            else
                frameArray := frames(action).frames;
            end if;

            if frameArray'Length = 1 then
                frameNum := 0;
            else
                if frames(action).looping then
                    frameNum := ((this.age - this.actionStart) / frames(action).frameDelay) mod frameArray'Length;
                else
                    frameNum := Integer'Min( ((this.age - this.actionStart) / frames(action).frameDelay), frameArray'Length - 1 );
                end if;
            end if;

            if reversed then
                Resolve_Id( id, frameArray(frameArray'Last - frameNum) );
            else
                Resolve_Id( id, frameArray(frameArray'First + frameNum) );
            end if;

            return id;
        end Get_Frame_Id;

        ------------------------------------------------------------------------

        id : Natural := 0;
    begin
        -- determine keen's current visible action
        case this.action is
            when Act_Moving =>
                if this.onPogo then
                    -- on the pogostick, in the air or on the ground
                    if this.grounded then
                        id := Get_Frame_Id( Pogoing_On_Ground, To_X( this.xv, this.dir ) );
                    else
                       id := Get_Frame_Id( Pogoing_Midair, To_X( this.xv, this.dir ) );
                    end if;
                elsif this.grounded then
                    -- standing or walking on the ground
                    if this.shooting then
                        id := Get_Frame_Id( Standing_and_Shooting, To_X( this.xv, this.dir ) );
                    elsif this.xv = 0.0 then
                        id := Get_Frame_Id( Standing, To_X( this.xv, this.dir ) );
                    else
                        id := Get_Frame_Id( Walking, To_X( this.xv, this.dir ) );
                    end if;
                else
                    -- in the air
                    if this.shooting then
                        id := Get_Frame_Id( Jumping_and_Shooting, To_D8( this.xv, this.dir ) );
                    elsif this.yv < 0.0 then
                        id := Get_Frame_Id( Jumping, To_X( this.xv, this.dir ) );
                    elsif this.yv < Get_Pref( "falling_speed" ) then
                        id := Get_Frame_Id( Jumping_Midair, To_X( this.xv, this.dir ) );
                    else
                        id := Get_Frame_Id( Falling, To_X( this.xv, this.dir ) );
                    end if;
                end if;

            when Act_Looking | Act_Looking_Back =>
                -- looking up or down
                if this.shooting then
                    id := Get_Frame_Id( Standing_and_Shooting, To_Y( this.dir ) );
                else
                    id := Get_Frame_Id( Looking, To_Y( this.dir ),
                                        reversed => this.action = Act_Looking_Back );
                end if;

            when Act_Teleporting =>
                -- walking into a teleporter
                id := Get_Frame_Id( Entering_Doorway );

            when Act_Idle =>
                id := Get_Frame_Id( idle_animations(this.idleNumber).visibleAction );

            when Act_Rising =>
                id := Get_Frame_Id( Rising_Bored );

            when Act_Dying =>
                id := Get_Frame_Id( Dying );

        end case;

        this.Set_Frame( id, notify );
    end Update_Frame;

begin

    Entities.Factory.Register_Class( "player.Platform", Create_Player'Access );

    Preferences.Set_Default( "development", "entity.player.keen.library", "keen" );

    Preferences.Set_Default( "development", "entity.player.keen.walk_speed",               92.0  );
    Preferences.Set_Default( "development", "entity.player.keen.walk_acceleration",        700.0 );
    Preferences.Set_Default( "development", "entity.player.keen.midair_acceleration",      500.0 );
    Preferences.Set_Default( "development", "entity.player.keen.midair_pogo_acceleration", 350.0 );

    Preferences.Set_Default( "development", "entity.player.keen.jump_speed",        192.0  );
    Preferences.Set_Default( "development", "entity.player.keen.jump_acceleration", 2400.0 );
    Preferences.Set_Default( "development", "entity.player.keen.falling_speed",     54.0   );

    Preferences.Set_Default( "development", "entity.player.keen.min_jump_height",   10.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_jump_height",   40.0 );
    Preferences.Set_Default( "development", "entity.player.keen.min_pogo_height",   16.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_pogo_height",   80.0 );
    Preferences.Set_Default( "development", "entity.player.keen.max_bounce_height", 26.0 );

    -- The scrolling speed in world coordinates per second when Keen looks up or down.
    Preferences.Set_Default( "development", "entity.player.keen.look_speed", 80.0 );

end Entities.Players.Keen;
