
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Text_IO;
with Applications;
with Debugging;                         use Debugging;
with GNAT.Directory_Operations.Iteration;
 use GNAT.Directory_Operations.Iteration;
with Preferences;                       use Preferences;
with Support.Paths;                     use Support.Paths;
with Version;
with Worlds;                            use Worlds;

pragma Elaborate_All( Applications );
pragma Elaborate_All( Preferences );
pragma Elaborate_All( Support.Paths );

package body Applications.UpdateMap is

    function Create_Application return A_Application is
        this : constant A_Application := new UpdateMap_Application;
    begin
        this.Construct;
        return this;
    end Create_Application;

    ----------------------------------------------------------------------------

    overriding
    function Get_Name( this : access UpdateMap_Application ) return String is
        pragma Unreferenced( this );
    begin
        return "UpdateMap";
    end Get_Name;

    ----------------------------------------------------------------------------

    overriding
    function Get_Window_Title( this : access UpdateMap_Application ) return String is
        pragma Unreferenced( this );
    begin
        return "UpdateMap";
    end Get_Window_Title;

    ----------------------------------------------------------------------------

    overriding
    function Init( this : access UpdateMap_Application ) return Boolean is
    begin
        return this.Init( 128, 64, app_scale => 1, app_windowed => True );
    end Init;

    ----------------------------------------------------------------------------

    overriding
    procedure Run( this       : access UpdateMap_Application;
                   returnCode : in out Integer ) is
        pragma Unreferenced( this );

        ------------------------------------------------------------------------

        procedure Display_Title is
        begin
            Ada.Text_IO.New_Line;
            Ada.Text_IO.Put_Line( "UpdateMap v" & Version.Release &
                                  " - Kevin Wellwood " &
                                  "Copyright (C) " & Version.Copyright_Year );
            Ada.Text_IO.New_Line;
        end Display_Title;

        ------------------------------------------------------------------------

        procedure Display_Usage is
        begin
            Ada.Text_IO.Put_Line( "Usage:" );
            Ada.Text_IO.Put_Line( "  updatemap <map_path> | all" );
            Ada.Text_IO.New_Line;
            returnCode := 1;
        end Display_Usage;

        ------------------------------------------------------------------------

        procedure Update_Map( path : String ) is
            world : A_World := null;
        begin
            world := Load_World( path );
            world.Save( path );
            Ada.Text_IO.Put_Line( "Updated " & path );
        exception
            when FILE_NOT_FOUND =>
                Delete( world );
                Ada.Text_IO.Put_Line( "Skipping " & path & " - File not found" );
            when READ_EXCEPTION =>
                -- this exception occurs if the world identifier is not found,
                -- the file version is not readable, or the format is otherwise
                -- damaged.
                Delete( world );
            when e : others =>
                Delete( world );
                Ada.Text_IO.Put_Line( "Skipping " & path & " - Error loading world" );
                pragma Debug( Dbg( e, level => ERROR ) );
        end Update_Map;

        ------------------------------------------------------------------------

        procedure Examine( path  : String;
                           index : Positive;
                           quit  : in out Boolean ) is
            pragma Unreferenced( index, quit );
        begin
            Update_Map( path );
        end Examine;

        ------------------------------------------------------------------------

        procedure Iterate is new Wildcard_Iterator( Examine );
    begin
        Display_Title;

        if Argument_Count < 1 then
            Display_Usage;
            return;
        end if;

        if To_Lower( Argument( 1 ) ) = "all" then
            Iterate( "*." & World_Extension );
            Iterate( Get_Pref( "application", "media" ) & "*." & World_Extension );
        else
            Update_Map( Add_Extension( Argument( 1 ), World_Extension ) );
        end if;
    end Run;

    ----------------------------------------------------------------------------

begin

    Register_Allocator( Create_Application'Access );

    Preferences.Set_Default( "application", "media", "media" & Slash );
    Preferences.Set_Default( "application", "editing_mode", True );

end Applications.UpdateMap;
